/*  
 *
 *  Embedded Planets RPX-Lite/HIOX Touch Panel Driver
 *  Copyright (c) 2000  Embedded Planet <mlocke@jps.net>
 *
 *  based on the following
 *
 *  VR41xx Touch Panel Driver
 *  Copyright (c) 1999,2000 Michael Klar <wyldfier@iname.com>
 *
 *  This file is subject to the terms and conditions of the GNU General
 *  Public License.  See the file "COPYING" in the main directory of 
 *  this archive for more details
 */

// if anyone ever figures out whats wrong with SPI
//#define CPM_SPI

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>
#include <linux/tpanel.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/mpc8xx.h>
#include <asm/io.h>
#include <asm/delay.h>

//#include <asm/rpxlite.h>
//#include <asm/rpxhiox.h>

#include "asm/commproc.h"

/*
 *  The BB ADS7843 and logic in the HIOX card generate pendown interrupts
 *  every 1, 10, or 100 ms.  10 ms interrupts seem reasonable.  The penup
 *  event is generate by timer interrupt set to timeout at 15 ms after the 
 *  last pendown.  The first packet read after a interrupt is garbage, so 
 *  read two packets per interrupt and throw the first away.
 *  The data to/from the 7843 looks like this if set for 12 bits
 *		zeros	 cmd wrd    data
 *		xxxxxxxx xxxxxxxx x1234567 89ABCxxx
 *
 */

/* 
 * Touch panel commands.
 *  for BB ADS7843
 */

#define TP_CMD_START            ((u_char)0x80)
#define TP_CMD_NOPOWERDOWN      ((u_char)0x03)
#define TP_CMD_8BIT             ((u_char)0x08)
#define TP_CMD_XP               ((u_char)0x10)
#define TP_CMD_YP               ((u_char)0x50)

#define TP_MEASURE_X (TP_CMD_START|TP_CMD_8BIT|TP_CMD_XP|TP_CMD_NOPOWERDOWN)
#define TP_MEASURE_Y (TP_CMD_START|TP_CMD_8BIT|TP_CMD_YP|TP_CMD_NOPOWERDOWN)

//#define SAMPLES_PER_INTR 4
#define SAMPLES_PER_INTR 51 // more samples removes jitter

/*
 * Timer commands
 */
#define TIMER_TGCR_EN1  0x0001
#define TIMER_TGCR_STP1 0x0002
#define TIMER_TGCR_EN2  0x0010
#define TIMER_TGCR_STP2 0x0020


// BUFFSIZE can be increased, but it must be a power of 2

#define BUFFSIZE 256 

struct tpanel_status {
	unsigned int	buffer[BUFFSIZE];
	unsigned char	head;
	unsigned char	tail;
	unsigned char	datavalid;
	unsigned char	lastcontact;
	wait_queue_head_t wait;
	struct fasync_struct *fasyncptr;
	int		active;
};

/* Since I can't make the microcode patch work for the SPI, I just
 * clock the bits using software.
 */
#ifndef CPM_SPI
static	void	sw_spi_init(void);
static	void	sw_spi_io(u_char *obuf, u_char *ibuf, uint bcnt);
#endif

#ifdef CPM_SPI
extern void cpm_spi_init(void);
extern ssize_t cpm_spi_read(struct file *, char *, size_t, loff_t *);
extern ssize_t cpm_spi_write(struct file *, char *, size_t, loff_t *);
#endif

static struct tpanel_status tpanel;

void tpanel_add_data(uint data1, uint data2)
{
	tpanel.buffer[tpanel.head++] = data1 | 
				(tpanel.datavalid << 31)|
				(tpanel.lastcontact << 30);
	tpanel.buffer[tpanel.head++] = data2;
	tpanel.head &= (BUFFSIZE - 1);
	
	if (tpanel.head == tpanel.tail) {
		tpanel.tail += 2;
		tpanel.tail &= (BUFFSIZE - 1);
		tpanel.buffer[tpanel.tail] |= 0x20000000;
	}
}
static void tpanel_pendn_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int data1 = 0, data2 = 0;
	loff_t dummy;
	char cmd[4];
	int i;
	volatile uint *hcsr;
	volatile immap_t	*immap;
	volatile cpmtimer8xx_t 	*timers;

	immap = (immap_t *)IMAP_ADDR;
	timers = (cpmtimer8xx_t *)&(immap->im_cpmtimer);
	hcsr = (volatile uint *)HIOX_CSR0_ADDR; 

	// TIMER - stop and clear counter
	timers->cpmt_tgcr |= (TIMER_TGCR_STP2);
	timers->cpmt_tcn2 = 0;
		
/* 
 *  acorrding to HIOX FW manual it is possible to get
 *  a interrupt with time-out still 0, so 
 *  check for interrupt
 */
	if (*hcsr & HIOX_CSR0_TIRQSTAT) {

		// throw away the 1st sample
		for(i = 0; i < SAMPLES_PER_INTR; i++)
		{
			cmd[0] = 0;
			cmd[1] = TP_MEASURE_Y;
			cmd[2] = 0;
			cmd[3] = 0;
		#ifdef CPM_SPI
			cpm_spi_write(NULL,cmd,2,&dummy);
			cpm_spi_read(NULL,&cmd[2],2,&dummy);
		#else
			sw_spi_io(&cmd[0],&cmd[0],4);
		#endif
			if(i> 0)
			data2 += ((cmd[2] << 1) + (cmd[3] >> 7));
		}
		for(i = 0; i < SAMPLES_PER_INTR; i++)
		{
			cmd[0] = 0;
			cmd[1] = TP_MEASURE_X;
			cmd[2] = 0;
			cmd[3] = 0;
	
		#ifdef CPM_SPI
			cpm_spi_write(NULL,cmd,2,&dummy);
			cpm_spi_read(NULL,&cmd[2],2,&dummy);
		#else
			sw_spi_io(&cmd[0],&cmd[0],4);
		#endif
			if(i> 0)
			data1 += ((cmd[2] << 1) + (cmd[3] >> 7));
		}
		
		tpanel.lastcontact = 1;
		tpanel.datavalid = 1;
		tpanel_add_data(data1/(SAMPLES_PER_INTR -1),
				data2/(SAMPLES_PER_INTR -1));

		// turn interrupts back on
		cmd[0] = 0;
		cmd[1] = TP_CMD_START;
		sw_spi_io(&cmd[0],&cmd[0],2);


		*hcsr |= HIOX_CSR0_TIRQSTAT;
		*hcsr |= HIOX_CSR0_PDOWN10;
	}

	wake_up_interruptible(&tpanel.wait);

	if (tpanel.fasyncptr)
		kill_fasync(&tpanel.fasyncptr, SIGIO, POLL_IN);

	// TIMER - restart
	timers->cpmt_tgcr &= ~(TIMER_TGCR_STP2);
}

static void tpanel_penup_intr(void *dev_id)
{
	volatile immap_t	*immap;
	volatile cpmtimer8xx_t 	*timers;

	immap = (immap_t *)IMAP_ADDR;
	timers = (cpmtimer8xx_t *)&(immap->im_cpmtimer);


	// pendown timeout
	// TIMER - stop
	timers->cpmt_tgcr |= (TIMER_TGCR_STP2);
	timers->cpmt_ter2 = 0xffff;
	timers->cpmt_tcn2 = 0;

	if(tpanel.lastcontact)  {
		tpanel.lastcontact = 0;
		tpanel.datavalid = 0;
		tpanel_add_data(0,0);
	}

	wake_up_interruptible(&tpanel.wait);
	
	if (tpanel.fasyncptr)
		kill_fasync(&tpanel.fasyncptr, SIGIO, POLL_IN);


}


static int fasync_tpanel(int fd, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &tpanel.fasyncptr);
	if (retval < 0)
		return retval;
	return 0;
}

static int close_tpanel(struct inode * inode, struct file * file)
{
	volatile uint *hcsr;

	hcsr = (volatile uint *)HIOX_CSR0_ADDR;

	*hcsr |= HIOX_CSR0_TIRQSTAT;
	*hcsr &= ~HIOX_CSR0_ENTOUCH;

	fasync_tpanel(-1, file, 0);
	if (--tpanel.active)
		return 0;
		// set for standby
	MOD_DEC_USE_COUNT;
	return 0;
}

static int open_tpanel(struct inode * inode, struct file * file)
{
	volatile uint *hcsr;
	loff_t dummy;
	unsigned char init_cmd[2];

	hcsr = (volatile uint *)HIOX_CSR0_ADDR;

	if (tpanel.active++)
		return 0;
	tpanel.tail = tpanel.head;
	tpanel.lastcontact = 0;
	tpanel.datavalid = 0;

/*
 *  clear interrupt, turn ADC chip on
 */
	*hcsr |= HIOX_CSR0_TIRQSTAT;
	*hcsr |= HIOX_CSR0_ENTOUCH;

/*
 *  turn pendown interrupts on
 */
	init_cmd[0] = (TP_CMD_START); 
	init_cmd[1] = 0;

#ifdef CPM_SPI
	cpm_spi_write(NULL,init_cmd,2,&dummy);
#else
	sw_spi_io(init_cmd,init_cmd,2);
#endif

/*
 *   set pen down timer/enable hiox interrupts
 */
	*hcsr |= (HIOX_CSR0_PDOWN10);

	MOD_INC_USE_COUNT;
	return 0;
}

/****************************************************************************
 *  NOTE:	This data format was defined by the VR41xx Touch Panel driver
 *		for Microwindows.  In order to be transparent to Microwindows
 *		I maintained the same format and added dummy values where 
 *		necessary.
 *
 *		x-,y- = 0;	The sample is already differential
 *		z = 901;	This is the minimum pressure reading defined
 *				by Microwindows/VR41xx
 ****************************************************************************
 *
 *  Read touch panel data.  Blocking read is not implemented, mostly to stay
 *  consistant with the way the mouse drivers do polling.
 *
 *  Data format:
 *  unsigned short status: bit 15 = xyz data valid (0 means contact state only)
 *                         bit 14 = pen contact state (1 means contact)
 *                         bit 13 = soft data lost flag: if this is 1, this point
 *                                  is valid (if bit 15 is 1), but data was lost
 *                                  between this point and the previous point
 *                                  due to a buffer overrun in the driver
 *                         bit 12 = hard data lost flag: if this is 1, this point
 *                                  is valid (if bit 15 is 1), but data was lost
 *                                  between this point and the previous point
 *                         bits 11-0 = reserved
 *  unsigned short x+ raw data (if status:15 = 1)
 *  unsigned short x- raw data (if status:15 = 1)
 *  unsigned short y+ raw data (if status:15 = 1)
 *  unsigned short y- raw data (if status:15 = 1)
 *  unsigned short z (pressure) raw data (if status:15 = 1)
 *
 *  x+, x-, y+, and y- are limited to range 0-1023 for this hardware.  Each +/-
 *  pair is somewhat redundant: the + value can be used as is, but using the
 *  difference (eg. (x+) - (x-)) will produce a more accurate result.  Further
 *  manipulation and/or statistical analysis may be required for best accuracy.
 *
 *  No calibration is done on the driver side, that is expected to be done on
 *  the user side.
 */

static ssize_t read_tpanel(struct file * file,
	char * buffer, size_t count, loff_t *ppos)
{
	int r = 0;
	unsigned int data1 = 0, data2 = 0;
	ulong flags;

	if (count < 12)
		return -EINVAL;

// We need to access the circular buffer atomic to
// anything else that will read or write it
	save_flags(flags); cli();

	while (tpanel.head == tpanel.tail) {
		if(file->f_flags & O_NONBLOCK) {
			restore_flags(flags);
			return -EAGAIN;
		}

		interruptible_sleep_on(&tpanel.wait);

		if (signal_pending(current))
			return -ERESTARTSYS;
		
		save_flags(flags); cli();
	}

	do {
		data1 = tpanel.buffer[tpanel.tail++];
		data2 = tpanel.buffer[tpanel.tail++];
		tpanel.tail &= BUFFSIZE - 1;

		restore_flags(flags);
	
		if (!access_ok(VERIFY_WRITE, buffer, 12))
			return -EFAULT;

		__put_user(((data1 >> 16) & 0xf000), (short*)buffer +r);
		__put_user((data1 & 0x0000ffff), (short*)(buffer +r + 2));
		__put_user(0x0000,(short *)(buffer +r +4));	
		__put_user((data2 & 0x0000ffff), (short*)(buffer +r + 6));
		__put_user(0x0000,(short *)(buffer +r +8));	
		__put_user(901,(short *)(buffer +r +10));	
		r += 12;
		if(r > (count -12))
		  return r;

		save_flags(flags); cli();
		
	} while (tpanel.head != tpanel.tail);
	
	
	restore_flags(flags);
	
	return r;
}

static unsigned int poll_tpanel(struct file *file, poll_table * wait)
{
  int flags,comp;

	poll_wait(file, &tpanel.wait, wait);

	save_flags(flags); cli();
	comp = (tpanel.head != tpanel.tail);
	restore_flags(flags);

	if (comp)
		return POLLIN | POLLRDNORM;
	return 0;
}


/*
 *  This is currently ignored but it does apply to us
 *  We have several parameters that the user should
 *  be allowed to set, # of bits, interrupt period
 */

static int ioctl_tpanel(struct inode * inode, struct file * file,
	unsigned int cmd, unsigned long arg)
{
	struct scanparam *sparm;
	unsigned int i1, i2;

	sparm = (struct scanparam *)arg;

	switch (cmd) {
	case TPGETSCANPARM:
		if (put_user(i1, &sparm->interval))
			return -EFAULT;
		if (put_user(i2, &sparm->settletime))
			return -EFAULT;
		return 0;
	case TPSETSCANPARM:
		if (get_user(i1, &sparm->interval))
			return -EFAULT;
		if (get_user(i2, &sparm->settletime))
			return -EINVAL;
		__put_user(i1 * 30, &sparm->interval);
		__put_user(i2 * 30, &sparm->settletime);
		return 0;
	default:
		return -EINVAL;
	}
}

static struct file_operations tpanel_fops = {
#ifdef MODULE
  owner: THIS_MODULE,
#endif
	read: read_tpanel,
	poll: poll_tpanel,
	ioctl: ioctl_tpanel,
	open: open_tpanel,
	release: close_tpanel,
	fasync: fasync_tpanel
};

static struct miscdevice tpanel_device = {
	TPANEL_MINOR, "tpanel", &tpanel_fops
};

int __init tpanel_init(void)
{
	volatile uint *hcsr4;
	volatile immap_t	*immap;
	volatile cpmtimer8xx_t 	*timers;

	hcsr4 = (volatile uint *)HIOX_CSR4_ADDR;
	immap = (immap_t *)IMAP_ADDR;
	timers = (cpmtimer8xx_t *)&(immap->im_cpmtimer);

/*
 *  Naming is a little confusing here
 *  HIOX IRQ2 is SIU IRQ4
 *  HIOX IRQ3 is SIU IRQ1
 */
	*hcsr4 &= ~(HIOX_CSR4_ENTIRQ2 | HIOX_CSR4_ENTIRQ3);
	*hcsr4 |= HIOX_CSR4_ENTIRQ2;

/*
 *  init TIMER2 for 15ms timeout
 */
	timers->cpmt_tgcr = 0x0000;
	timers->cpmt_tmr2 = 0xff14;  // divide by 16, ps by 256
	timers->cpmt_tcn2 = 0;
	timers->cpmt_trr2 = 241;
	timers->cpmt_ter2 = 0xffff;
	timers->cpmt_tgcr |= TIMER_TGCR_EN2;
	timers->cpmt_tgcr |= TIMER_TGCR_STP2;
	timers->cpmt_tcn2 = 0;

#ifdef CPM_SPI
        cpm_spi_init();
#else
	sw_spi_init();
#endif

	tpanel.active = 0;
	tpanel.head = tpanel.tail = 0;
	tpanel.fasyncptr = NULL;
//	tpanel.wait = NULL;
	init_waitqueue_head (&tpanel.wait);

/*  
 *  Make IRQ4 edge sensitive.
 */
        ((immap_t *)IMAP_ADDR)->im_siu_conf.sc_siel |= 0x00800000;

	if(request_8xxirq(SIU_IRQ4, tpanel_pendn_intr,0,
						"tpanel",NULL) != 0)
	  printk("Could not allocate tpanel IRQ\n");

	cpm_install_handler(CPMVEC_TIMER2, tpanel_penup_intr,NULL);

	if(misc_register(&tpanel_device))
		printk("unable to register TPANEL\n");

	printk("TPANEL v.80 driver installed\n");
	return 0;
}

#ifdef MODULE

int init_module(void)
{
	return tpanel_init();
}

void cleanup_module(void)
{
	free_irq(SIU_IRQ4, NULL);
	cpm_free_handler(CPMVEC_TIMER2);
	misc_deregister(&tpanel_device);
}
#else
__initcall(tpanel_init);
#endif


#ifndef CPM_SPI
/* Software SPI functions.
 * These are on Port B.
 */
#define PB_SPICLK	((uint)0x00000002)
#define PB_SPIMOSI	((uint)0x00000004)
#define PB_SPIMISO	((uint)0x00000008)

static
void	sw_spi_init(void)
{
	volatile cpm8xx_t	*cp;
	volatile uint		*hcsr0;

	hcsr0 = (volatile uint *)HIOX_CSR0_ADDR;
	cp = cpmp;	/* Get pointer to Communication Processor */

	/* Disable touch panel SPI select.
	*/
	*hcsr0 &= ~HIOX_CSR0_TSELSPI;

	/* Make these Port B signals general purpose I/O.
	 * First, make sure the clock is low.
	 */
	cp->cp_pbdat &= ~PB_SPICLK;
	cp->cp_pbpar &= ~(PB_SPICLK | PB_SPIMOSI | PB_SPIMISO);

	/* Clock and Master Output are outputs.
	*/
	cp->cp_pbdir |= (PB_SPICLK | PB_SPIMOSI);

	/* Master Input.
	*/
	cp->cp_pbdir &= ~PB_SPIMISO;

}

static
void	sw_spi_io(u_char *obuf, u_char *ibuf, uint bcnt)
{
	int	bits, i;
	u_char	outbyte, inbyte;
	volatile cpm8xx_t	*cp;
	volatile uint		*hcsr0;

	hcsr0 = (volatile uint *)HIOX_CSR0_ADDR;
	cp = cpmp;	/* Get pointer to Communication Processor */

	/* The timing on the bus is pretty slow.  Code inefficiency
	 * and eieio() is our friend here :-).
	 */
	cp->cp_pbdat &= ~PB_SPICLK;
	*hcsr0 |= HIOX_CSR0_TSELSPI;	/* Enable SPI select */
	eieio();

	/* Clock in/out the bytes.  Data is valid on the rising edge
	 * of the clock.  Data is MSB first.
	 */
	for (i=0; i<bcnt; i++) {
		outbyte = *obuf++;
		inbyte = 0;
		for (bits=0; bits<8; bits++) {
			if (outbyte & 0x80)
				cp->cp_pbdat |= PB_SPIMOSI;
			else
				cp->cp_pbdat &= ~PB_SPIMOSI;
			eieio();
			cp->cp_pbdat |= PB_SPICLK;
			eieio();
			outbyte <<= 1;
			inbyte <<= 1;
			if (cp->cp_pbdat & PB_SPIMISO)
				inbyte |= 1;
			cp->cp_pbdat &= ~PB_SPICLK;
			eieio();
		}
		*ibuf++ = inbyte;
	}

	*hcsr0 &= ~HIOX_CSR0_TSELSPI;	/* Disable SPI select */
	eieio();
}
#endif
