/*
 * 
 * rd.untar.c
 * 
 * 14-APR-2000
 * Partly re-written by Robert Kaiser (rkaiser@sysgo.de) to support
 * uncompressing/extracting the tar image in one go
 *
 * Dimitri Maziuk
 * and
 * Dave Cinege
 * 
 * Parts stolen from GNU 'Let's make the same archive 10 different ways' tar
 * 
 * 1998-01-10
 * This version of untar is very limited. It expects a fairly good archive, 
 * as well as expansion to a clean root.
 * 
 * It does not:
 * 
 * Look for a leading '/'
 * Check before making links
 * Look at the header checksum
 * Check or create any leading directory structure
 * 
 * 1998-01-17
 * Is GNUtar the biggest fuck-up in released software since MS-DOS? 
 * It could be.
 * 
 * Tar decides it sometimes likes to make a dir entry that is blank or "./".
 * We now check for this, and modified our exit pattern to deal with it.
 * 
 * Although both Dima and I thought it wasn't possible, tar can  
 * stor files without first storing the dir they require.
 * If your nice freshly made archive doesn't work don't be surprised.
 * Try specing a different way to make it.
 *
 * Dima has a forthcoming mini tar archive creator (ctar) that should
 * better deal with these problems. You may want to look for it if normal
 * tar does not suite your needs when using this kernel feature.
 * 
 * 1998-02-22
 * Convert to 2.1 file ops. Open /dev/ram1 locally. With 2.1 outfile in rd.c
 * was not holding it's value when we entered this code. I think it has
 * to do with the newer __initfunc(); thingy. Couldn't figure out how to
 * make it work. Did this instead.
 *
 * 1998-10-17
 * Move to wrapper function.
 *
 */

#include <linux/unistd.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/stddef.h>  
#include <linux/utime.h>

//DEBUG_UNTAR only
// #include <linux/delay.h>


static unsigned long simple_strtoul_wrapper(const char *cp,char **endp,unsigned int base);

//#define DEBUG_UNTAR

#define E_NAMETOOLONG	1
#define E_CANNOTSTAT	2
#define E_READLINK	3
#define E_SHORTWRITE	4
#define E_FOPEN		5
#define E_MALLOC	6
#define E_OPENDIR	7
#define E_WRONG_FTYPE	8

#define	TARBLOCKSIZE	512

#define NAME_FIELD_SIZE   100
#define PREFIX_FIELD_SIZE 155
#define UNAME_FIELD_SIZE   32
#define GNAME_FIELD_SIZE   32

#define TMAGIC   "ustar "	/* ustar and a null */
#define TMAGLEN  6
#define TVERSION " \0"		/* 00 and no null */
#define TVERSLEN 2

#define REGTYPE	 '0'		/* regular file */
#define AREGTYPE '\0'		/* regular file */
#define LNKTYPE  '1'		/* link */
#define SYMTYPE  '2'		/* reserved */
#define CHRTYPE  '3'		/* character special */
#define BLKTYPE  '4'		/* block special */
#define DIRTYPE  '5'		/* directory */
#define FIFOTYPE '6'		/* FIFO special */
#define CONTTYPE '7'		/* reserved */

#define BUF_SIZE (1024*64)

//#define strtoul simple_strtoul
#define strtoul simple_strtoul_wrapper

/* POSIX header  */
struct fileHeader
{					/* byte offset */
  char name[NAME_FIELD_SIZE];		/*   0 */
  char mode[8];				/* 100 */
  char uid[8];				/* 108 */
  char gid[8];				/* 116 */
  char size[12];			/* 124 */
  char mtime[12];			/* 136 */
  char chksum[8];			/* 148 */
  char typeflag;			/* 156 */
  char linkname[NAME_FIELD_SIZE];	/* 157 */
  char magic[TMAGLEN];			/* 257 */
  char version[TVERSLEN];		/* 263 */
  char uname[UNAME_FIELD_SIZE];		/* 265 */
  char gname[GNAME_FIELD_SIZE];		/* 297 */
  char devmajor[8];			/* 329 */
  char devminor[8];			/* 337 */
  char prefix[PREFIX_FIELD_SIZE];	/* 345 */
					/* 500 */
};

typedef union TarInfo 
{
  char buf[TARBLOCKSIZE];
  struct fileHeader header;
} TarInfo;

enum untar_status
{
	READING_HEADER,
	READING_DATA,
	SKIPPING_REST
};

struct untar_context
{
	TarInfo           tarInfo;
	enum untar_status state;
	mode_t            fmode;
	unsigned long     fsize;
	int               out;
	int               rotate;
} *Untar;


extern int sys_write(unsigned int fd, const char * buf,unsigned int count);
extern int sys_chmod(const char * filename, mode_t mode);
extern int sys_chown(const char * filename, uid_t user, gid_t group);
extern int sys_mknod(const char * filename, int mode, dev_t dev);
extern int sys_mkdir(const char * pathname, int mode);
extern int sys_chdir(const char * filename);
extern int sys_link(const char * oldname, const char * newname);
extern int sys_symlink(const char * oldname, const char * newname);
extern int sys_utime(char * filename, struct utimbuf * times);
extern int sys_time(int * tloc);
extern int sys_fchown(unsigned int fd, uid_t user, gid_t group);

/*
** This is called by flush_window to flush the gzip output
** buffer (containing part of a tar image).
*/
__initfunc(static ssize_t untar_write(
	struct file *fp,		/* not used */
	const char  *buf,
	size_t      count,
	loff_t      *fpos))		/* not used */
{
	int i, scoop, err, file_written = 0;
	char *to;
	kdev_t fdev;
	struct utimbuf ut;
	struct untar_context *Untar;
#ifndef DEBUG_UNTAR
#ifndef CONFIG_ARCH_S390
	char rotator[4] = { '|' , '/' , '-' , '\\' };

	/* pointer to our context structure is passed thru ioctl fop */
	Untar = (struct untar_context*)fp->f_op->ioctl;

	printk("%c\b", rotator[Untar->rotate & 0x3]);
	Untar->rotate++;
#endif
#endif

	/*
	** Note: the gunzip rotuines call this only when
	**       a) the gunzip buffer is full -> count = WSIZE = 64 * TARBLOCKSIZE
	** or    b) to flush the last bytes out
	*/
	while(count)
	{	/*
		** What is our current state ?
		*/
		err = 0;
		switch(Untar->state)
		{
			case READING_HEADER:
			{	/* get the tar header */
				if(count < TARBLOCKSIZE)
				{
					count = 0;
					break;
				}
				to = (char*)&Untar->tarInfo;
				for(i = 0; i < TARBLOCKSIZE; i++)
				{
					*to++ = *buf++;
					--count;
				}
				/*
				** Tar usually pads the output byte to a
				** multiple of it's block size usually
				** 10 KB, appending zeroes if necessary.
				** here we skip those zeroes:
				*/
				if(Untar->tarInfo.header.name[0] == '\0'  &&
				   Untar->tarInfo.header.mode[0] == '\0' &&
				   Untar->tarInfo.header.mode[0] == '\0')
				{
					if(count < TARBLOCKSIZE)
						count = 0;
					break;
				}
				/* determine mode and size */
				Untar->fmode = strtoul(Untar->tarInfo.header.mode,NULL,8);
				Untar->fsize = strtoul(Untar->tarInfo.header.size,NULL,8);
#ifdef DEBUG_UNTAR
				printk("\ntype: %c ",Untar->tarInfo.header.typeflag);   
				printk("name: %s ",Untar->tarInfo.header.name);
				printk("Untar->fsize: %ld ",fsize); 
				udelay(100000);
#endif
				/* check the type */
				switch(Untar->tarInfo.header.typeflag)
				{
					case AREGTYPE : 
					case REGTYPE  :
						Untar->out = sys_open(Untar->tarInfo.header.name,
					             O_CREAT|O_WRONLY|O_TRUNC, Untar->fmode);
						if(Untar->out < 0)
						{
							err = Untar->out;
							break;
						}
						
						err = sys_fchown(Untar->out,
							(uid_t)strtoul(Untar->tarInfo.header.uid, NULL, 8),
							(uid_t)strtoul(Untar->tarInfo.header.gid, NULL, 8));
						
						Untar->state = READING_DATA;
						break;
					case LNKTYPE  :
						err = sys_link(Untar->tarInfo.header.linkname,
						                  Untar->tarInfo.header.name);
						break;

					case SYMTYPE  :
						err = sys_symlink(Untar->tarInfo.header.linkname,
						                 Untar->tarInfo.header.name);
						break;

					case CHRTYPE  :	
					case BLKTYPE  :	
					case FIFOTYPE :	
						fdev = MKDEV(strtoul(Untar->tarInfo.header.devmajor,NULL,8),
						   strtoul(Untar->tarInfo.header.devminor,NULL,8));
						err = sys_mknod(Untar->tarInfo.header.name,Untar->fmode,fdev);
   	    				break;

					case DIRTYPE  :
						/* Skip if name is "" "./" "." ".." */
						if((Untar->tarInfo.header.name[0] != '.' && 
						    Untar->tarInfo.header.name[0] != '\0') &&
						   (Untar->tarInfo.header.name[1] != '/' &&
						    Untar->tarInfo.header.name[1] != '\0') &&
						   (Untar->tarInfo.header.name[1] != '\0')) {
							err = sys_mkdir(Untar->tarInfo.header.name,Untar->fmode);
							/* don't treat existing directories as errors */
							if (err == -EEXIST)
							  err = 0;
						}
						break;
					default:
						printk(KERN_ERR "RAMDISK: corrupt tar archive\n");
						Untar->state = SKIPPING_REST;
						return(-1);
				}
				break;
			}
			case READING_DATA:
			{
				scoop = 0;
				while(count > 0 && Untar->fsize > 0)
				{
					int wrote;
					scoop = Untar->fsize > TARBLOCKSIZE ? TARBLOCKSIZE : Untar->fsize;
					wrote = sys_write(Untar->out, buf, scoop);
					if (wrote < 0)
					{
						/* fatal error, skip this file */
						Untar->fsize = 0;
						err = wrote;
						break;
					}
					count -= wrote;
					buf   += wrote;
					Untar->fsize -= wrote;
				}
				if(Untar->fsize == 0)
				{	/* skip unto the next tar block */
					sys_close(Untar->out);
					scoop = count % TARBLOCKSIZE;
					buf   += scoop;
					count -= scoop;
					Untar->state = READING_HEADER;
					file_written = 1;
				}
				break;
			}
			case SKIPPING_REST:
				count = 0;
				break;
		}
		if(err)
		{
			/* 
			 * Tell the user what went wrong. If the RAM disk was full, we
			 * skip all remaining data. The switch() should be maintained
			 * to contain the most common types of errors in order to help
			 * the user to track down configuration problems.
			 */
			printk("\nError %d making %s", -err, Untar->tarInfo.header.name);
			switch(err)
			{
				case -ENOSPC:
					printk(": No space left on device\n");
					Untar->state = SKIPPING_REST;
			   	break;
				case -EPERM:
					printk(": Operation not permitted\n");
			   	break;
				case -EACCES:
					printk(": Permission denied\n");
			   	break;
				default:
				break;
			}
		}
		if(file_written)
		{
			if(Untar->tarInfo.header.typeflag != LNKTYPE &&
			   Untar->tarInfo.header.typeflag != SYMTYPE)
			{
				err = sys_chmod(Untar->tarInfo.header.name, Untar->fmode);
			}
			if(err)
				printk("\nError chown and/or chmod %s", Untar->tarInfo.header.name);
			ut.actime=sys_time(NULL);
			ut.modtime=strtoul(Untar->tarInfo.header.mtime,NULL,8);
			sys_utime(Untar->tarInfo.header.name,&ut);
			file_written = 0;
		}
	}
	return(0);
}

typedef int (*ioctl_fop_t)(struct inode *, struct file *, unsigned int, unsigned  long);


__initfunc(static int untar(void))
{
 	struct               inode inode;
	struct               file infile, outfile;
	struct               file_operations out_fops;
	struct               dentry in_dentry, out_dentry;
	kdev_t               device;
	int                  err;
	struct untar_context *Untar;


	/*
	** get buffers for tar extractor
	*/
	if(!(Untar = kmalloc(sizeof(*Untar), GFP_KERNEL)))
	{
		printk(KERN_ERR "RAMDISK: Couldn't allocate untar buffer\n");
		return(-1);
	}
	Untar->state  = READING_HEADER;	/* begin reading header */
	Untar->rotate = 0;

	/*
	** load from initrd:
	** Note: initrd has already been opened (and left open)
	**       by rd_load_image(). Therefore, we don't open
	**       it again here, we just copy the initrd_fops
	**       pointer.
	*/
	device = MKDEV(MAJOR_NR, INITRD_MINOR);
	memset(&infile, 0, sizeof(infile));
	memset(&inode, 0, sizeof(inode));
	memset(&in_dentry, 0, sizeof(in_dentry));
	inode.i_rdev      = device;
	infile.f_mode     = 1; /* read only */
	infile.f_dentry   = &in_dentry;
	infile.f_op       = &initrd_fops;
	in_dentry.d_inode = &inode;

	/*
	** build a fake outfile descriptor for use with gunzip()
	** all gunzip will ever do with it is to call the
	** write function. We misuse the fops.iocl entry
	** here to pass a pointer to Untar down to untar_write()
	*/
	memset(&outfile, 0, sizeof(outfile));
	memset(&out_inode, 0, sizeof(out_inode));
	memset(&out_dentry, 0, sizeof(out_dentry));
	memset(&out_fops, 0, sizeof(out_fops));
	out_inode.i_rdev   = 0;		/* don't care */
	outfile.f_mode     = 3; 	/* don't care */
	out_fops.write     = untar_write;
	out_fops.ioctl     = (ioctl_fop_t)Untar;
	outfile.f_dentry   = &out_dentry;
	outfile.f_op       = &out_fops;
	out_dentry.d_inode = &out_inode;

	printk(KERN_NOTICE "RAMDISK: Extracting root archive: ");

	/* now uncompress & untar the initrd image */
	err = crd_load(&infile, &outfile);

	/* free buffer */
	kfree(Untar);

	/* now free the initrd image */
	invalidate_buffers(device);

	/*
	** now close the initrd device, freeing it's space
	*/
#ifdef DEBUG_UNTAR
	printk("\ninitrd_users = %d ",initrd_users);   
#endif
	if(infile.f_op->release)
		infile.f_op->release(&inode, &infile);
#ifdef DEBUG_UNTAR
	printk("-> %d\n",initrd_users);   
#endif

	if(!err)
		printk("done.\n");

	return(err);
}

/*
 * simple_strtoul() does not strip leading spaces (0x20) from the input
 * string. The 'real' strtoul does. This wrapper works around this.
 * NOTE: This probably should be fixed in: lib/vsprintf.c 
 *
 */

__initfunc(static unsigned long simple_strtoul_wrapper(
	const char   *cp,
	char         **endp,
	unsigned int base))
{
	while (*cp == ' ') {
		cp++;
	}

	return ( simple_strtoul(cp,endp,base) );
}
