/*
 * Men824x Watchdog: A Simple Hardware Watchdog Device
 * Based on MixCom Watchdog driver by Gergely Madarasz
 * for Men 824x boards.
 *
 * Copyright (c) 2001 MEN MIKRO ELEKTRONIK <info@men.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This driver is for the ALI824x MEN series boards.
 *
 * On the F1N, the watchdog is started when /dev/watchdog is
 * open.
 * On the A12/D3/SC13, the watchdog have to be started by the
 * MenMon (get help with the command 'h wdog' on MenMon prompt.
 * As soon as the MenMon start the Linux Kernel it must toggle
 * the watchdog trigger line. To manage this during startup
 * this files have been modified:
 * - include/linux/init.h
 * - init/main.c
 * - kernel/printk.c
 * - arch/ppc/boot/common/misc-simple.c
 * - arch/ppc/boot/men824xali/Makefile
 * And this files added:
 * - arch/ppc/boot/men824xali/men824xali_wdog.c
 * - arch/ppc/boot/men824xali/men824xali_wdog_timer.c
 *
 * If CONFIG_WATCHDOG_NOWAYOUT is defined, then once the file
 * /dev/watchdog have been open, it should never be closed.
 * If for any reason it is closed, the the watchdog will
 * expire.
 * If CONFIG_WATCHDOG_NOWAYOUT is not defined the on close of
 * /dev/watchdog
 * -for the F1N: the watchdog is stopped.
 * -for the A12/D3/SC13: the watchdog can't be stopped, then
 * the timer is started again.
 *
 * Version 0.1 (10/22/01):
 *	- first version
 * Version 0.2 (03/24/02):
 *	- pba@sysgo.de: added support for reading watch dog timeout from SMS24

 *
 */

#define VERSION "0.2" 

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/ioport.h>
#include <linux/watchdog.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/pci.h>
#include <linux/i2c.h>

#define DRIVER_NAME	"men824xwd"

#if defined(CONFIG_MENB11) || defined(CONFIG_MENA12)
#define SMS24_I2C_ADDR      	0x48
static unsigned int G_timeout;
extern int men824xwd_enable_ping_early;
#elif defined(CONFIG_MENF1N)
# define ADDR_STROBE1 0x70
# define ADDR_STROBE2 0x72
# define DATA_PORT    0x71
# define WATCHDOG     0x1ff7
# define RTC_ADR_L    0x007f & WATCHDOG
# define RTC_ADR_H    (0xff80 & WATCHDOG ) >> 7

static char G_watchReg;
#endif

static long men824xwd_opened;
static struct timer_list men824xwd_timer;
struct timer_list men824xwd_timer_1;
static struct pci_dev *m1543_pci_dev = NULL;

/******************* I2C client for reading watch dog timeout *******************************/
static int i2c_sms24_attach_adapter(struct i2c_adapter *adap);
/*
** dummy for i2c command
*/
static int i2c_sms24_command(struct i2c_client *client, unsigned int cmd,
                           void *arg)
{
  return -1;
}

/*
** dummy for i2c detach client
*/
static int i2c_sms24_detach_client(struct i2c_client *client)
{
  return 0;
} 

/*
** i2c driver structure
*/
static struct i2c_driver i2c_sms24_driver = {
  name:			DRIVER_NAME,
  id:			I2C_HW_B_MEN824XALI,
  flags:		I2C_DF_DUMMY,
  attach_adapter:	i2c_sms24_attach_adapter,
  detach_client:	i2c_sms24_detach_client,
  command:		i2c_sms24_command,
};

/*
** i2c client structure
*/
static struct i2c_client client = {
  name:     "WD read tout client",
  id:       1,
  flags:    0,
  addr:    -1,
  adapter:  NULL,
  driver:   &i2c_sms24_driver,
  data:     NULL 
};

/*
** register the adapter we use
*/
static int i2c_sms24_attach_adapter(struct i2c_adapter *adap)
{
  if (client.adapter == adap) {
    /* This is actually a detach_adapter call! */
    client.adapter = NULL;
  } else {
    client.adapter = adap;
  }
  return 0;
}
/****************************************************************************/

static void men824xwd_ping(void)
{
#if defined(CONFIG_MENB11) 
	unsigned char tmpValue = 0;

	pci_read_config_byte(m1543_pci_dev, 0x7e, &tmpValue );
	tmpValue = (tmpValue & 0xfb) | ((tmpValue & 0x04) ^ 0x04);
	pci_write_config_byte(m1543_pci_dev, 0x7e, tmpValue );
#elif defined(CONFIG_MENA12)	
	unsigned char tmpValue = 0;

	pci_read_config_byte(m1543_pci_dev, 0xc3, &tmpValue );
	tmpValue = (tmpValue & 0xfd) | ((tmpValue & 0x02) ^ 0x02);
	pci_write_config_byte(m1543_pci_dev, 0xc3, tmpValue );
#elif defined(CONFIG_MENF1N)
	outb_p(RTC_ADR_L, ADDR_STROBE1);
	outb_p(RTC_ADR_H, ADDR_STROBE2);
	outb_p(G_watchReg, DATA_PORT);
#else
# error No MEN board defined
#endif
	return;
}

static void men824xwd_timerfun(unsigned long d)
{
	men824xwd_ping();
	
	mod_timer(&men824xwd_timer,jiffies + G_timeout);
}

static int men824xwd_open(struct inode *inode, struct file *file)
{
	if(test_and_set_bit(0,&men824xwd_opened)) {
		return -EBUSY;
	}
#if defined(CONFIG_MENB11) || defined(CONFIG_MENA12)
	del_timer(&men824xwd_timer);
#endif

	men824xwd_ping();

	return 0;
}

static int men824xwd_release(struct inode *inode, struct file *file)
{
	men824xwd_ping();
	lock_kernel();
#if !defined(CONFIG_WATCHDOG_NOWAYOUT)
# if defined(CONFIG_MENB11) || defined(CONFIG_MENA12)
	add_timer(&men824xwd_timer);
# elif defined(CONFIG_MENF1N)
	outb_p(RTC_ADR_L, ADDR_STROBE1);
	outb_p(RTC_ADR_H, ADDR_STROBE2);
	outb_p(G_watchReg0x80, DATA_PORT);
# endif
#endif
	clear_bit(0,&men824xwd_opened);
	unlock_kernel();
	return 0;
}


static ssize_t men824xwd_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	if (ppos != &file->f_pos) {
		return -ESPIPE;
	}

	if(len)
	{
		men824xwd_ping();
		return 1;
	}
	return 0;
}

static int men824xwd_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	int status;
	static struct watchdog_info ident = {
		WDIOF_KEEPALIVEPING, 1, "MEN 824x watchdog"
	};
                                        
	switch(cmd)
	{
		case WDIOC_GETSTATUS:
			status=men824xwd_opened;
			if (copy_to_user((int *)arg, &status, sizeof(int))) {
				return -EFAULT;
			}
			break;
		case WDIOC_GETSUPPORT:
			if (copy_to_user((struct watchdog_info *)arg, &ident, 
				sizeof(ident))) {
					return -EFAULT;
				}
			break;
		case WDIOC_KEEPALIVE:
			men824xwd_ping();
			break;
		default:
			return -ENOIOCTLCMD;
	}
	return 0;
}

static struct file_operations men824xwd_fops=
{
	owner:		THIS_MODULE,
	write:		men824xwd_write,
	ioctl:		men824xwd_ioctl,
	open:		men824xwd_open,
	release:	men824xwd_release,
};

static struct miscdevice men824xwd_miscdev=
{
	WATCHDOG_MINOR,
	"watchdog",
	&men824xwd_fops
};

static int __init men824xwd_init(void)
{
	int ret;
	u8 buf[1];
	int sms24_timeouts[8] = {0,0,0,400,800,1600,3200,6400};
	
	printk("MEN 824x watchdog driver v%s\n", VERSION);	
	
	m1543_pci_dev = pci_find_device(PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_M7101, m1543_pci_dev);
	if (m1543_pci_dev == NULL) {
		printk(DRIVER_NAME ": No device detected.\n");
		return -ENODEV;
	}

#if defined(CONFIG_MENB11) || defined(CONFIG_MENA12)
/*
# if defined(MEN_TOUT_800)
	G_timeout = 800;
# elif defined(MEN_TOUT_1600)
	G_timeout = 1600;
# elif defined(MEN_TOUT_3200)
	G_timeout = 3200;
# elif defined(MEN_TOUT_6400)
	G_timeout = 6400;
# endif
*/
	
	if (i2c_add_driver(&i2c_sms24_driver)) {
	    printk(KERN_INFO DRIVER_NAME ": I2C X-fer add_driver error.\n");
	    return -EFAULT;
	}
	if(!client.adapter) {
	    printk(KERN_INFO DRIVER_NAME ": I2C X-fer send without adapter.\n");
	    return -EFAULT;	    
	}
	if (client.adapter->inc_use)
	    client.adapter->inc_use(client.adapter);

	client.addr = SMS24_I2C_ADDR;
	buf[0] = 1; /* Program register 1 */
	i2c_master_send(&client, buf, 1);
	i2c_master_recv(&client, buf, 1);	

	if (client.adapter->dec_use)
	    client.adapter->dec_use(client.adapter);
	
	i2c_del_driver(&i2c_sms24_driver);	
	G_timeout = sms24_timeouts[buf[0] & 0x07];
	if (G_timeout == 0) {
	    printk(KERN_INFO DRIVER_NAME ": Watch dog disabled by MenMon\n");	
	    G_timeout = 6400;
	} else {
	    printk(KERN_INFO DRIVER_NAME ": Watch dog timeout: %dms\n", G_timeout);
	}    
	
	G_timeout = (((G_timeout * 8)/10) * HZ) / 1000;	
#elif defined(CONFIG_MENF1N)

	{
	int resolution, multiplier;
	const unsigned long msTbl[] = { 62, 250, 1000, 4000 };

	for( resolution=0; resolution < 4; resolution++ ){

		multiplier = MEN_TIMEOUT_VALUE / msTbl[resolution];
			if( multiplier < 32 )
				break;
		}
		if( resolution >= 4 )
			return -EINVAL;		/* timeout too large */

		if( multiplier == 0 )
			multiplier = 1;

		G_watchReg = 0x80 | (multiplier<<2) | resolution;
	}
#else
# error No MEN board defined
#endif

	if (ret = misc_register(&men824xwd_miscdev)) {
	    printk(DRIVER_NAME ": register misc device error.\n");
	    return ret;
	}    
	
#if defined(CONFIG_MENB11) || defined(CONFIG_MENA12)
	men824xwd_enable_ping_early = 0;
	men824xwd_ping();

	init_timer(&men824xwd_timer);
	men824xwd_timer.expires=jiffies + G_timeout;
	men824xwd_timer.function=men824xwd_timerfun;
	men824xwd_timer.data=0;
	add_timer(&men824xwd_timer);
	clear_bit(0,&men824xwd_opened);
#endif

	return 0;
}	

static void __exit men824xwd_exit(void)
{
#ifndef CONFIG_WATCHDOG_NOWAYOUT
	printk(KERN_WARNING DRIVER_NAME ": I quit now, hardware will"
		" probably reboot!\n");
	del_timer(&men824xwd_timer);
#endif
	misc_deregister(&men824xwd_miscdev);
}

module_init(men824xwd_init);
module_exit(men824xwd_exit);
