/***************************************************************************
                                  isp1181.c
                              -------------------
    project:  USB device for Philips ISP1181A

    begin                : Wed Sep 4 2002
    copyright            : (C) 2002 by HBM GmbH, Siegfried Wessler
    email                : siegfried.wessler@hbm.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include  <asm/io.h>

#include  "types.h"
#include  "global.h"
#include  "isp1181.h"
#include  "usbdev.h"
#include  "descript.h"
#include  "fifo.h"   // rep_outs



UINT8 usb_rd1b(void)

{ // Read only one data byte and return it

  return  inportb(ISP_DATA);
}


UINT16 usb_rd1w(void)
{ // Read only one data word and return it

  return (inportw(ISP_DATA));
}


void  usb_wr1b(UINT8 data)
{ // Simply write 1 data byte
  outportb(ISP_DATA, data);
}


void  usb_wr1w(UINT16 data)
{ // Simply write 1 data word
  outportw(ISP_DATA, data);
}


UINT8 usb_wrc_rd1b(UINT8 addr)
{ // write command with address 'addr' and return 1 data byte
  outportb(ISP_COMMAND, addr);

  return (inportb(ISP_DATA));
}


UINT16 usb_wrc_rd1w(UINT8 addr)
{ // write command with address 'addr' and return 1 data word
  outportb(ISP_COMMAND, addr);

  return (inportw(ISP_DATA));
}


UINT16 usb_wrc_rd2b(UINT8 addr)
{ // write command with address 'addr' and receive 2 data bytes
  UINT8   data1;
  UINT16  data2;

  outportb(ISP_COMMAND, addr);
  data1 = inportb(ISP_DATA);
  data2 = inportb(ISP_DATA);

  data2 = (data2 << 8) | data1;

  return data2;
}

UINT32 usb_wrc_rd2w(UINT8 addr)


{ // write command with address 'addr' and receive 2 data words
  UINT16  data1;
  UINT32  data2;
  outportb(ISP_COMMAND, addr);
  data1 = inportw(ISP_DATA);
  data2 = inportw(ISP_DATA);

  data2 = (data2 << 16) | data1;

  return data2;
}


UINT32 usb_wrc_rd1l(UINT8 addr)
{ // write command with address 'addr' and receive 1 data long
  UINT32  data;
  outportb(ISP_COMMAND, addr);
  data = inportl(ISP_DATA);

  return data;
}


void usb_wrc(UINT8 addr)
{ // write command with address 'addr'
  outportb(ISP_COMMAND, addr);
}

void usb_wrc_wr1b(UINT8 addr, UINT8 data)
{ // write command, write 1 data byte
  outportb(ISP_COMMAND, addr);
  outportb(ISP_DATA, data);
}

void usb_wrc_wr1w(UINT8 addr, UINT16 data)
{ // write command, write 1 data word
  outportb(ISP_COMMAND, addr);
  outportw(ISP_DATA, data);
}

void usb_wrc_wr2w(UINT8 addr, UINT16 data1, UINT16 data2)
{ // write command, write 2 data word
  outportb(ISP_COMMAND, addr);
  outportw(ISP_DATA, data1);
  outportw(ISP_DATA, data2);
}

void usb_wrc_wr1l(UINT8 addr, UINT32 data)
{ // write command, write 1 data long
  outportb(ISP_COMMAND, addr);
  outportl(ISP_DATA, data);
}

// ########################################


void  calc_endpoints(void)
{
UINT16  i;

  num_active_endpoints = 2;             // only bulk endpoints MAIN_IN and MAIN_OUT at startup
  for (i=FIRST_MAIN_EP_NUMBER; i<MAX_ENDPOINTS; i++)
  {
    if (EndpointDescrAS1[i].wMaxPacketSize)
      num_active_endpoints++;
  }

  config_descriptor_length = USB_CONFIGURATION_DESCRIPTOR_LENGTH
                               +NUM_ALTERNATE_SETTINGS * USB_INTERFACE_DESCRIPTOR_LENGTH
                               +((2 +num_active_endpoints) * USB_ENDPOINT_DESCRIPTOR_LENGTH);

  Config0Descr.wTotalLength = (UINT16) config_descriptor_length;

  Interface0DescrAS1.bNumEndpoints = (UINT8) num_active_endpoints;

  printd1(" calc_endpoints: num_active_endpoints=%d, config_descr_len=%d\n", num_active_endpoints, config_descriptor_length);
}


void  isp_init(boolean reset)
{
  UINT16  wtemp, i;
  UINT32  ltemp;

  DISABLE_INTERRUPT;

  if (reset)
  {
    DPRINT("%s: Initialize USB chip, USB-SerialNo=%s\n", MODULE_NAME, USB_SerialNo_Str);

    DPRINT("%s: Reading Interrupt edge/level register: I/O port 0x4D1=0x%X, 0x4D0=0x%X\n", MODULE_NAME, inb(0x4d1), inb(0x4d0));
    DPRINT("%s: Setting USB interrupt to level sensitivity.\n", MODULE_NAME);

    SET_INTERRUPT_LEVEL;      // Set processors interrupt sensitivity for use of USB irq to level

    DPRINT("%s: Reading Interrupt edge/level register: I/O port 0x4D1=0x%X, 0x4D0=0x%X\n", MODULE_NAME, inb(0x4d1), inb(0x4d0));

    wtemp = usb_wrc_rd1w(ISP_RD_CHIP_ID);

    if ( (wtemp == ISP_CHIPID) || (wtemp == ISP_CHIPID_B) )
    {
      DPRINT("%s: USB chip found, Chip ID = %04X\n", MODULE_NAME, wtemp);
    }
    else
    {
      printk("%s: USB chip not found! Expected: 0x%04X, found: 0x%04X\n", MODULE_NAME, ISP_CHIPID, wtemp);
      printk("%s: USB port cannot be initialized.\n", MODULE_NAME);
      return;
    }

    usb_wrc(ISP_RESET);
    delayc(1000);

  }
  else
  {
    printk("%s: Reconfigure USB chip using AlternateSetting %d\n", MODULE_NAME, AlternateSetting);
  }

  calc_endpoints();

  if (reset)
  {
    usb_wrc_wr1b(ISP_WR_MODE, ISP_MODE_SOFTCT);
    delayc(1000);
  }

  // It is truely needed to do the initialization twice to insure that the secondary buffer is not used
  //  when switching to non-double buffered mode:
  isp_SetAllMainEndpointStatus(ISP_ECR_FIFOEN);
  isp_SetAllMainEndpointStatus(0x00);
  isp_SetAllMainEndpointStatus(ISP_ECR_FIFOEN);

  ltemp = ISP_INT_CTRL_OUT | ISP_INT_CTRL_IN | ISP_INT_MAIN_OUT | ISP_INT_MAIN_IN | ISP_INT_RESET;

//  if (EnableSOFInt) // Always enable SOF interrupt to ensure regularly call of tasklet. See "tasklet_jobs".

                      // To prevent race errors, all access to ISP except initialization is done within one tasklet.
                      // To ensure that the tasklet isn't called repetitive, it is called only after an interrupt occured.
                      // To ensure that interrupts are there to keep the tasklet run through regularly, we allow every
                      //   Millisecond an SOF interrupt:
  ltemp |= ISP_INT_SOF;   // Enables interrupt on detection of SOF

  if (AlternateSetting == 1)
    for (i=EP_ISO1; i<=LAST_MAIN_EP_NUMBER; i++)
      if (EndpointDescrAS1[i- (EP_ISO1 -FIRST_ISO_EP_INDEX)].wMaxPacketSize)
      {
        ltemp |= ( ISP_INT_BASE_BIT << i );   // enable interrupt for this endpoint
      }

  DPRINT("interrupt mask: ltemp=0x%04X %04X\n", (UINT16)(ltemp >> 16), (UINT16)(ltemp & 0xffff));
  usb_wrc_wr1l(ISP_WR_IER, ltemp);
  delayc(1000);

  if (reset)
  {
    usb_wrc_wr1w(ISP_WR_HCR, ISP_HCR_PWROFF | ISP_HCR_NOLAZY | ISP_HCR_CLKRUN | ISP_HCR_DRQPOL | ISP_HCR_WKUPCS | ISP_HCR_CKDIV);
    delayc(1000);

    usb_wrc_wr1b(ISP_WR_MODE, ISP_MODE_INTENA | ISP_MODE_SOFTCT);
    delayc(1000);
  }

  MAINBuffer.FifoSizeIN = EndpointDescrAS1[EP_MAIN_IN -2].wMaxPacketSize;
  MAINBuffer.FifoSizeOUT = EndpointDescrAS1[EP_MAIN_OUT -2].wMaxPacketSize;

  ENABLE_INTERRUPT;

}   // isp_init()



void  isp_SetAllMainEndpointStatus(UINT8 status)
{
UINT8   i;
UINT8   fifo;
UINT16  buffersize;
UINT8   btemp;

  // Setting up CTRL_IN and CTRL_OUT:
  usb_wrc_wr1b(ISP_WR_ECR_CTRL_OUT, ISP_ECR_FIFOEN | ISP_ECR_FFOSZ_NI64);
  usb_wrc_wr1b(ISP_WR_ECR_CTRL_IN, ISP_ECR_FIFOEN | ISP_ECR_EPDIR_IN | ISP_ECR_FFOSZ_NI64);

  // Setting up Bulk MAIN_IN:
  btemp = status | (EnableBulkDoubleBuffer ? ISP_ECR_DBLBUF : 0x00);  // enable/disable double buffering

  if (AlternateSetting == 1)
    buffersize = EndpointDescrAS1[EP_MAIN_IN -2].wMaxPacketSize;
  else
    buffersize = EndpointDescrAS0[EP_MAIN_IN -2].wMaxPacketSize;

  if (isp_GetNonIsoFifoSize(buffersize, &fifo) > 0)
    btemp |= fifo;

  usb_wrc_wr1b(ISP_WR_ECR_MAIN_IN, btemp | ISP_ECR_EPDIR_IN);

  // Setting up Bulk MAIN_OUT:
  btemp = status;
  // Disable double buffering for BULK OUT, since we got race errors while firmware download!
  // See versions before date 04.03.2003 to see how I tried to deal with bulk OUT double buffering.

  if (AlternateSetting == 1)
    buffersize = EndpointDescrAS1[EP_MAIN_OUT -2].wMaxPacketSize;
  else
    buffersize = EndpointDescrAS0[EP_MAIN_OUT -2].wMaxPacketSize;

  if (isp_GetNonIsoFifoSize(buffersize, &fifo) > 0)
    btemp |= fifo;

  usb_wrc_wr1b(ISP_WR_ECR_MAIN_OUT, btemp);

  //Setting up isochronous and not used endpoints:
  for (i=EP_ISO1; i<=LAST_MAIN_EP_NUMBER; i++)
  {
    if (AlternateSetting == 1)
    {
      buffersize = EndpointDescrAS1[i-(EP_ISO1 -FIRST_ISO_EP_INDEX)].wMaxPacketSize;

      btemp = status | ISP_ECR_FFOISO | ISP_ECR_DBLBUF;
      if (isp_GetIsoFifoSize(buffersize, &fifo) == 0)
        btemp = 0x00;
      else
      {
        btemp |= ((EndpointDescrAS1[i-(EP_ISO1 -FIRST_ISO_EP_INDEX)].bEndpointAddress & USB_ENDPOINT_DIRECTION_MASK) ? ISP_ECR_EPDIR_IN : 0x00);
        btemp |= fifo;
      }
    }
    else
      btemp = 0x00;

    usb_wrc_wr1b(ISP_WR_ECR +i, btemp);        // ENDPX3 .. ENDPX14
    printd("i=%d, set status=0x%02X, fifo type=%d\n", i, btemp, fifo);
  }

  delayc(10000);

  for (i=EP_MAIN_IN; i<=LAST_MAIN_EP_NUMBER; i++)
  {
    btemp = usb_wrc_rd1b(ISP_RD_ECR +i);    // Get endpoint configuration
    if (btemp & ISP_ECR_FIFOEN)
    {   // this endpoint is enabled:
      usb_wrc(ISP_STALL_EP +i);   // stall/unstall endpoint to reset PID
      delayc(20000);
      usb_wrc(ISP_UNSTALL_EP +i);
    }
  }

}   // isp_SetAllMainEndpointStatus()


const UINT16   isp_fifo[17] = { 1023, 896, 768, 640, 512, 384, 320, 256, 192, 160, 128, 96, 64, 48, 32, 16, 0 };

UINT16 isp_GetIsoFifoSize(UINT16 buffersize, UINT8 *fifo)
{
UINT16  size;
UINT8   i;

  *fifo = 0;
  size = 0;

  if ((buffersize > 0) && (buffersize <= 1023))
  {
    if (buffersize == 1023)
    { *fifo = 15; size = 1023; }
    else
    {
      for (i = 1; i < 17; i++)
      {
        if (buffersize > isp_fifo[i])
        {
          *fifo = (16 -i);
          size = isp_fifo[i-1];
          break;
        }
      }
    }
  }

  return size;

}   // isp_GetIsoFifoSize


UINT16 isp_GetIsoFreeFifoSize(UINT16 buffersize)
{
UINT16  size;
UINT8   i;

  size = 0;

  if (buffersize > 0)
    for (i = 0; i < 17; i++)
      if (buffersize >= isp_fifo[i])
      {
        size = isp_fifo[i];
        break;
      }

  return size;

}   // isp_GetIsoFreeFifoSize


UINT16 isp_GetNonIsoFifoSize(UINT16 buffersize, UINT8 *fifo)
{
UINT16  size;

  if (buffersize > 64)
  { *fifo = 0; size = 0; }
  else if (buffersize > 32)
    { *fifo = ISP_ECR_FFOSZ_NI64; size = 64; }
    else if (buffersize > 16)
      { *fifo = ISP_ECR_FFOSZ_NI32; size = 32; }
      else if (buffersize > 8)
        { *fifo = ISP_ECR_FFOSZ_NI16; size = 16; }
        else if (buffersize > 0)
          { *fifo = ISP_ECR_FFOSZ_NI8; size = 8; }
          else { *fifo = 0; size = 0; }

  return size;

}   // isp_GetNonIsoFifoSize


UINT16 isp_GetCtrlEndpoint(UINT8 *buf, UINT16 max_bytes)
{ // *buf: Buffer for read data words.
  // max_bytes: number of bytes to be read.
  // return: number of bytes received.

UINT16 packet_length, get_words, got_bytes;
UINT16 wtemp, i;


  DPRINT("- ReadEPCtrlOut ");

  packet_length = usb_wrc_rd1w(ISP_RD_EPBUF_CTRL_OUT);

  if (packet_length)
  {
    if (packet_length > max_bytes)
    {
      get_words = (max_bytes +1) / 2;
      got_bytes = max_bytes;
    }
    else
    {
      get_words = (packet_length +1) / 2;
      got_bytes = packet_length;
    }

    for (i=0; i<get_words; i++)
    {
      wtemp = usb_rd1w();
      *(UINT16*)buf = wtemp; buf += 2;
      DPRINT("%04X ", (UINT16)wtemp);
    }

  }
  else
    got_bytes = 0;


  usb_wrc(ISP_CLEAR_BUFFER_CTRL_OUT);

  DPRINT("%d\n", got_bytes);

  return got_bytes;

}   // isp_GetCtrlEndpoint



UINT16  isp_WriteCtrlEndpoint(UINT8 *buf, UINT16 bytes_to_write)
{
UINT16  bytes_written;


  if (bytes_to_write > 0)
  {
    if (bytes_to_write > EP_CTRL_IN_FIFO_SIZE)
      bytes_written = EP_CTRL_IN_FIFO_SIZE;
    else
      bytes_written = bytes_to_write;

    DPRINT(" - WriteCtrlEP: Daten (%d of %d):\n", bytes_written, bytes_to_write);

    usb_wrc_wr1w(ISP_WR_EPBUF_CTRL_IN, bytes_written);
    outsw(ISP_DATA, buf, (bytes_written +1) /2);

#ifdef DEBUG_ON
{
UINT16  i;
UINT16  *pbuf;

    pbuf = (UINT16*)buf;
    for (i=0; i<(bytes_written +1)/2; i++)
    {
      DPRINT("%04X ", *(pbuf++));
    }  
}
#endif

  }
  else
  {
    bytes_written = 0;
    usb_wrc_wr1w(ISP_WR_EPBUF_CTRL_IN, 0);
  }

  usb_wrc(ISP_VALIDATE_BUFFER_CTRL_IN);

  DPRINT(" sent: %d\n", bytes_written);

  return bytes_written;

} //  isp_WriteCtrlEndpoint



UINT16 isp_GetBulkEndpoint(tBulkBuffer *BulkData, UINT8 *buf, UINT16 max_bytes)
{
UINT16  packet_length;
UINT16  got_bytes, get_words, i, wtemp;
UINT8   *bufp;

  packet_length = usb_wrc_rd1w(ISP_RD_EPBUF +BulkData->EndpointNoOUT);
  got_bytes = 0;

  printd("- ReadMainEP: packet_length=%d", packet_length);

  if (packet_length)
  {
    bufp = buf;

    if (packet_length > (max_bytes -got_bytes))
    {
      get_words = (max_bytes -got_bytes +1) / 2;
      got_bytes = (max_bytes -got_bytes);
    }
    else
    {
      get_words = (packet_length +1) / 2;
      got_bytes = packet_length;
    }

    DPRINT(", get_words=%d, got_bytes=%d\n", get_words, got_bytes);

    for (i=0; i<get_words; i++)
    { // Attention! Don't use insw()
      wtemp = usb_rd1w();
      *(bufp++) = wtemp & 0x00FF;
      *(bufp++) = (wtemp >> 8);
      printd("%04X ", wtemp);
    }
  }

  printd("\n");

  usb_wrc(ISP_CLEAR_BUFFER +BulkData->EndpointNoOUT);

  printd("got_bytes=%d\n", got_bytes);
  printd("GetBulkEP%d (%d):%.20s\n", BulkData->EndpointNoOUT, got_bytes, buf);

  return got_bytes;

}   // isp_GetBulkEndpoint



UINT16 isp_WriteBulkEndpoint(tBulkBuffer *BulkData, UINT8 *buf, UINT16 bytes_to_write)

{
UINT16  bytes_written;

  if (!(bytes_to_write))
  { // there is nothing to transmit:
    bytes_written = 0;

    if (BulkData->FullblockWritten)
    { // last transmission was exactly 64 bytes long. Send zero length packet to terminate this block:
      usb_wrc_wr1w(ISP_WR_EPBUF +BulkData->EndpointNoIN, 0);
      usb_wrc(ISP_VALIDATE_BUFFER +BulkData->EndpointNoIN);
      BulkData->FullblockWritten = FALSE;

      printd("  written: %d\n", bytes_written);
    }
  }
  else
  { // there is some data to transmit:

    printd1(" - WriteBulkEndpoint%d (%d-%d):%.20s\n", BulkData->EndpointNoIN, bytes_to_write, BulkData->FifoSizeIN, buf);

    if (bytes_to_write >= BulkData->FifoSizeIN)
    {
      // write full block:
      bytes_written = BulkData->FifoSizeIN;
      BulkData->FullblockWritten = TRUE;
    }
    else
    {
      bytes_written = bytes_to_write;
      BulkData->FullblockWritten = FALSE;
    }

    DPRINT("WriteBulkEP: (%d of %d)\n", bytes_written, bytes_to_write);

    usb_wrc_wr1w(ISP_WR_EPBUF +BulkData->EndpointNoIN, bytes_written);
    outsw(ISP_DATA, buf, (bytes_written +1) /2);
    usb_wrc(ISP_VALIDATE_BUFFER +BulkData->EndpointNoIN);

    printd1("  written: %d\n", bytes_written);
  }

  return bytes_written;

} //  isp_WriteBulkEndpoint



UINT16  isp_WriteIsoEndpoint(UINT8 endp, UINT8 *buffer, INT16 bytes_to_write)
{
INT16   bytes_written;
INT16   isp_fifosize;
UINT16  *buf16;

DPRINT(" - isp_WriteIsoEndpoint: endp=%d, bytes_to_write=%d\n", endp, bytes_to_write);

  if (bytes_to_write < 0)
    return 0;

  if (!(bytes_to_write))
  { // send zero length packet:
    bytes_written = 0;

    usb_wrc_wr1w(ISP_WR_EPBUF +endp, 0);
    usb_wrc(ISP_VALIDATE_BUFFER +endp);
  }
  else
  { // there is some data to transmit:

    isp_fifosize = EndpointDescrAS1[endp -EP_ISO1 +2].wMaxPacketSize;

    if (bytes_to_write >= isp_fifosize)
    {
      bytes_written = isp_fifosize;
    }
    else
    {
      bytes_written = bytes_to_write;
    }

    usb_wrc_wr1w(ISP_WR_EPBUF +endp, bytes_written);
    buf16 = (UINT16*)buffer;
    outsw(ISP_DATA, buf16, (bytes_written +1) /2);
    usb_wrc(ISP_VALIDATE_BUFFER +endp);
  }

  DPRINT("  written: %d\n", bytes_written);

  return bytes_written;

} //  isp_WriteIsoEndpoint

