/*
 * jtag.c - JTAG device for Simmux2
 *
 * This file contains the routines to output values to and read it
 * from the JTAG ports.
 *
 */


/* $Id: jtag_core.c,v 1.1.1.1 2001/12/13 13:09:21 kanne Exp $ */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/miscdevice.h>	/* JTAG_MINOR */
#include <asm/uaccess.h>

#include "jtag_port.h"

#define REVISION "$Revision: 1.1.1.1 $"
static char jtag_version[] =
		"$Id: jtag_core.c,v 1.1.1.1 2001/12/13 13:09:21 kanne Exp $"
		" (built " __DATE__ " " __TIME__ ")";

//#undef KERN_DEBUG
//#define KERN_DEBUG
//#define __DEBUG__


/*
 *  sl
 *  define global values for base-port and xcvf-file, so that we can change it at load time!
 *	( 'insmod mydriver myvar1=wert1 myvar2="Hallo Welt"'
 */

static int my_use_count = 0;

static int jtag_open(struct inode *inode, struct file *file)
{
    if (my_use_count >= 1) {
	printk(KERN_ERR "jtag_open: Device is busy (my_use_count=%d)!\n", my_use_count);
	return -EBUSY;
    }

    my_use_count++;
    MOD_INC_USE_COUNT;
	
#ifdef __DEBUG__
    printk(KERN_INFO "jtag_open: done (my_use_count=%d)\n", my_use_count);
#endif

    return 0;
}

static int jtag_release(struct inode *inode, struct file *file)
{
//    SetAllDirectionReg(_READ_);          // Clear all (write mode)

    my_use_count--;
    MOD_DEC_USE_COUNT;
	
#ifdef __DEBUG__
    printk(KERN_INFO "jtag: released (my_use_count=%d).\n", my_use_count);
#endif

    return 0;
}

static ssize_t jtag_read(struct file * file, char * buf, size_t count, loff_t *ppos)
{
    jtag_byte byte;

    byte = read_jtag_byte();
#ifdef __DEBUG__
    printk(KERN_INFO "jtag_read: readed %.02X\n", byte.value);
#endif

    if (copy_to_user(buf, &byte.value, 1)) {
	printk(KERN_ERR "jtag_read: cann't copy data to userspace.\n");
	return -EFAULT;
    }
    
    return 1;
}


static ssize_t jtag_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
    /*unsigned*/ char buffer[len];

    jtag_byte byte;
    int count = 0;

#if 0
	/*  Can't seek (pwrite) on this device  */
    if (ppos != &file->f_pos) {
#ifdef __DEBUG__    
	printk(KERN_ERR "jtag_write: cann't seek\n");
#endif
	return -ESPIPE;
    }
#endif

	
    if (len > 0) {
	if (copy_from_user(buffer,data,len) != 0) {
	    printk(KERN_ERR "jtag_write: cann't copy data from userspace!\n");
	    return -EFAULT;
	}

        for (count=0; count < len; count++) {
	    byte.value = buffer[count];
	    write_jtag_byte(byte);
#ifdef __DEBUG__
	    printk(KERN_INFO "jtag_write: writing %.02X\n", byte.value);
#endif
        }
    }
    
    return len;
}

static struct file_operations jtag_fops=
{
    owner:	THIS_MODULE,
    read:	jtag_read,
    write:	jtag_write,
//    ioctl:	jtag_ioctl,
    open:	jtag_open,
    release:	jtag_release,
};


static struct miscdevice jtag_miscdev=
{
	JTAG_MINOR,
	"jtag",
	&jtag_fops
};

static int __init jtag_module_init(void)
{
    int error = 0;
    my_use_count = 0;

#if 0	
	// Ports initialisieren, Default ist Low (0)
	SetAllDirectionReg(_WRITE_);
	SetDirectionRegister(1, _WRITE_);
	SetTestHighLow(1/*an_aus*/);
    //SetTestHighLow(0);
	SetAllDirectionReg(_WRITE_);

#endif
    setup_ports();

    error = misc_register(&jtag_miscdev);
    if (error) {
	printk(KERN_ERR "jtag: cann't register driver\n");
    }
#ifdef __DEBUG__
    else {
	printk(KERN_INFO "jtag: driver loaded, " REVISION "\n");
    }
#endif
    return (error);
}

static void __exit jtag_module_exit(void)
{
    if (misc_deregister(&jtag_miscdev)) {
	printk(KERN_ERR "jtag: cann't unregister driver\n");
    }
#ifdef __DEBUG__
    else {
	printk(KERN_INFO "jtag: driver unloaded\n");
    }
#endif
}

module_init(jtag_module_init);
module_exit(jtag_module_exit);
