/*
 * $Id$
 *
 * Handle mapping of the NOR flash on XL-Web 855T boards
 *
 * Copyright 2002 SYSGO Real-Time Solutions GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define WINDOW_ADDR 0x40000000		/* physical address of flash */
#define WINDOW_SIZE 0x00100000		/* for now the flash size is 1MB */
#define BUSWIDTH 2 

/* can be { "cfi_probe", "jedec_probe", "map_rom", 0 }; */
#define PROBETYPES { "jedec_probe", 0 }

#define MSG_PREFIX "XLWeb: "   /* prefix for our printk()'s */

static struct mtd_info *mymtd;

__u8 xlweb_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 xlweb_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 xlweb_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

#ifdef CFI_WORD_64
__u64 xlweb_read64(struct map_info *map, unsigned long ofs)
{
	return __raw_readll(map->map_priv_1 + ofs);
}
#endif

void xlweb_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void xlweb_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void xlweb_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void xlweb_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

#ifdef CFI_WORD_64
void xlweb_write64(struct map_info *map, __u64 d, unsigned long adr)
{
	__raw_writell(d, map->map_priv_1 + adr);
	mb();
}
#endif

void xlweb_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info xlweb_map = {
	name: "XLWeb NOR Flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: xlweb_read8,
	read16: xlweb_read16,
	read32: xlweb_read32,
#ifdef CFI_WORD_64
	read64: xlweb_read64,
#endif
	copy_from: xlweb_copy_from,
	write8: xlweb_write8,
	write16: xlweb_write16,
	write32: xlweb_write32,
#ifdef CFI_WORD_64
	write64: xlweb_write64,
#endif
	copy_to: xlweb_copy_to
};

#ifdef CONFIG_MTD_PARTITIONS

static struct mtd_partition static_partitions[] =
{
		{
				name:	"ppcboot1",
				size:	0x4000,
				offset:	0x00000000,
				mask_flags: MTD_WRITEABLE
		},
		{
				name:	"ppcboot_env",
				size:	0x2000,
				offset:	0x00004000
		},
		{
				name:	"ppcboot2",
				size:	0x1A000,
				offset:	0x00006000,
				mask_flags: MTD_WRITEABLE
		},
		{
				name:	"linux1",
				size:	0x70000,
				offset:	0x00020000
		},
		{
				name:	"linux2",
				size:	0x70000,
				offset:	0x00090000
		},
};

#define NB_OF(x) (sizeof (x) / sizeof (x[0]))

#ifdef CONFIG_MTD_CMDLINE_PARTS
int parse_cmdline_partitions(struct mtd_info *master,
					struct mtd_partition **pparts,
					const char *mtd_id);

static struct mtd_partition *mtd_parts = 0;
static int                   mtd_parts_nb = 0;
#endif

#endif

int __init init_xlweb(void)
{
	static const char *rom_probe_types[] = PROBETYPES;
	const char **type;
	const char *part_type = 0;

	printk(KERN_NOTICE MSG_PREFIX "flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	xlweb_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	if (!xlweb_map.map_priv_1) {
		printk(MSG_PREFIX "Failed to ioremap\n");
		return -EIO;
	}
	
	mymtd = 0;
	type = rom_probe_types;
	for(; !mymtd && *type; type++) {
		mymtd = do_map_probe(*type, &xlweb_map);
	}
	if (mymtd) {
		mymtd->module = THIS_MODULE;

		add_mtd_device(mymtd);
#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
		mtd_parts_nb = parse_cmdline_partitions(mymtd, &mtd_parts, 
							"xlweb");
		if (mtd_parts_nb > 0)
				part_type = "command line";
#endif
		if (mtd_parts_nb <= 0)
		{
				mtd_parts = static_partitions;
				mtd_parts_nb = NB_OF(static_partitions);
				part_type = "static";
		}
		if (mtd_parts_nb <= 0)
		{
				printk(KERN_NOTICE MSG_PREFIX
							"no partition info available\n");
		}
		else
		{
				printk(KERN_NOTICE MSG_PREFIX
			       "Using %s partition definition\n",
						 part_type);
				add_mtd_partitions (mymtd, mtd_parts, mtd_parts_nb);
		}
#endif
		return 0;
	}

	iounmap((void *)xlweb_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_xlweb(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (xlweb_map.map_priv_1) {
		iounmap((void *)xlweb_map.map_priv_1);
		xlweb_map.map_priv_1 = 0;
	}
}

module_init(init_xlweb);
module_exit(cleanup_xlweb);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Armin Schindler <acs@sysgo.de>");
MODULE_DESCRIPTION("MTD map driver for XL-WEB 855T");
