/*
 * ppc405.c
 *
 * Ethernet driver for the built in ethernet on the IBM 405 PowerPC
 * processor.
 *
 * Author: Johnnie Peters
 *         jpeters@mvista.com
 *
 * Copyright 2000 MontaVista Softare Inc. 
 *  
 * This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.    
 *  
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/processor.h>      /* Processor type for cache alignment. */
#include <asm/bitops.h>
#include <asm/io.h>

#include  <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include "ppc405_enet.h"

#define MDIO_READ(phy, reg, val)        ppc405_enet_mdio_read(phy, reg, val)
#define MDIO_WRITE(phy, reg, val)       ppc405_enet_mdio_write(phy, reg, val)


int	ppc405_enet_probe(struct net_device *);
static int	ppc405_enet_probe1(struct net_device *);
static int	ppc405_phy_speed(void);
static int	ppc405_phy_duplex(void);
static int      ppc405_enet_mdio_read(int phy, int reg, uint16_t *value);
static int      ppc405_enet_mdio_write(int phy, int reg, uint16_t value);
static int      ppc405_enet_ioctl(struct net_device *, struct ifreq *rq, int cmd);
static int	ppc405_phy_dump(char *log_level);
static int	ppc405_enet_open(struct net_device *);
static int	ppc405_enet_start_xmit(struct sk_buff *, struct net_device *);
static struct net_device_stats *ppc405_enet_stats(struct net_device *);
static int	ppc405_enet_close(struct net_device *);
static void ppc405_enet_set_multicast_list(struct net_device *);

static void	ppc405_eth_wakeup(int, void *, struct pt_regs *);
static void	ppc405_eth_serr  (int, void *, struct pt_regs *);
static void	ppc405_eth_txeob (int, void *, struct pt_regs *);
static void	ppc405_eth_rxeob (int, void *, struct pt_regs *);
static void	ppc405_eth_txde  (int, void *, struct pt_regs *);
static void	ppc405_eth_rxde  (int, void *, struct pt_regs *);
static void	ppc405_eth_mac   (int, void *, struct pt_regs *);

#ifdef CONFIG_PPC405_ENET_ERROR_MSG
static void	ppc405_eth_desc_dump(char *log_level);
static void	ppc405_eth_emac_dump(char *log_level);
static void	ppc405_eth_mal_dump (char *log_level);
#endif

static unsigned long crc32(char *, int);

/* Physical mapping of ethernet register space. */
static struct	ppc405_enet_regs *ppc405_enet_regp =
			(struct ppc405_enet_regs *)PPC405_EM0_REG_ADDR;

/* structures to define the ring buffer access. */
static mal_desc_t	*ppc405_enet_tx;
static mal_desc_t	*ppc405_enet_rx;
static struct sk_buff	*ppc405_skb_rx[NUM_RX_BUFF];

static volatile int	ppc405_enet_tx_slot = 0;
static volatile int	ppc405_enet_ack_slot = 0;
static volatile int	ppc405_enet_rx_slot = 0;


/*
 * Maintain the initial probe as a sub function just in case it is wished
 * to have this driver as a module later.
 */
int ppc405_enet_probe_done = 0;

int
ppc405_enet_probe(struct net_device *dev)
{
	if (ppc405_enet_probe_done)
		return -ENODEV;

	ppc405_enet_probe1(dev);

	ppc405_enet_probe_done = 1;
	return 0;
}


static int
ppc405_enet_open(struct net_device *dev)
{
	int loop;
	unsigned long mode_reg;
	struct ppc405_enet_private *lo_priv = dev->priv;
	dma_addr_t rx_phys_addr;
	dma_addr_t tx_phys_addr;
#if 1
	int *rx_cached_addr;
	int *tx_cached_addr;
#endif
	unsigned long emac_ier;
	unsigned short ctrl;

	emac_ier = 0;

	/* although the probe has already reset the chip we do it again here */
	printk(KERN_NOTICE "Reset ethernet interfaces\n");
	/* Reset the MAL */
	mtdcr(DCRN_MALCR, MALCR_MMSR);

	/* Reset the EMAC */
	ppc405_enet_regp->em0mr0 = EMAC_M0_SRST;
	eieio();
	for (loop = 0; loop < 1000; loop++);
	ppc405_enet_regp->em0mr0 = ppc405_enet_regp->em0mr0 & ~EMAC_M0_SRST;
	eieio();


	/* Set the MAL configuration register */
	mtdcr(DCRN_MALCR, MALCR_PLBB | MALCR_OPBBL | MALCR_LEA |
							MALCR_PLBLT_DEFAULT);

	/* Allocate memory for the transmit descriptors and save its physical
	 * address.
	 */

#if 0
	ppc405_enet_tx = (mal_desc_t *) consistent_alloc(GFP_KERNEL, PAGE_SIZE,
						&tx_phys_addr);

	ppc405_enet_rx = (mal_desc_t *) consistent_alloc(GFP_KERNEL, PAGE_SIZE,
						&rx_phys_addr);
#else
	/* Allocate memory for the transmit descriptors and save its physical
	 * address.
	 */
	tx_cached_addr = (int *)(void *)get_free_page(GFP_KERNEL);
	tx_phys_addr   = (dma_addr_t)virt_to_phys(tx_cached_addr);

	/* Map the transmit descriptors into non cached memory and then flush
	 * original map out of cache.
	 */
	ppc405_enet_tx = (mal_desc_t *) ioremap((unsigned long)
				tx_phys_addr, PAGE_SIZE);

	dma_cache_wback_inv((unsigned long)tx_cached_addr, PAGE_SIZE);
	dma_cache_wback_inv((unsigned long)ppc405_enet_tx, PAGE_SIZE);

	/* Repeat the previous steps for the receive descriptors. */
	rx_cached_addr = (int *)(void *)get_free_page(GFP_KERNEL);
	rx_phys_addr   = (dma_addr_t)virt_to_phys(rx_cached_addr);

	ppc405_enet_rx = (mal_desc_t *) ioremap((unsigned long)
				rx_phys_addr, PAGE_SIZE);

	dma_cache_wback_inv((unsigned long)rx_cached_addr, PAGE_SIZE);
	dma_cache_wback_inv((unsigned long)ppc405_enet_rx, PAGE_SIZE);
#endif
   
	/* Fill in the transmit descriptor ring.  */
	for (loop = 0; loop < NUM_TX_BUFF; loop++) {
		ctrl = 0;
		if ((NUM_TX_BUFF - 1) == loop)
			ctrl |= MAL_TX_CTRL_WRAP;
		ppc405_enet_tx[loop].ctrl     = ctrl;
		ppc405_enet_tx[loop].data_len = 0;
		ppc405_enet_tx[loop].data_ptr = NULL;
	}

	/* Format the receive descriptor ring. */
	for (loop = 0; loop < NUM_RX_BUFF; loop++) {
		ctrl =  MAL_RX_CTRL_EMPTY | MAL_RX_CTRL_INTR;
		if ((NUM_RX_BUFF - 1) == loop)
			ctrl |= MAL_RX_CTRL_WRAP;
		ppc405_skb_rx[loop]           = dev_alloc_skb(DESC_BUF_SIZE);
		ppc405_enet_rx[loop].data_len = 0;
		ppc405_enet_rx[loop].data_ptr =
				(char *)virt_to_phys(ppc405_skb_rx[loop]->data);
		ppc405_enet_rx[loop].ctrl     =  ctrl;

		dma_cache_wback_inv((unsigned long)
				ppc405_skb_rx[loop]->data, DESC_BUF_SIZE);
	}

	ppc405_enet_rx_slot  = 0;
	ppc405_enet_tx_slot  = 0;
	ppc405_enet_ack_slot = 0;

	/* setup MAL tx and rx channel pointers */
	mtdcr(DCRN_MALTXCTP0R, tx_phys_addr);
	mtdcr(DCRN_MALRXCTP0R, rx_phys_addr);

	/* Do not need to worry about shared irq's.  this interface has
	 * multiple interrupt handlers and each has a unique irq.  It will
	 * be noticed that the actual interupt handler for irq 9 (wakeup)
	 * is empty.
	 */
	request_irq( 9, &ppc405_eth_wakeup, 0, "405eth Wakeup",   dev);
	request_irq(10, &ppc405_eth_serr,   0, "405eth MAL SERR", dev);
	request_irq(11, &ppc405_eth_txeob,  0, "405eth TX EOB",   dev);
	request_irq(12, &ppc405_eth_rxeob,  0, "405eth RX EOB",   dev);
	request_irq(13, &ppc405_eth_txde,   0, "405eth TX DE",    dev);
	request_irq(14, &ppc405_eth_rxde,   0, "405eth RX DE",    dev);
	request_irq(15, &ppc405_eth_mac  ,  0, "405eth MAC",      dev);

	/* set the high address */
	ppc405_enet_regp->em0iahr = (dev->dev_addr[0] << 8) | dev->dev_addr[1];
	eieio();

	/* set the low address */
	ppc405_enet_regp->em0ialr =
		      (dev->dev_addr[2] << 24)
	            | (dev->dev_addr[3] << 16)
		    | (dev->dev_addr[4] <<  8)
		    |  dev->dev_addr[5];
	eieio();

	/* reset the transmit and receive channels */
	mtdcr(DCRN_MALRXCARR, 0x80000000);
	mtdcr(DCRN_MALTXCARR, 0xc0000000);

	/* set the rx buffer size */
	/* revisit: 128 bytes???? why not DESC_BUF_SIZE? */
	mtdcr(DCRN_MALRCBS0, DESC_BUF_SIZE_REG);

	/* set receive fifo to 4k and tx fifo to 2k */
	mode_reg = EMAC_M1_RFS_4K | EMAC_M1_TX_FIFO_2K | EMAC_M1_APP |
		   EMAC_M1_TR0_MULTI;

	/* set speed (default is 10Mb) */
	if (lo_priv->ep_speed == _100BASET)
		mode_reg = mode_reg | EMAC_M1_MF_100MBPS;

	if( lo_priv->ep_duplex == FULL)
		mode_reg = mode_reg | EMAC_M1_FDE | EMAC_M1_EIFC;

	ppc405_enet_regp->em0mr1 = mode_reg;
	eieio();

	/* enable broadcast and individual address */
	ppc405_enet_regp->em0rmr = EMAC_RMR_IAE | EMAC_RMR_BAE;
	eieio();

	/* set transmit request threshold register */
	ppc405_enet_regp->em0trtr = EMAC_TRTR_256;
	
	eieio();
	/* set receive low/high water mark register */
	ppc405_enet_regp->em0rwmr = 0x0f002000;
	eieio();

	/* set frame gap */
	ppc405_enet_regp->em0ipgvr = CONFIG_PPC405_ENET_GAP;
	eieio();

	netif_start_queue(dev);

	for (loop = 0; loop < NUM_TX_BUFF; loop++) {
		lo_priv->ep_xmit_skb[loop] = (struct sk_buff *)NULL;
	}

	lo_priv->ep_xmit_pend         = (struct ppc405_skb_list *)NULL;
	lo_priv->ep_xmit_pend_cnt     = 0;
	lo_priv->ep_xmit_active_count = 0;


	/* set the MAL IER */
	mtdcr(DCRN_MALIER, MALIER_DE |
			MALIER_NE | MALIER_TE | MALIER_OPBE | MALIER_PLBE);

	/* set the EMAC IER */
	emac_ier = EMAC_ISR_PP   | EMAC_ISR_BP   | EMAC_ISR_SE |
		   EMAC_ISR_PTLE | EMAC_ISR_BFCS |
		   EMAC_ISR_ORE  | EMAC_ISR_IRE;

	ppc405_enet_regp->em0iser = emac_ier;
	eieio();

	/* enable MAL transmit channel 0 and receive channel 0 */
	mtdcr(DCRN_MALRXCASR, 0x80000000);
	mtdcr(DCRN_MALTXCASR, 0x80000000);

	/* set transmit and receive enable */
	ppc405_enet_regp->em0mr0 = EMAC_M0_TXE | EMAC_M0_RXE;
	eieio();

	printk(KERN_NOTICE "%s: PPC405 Enet open completed\n", dev->name);

	return 0;
}

/*
 * static int ppc405_enet_mdio_read()
 *
 * Description:
 *   This routine reads from a specified register on a PHY over the MII
 *   Management Interface.
 *
 * Input(s):
 *   phy    - The address of the PHY to read from. May be 0 through 31.
 *   reg    - The PHY register to read. May be 0 through 31.
 *   *value - Storage for the value to read from the PHY register.
 *
 * Output(s):
 *   *value - The value read from the PHY register.
 *
 * Returns:
 *   0 if OK, otherwise -1 on error.
 *
 */
static int
ppc405_enet_mdio_read(int phy, int reg, uint16_t *value)
 {
	register int i = 0;
	uint32_t stacr;

	/* Wait for data transfer complete bit */
	i = 0;
        while ((ppc405_enet_regp->em0stacr & EMAC_STACR_OC) == 0) {
                udelay(7);
		if (i == 5) {
		       return (-1);
		}
		i++;
	}

	/* Clear the speed bits and make a read request to the PHY */

	stacr = reg & 0x1F;
	stacr = (stacr | EMAC_STACR_READ) & ~EMAC_STACR_CLK_100MHZ;
	stacr = stacr | ((phy & 0x1F) << 5);

	ppc405_enet_regp->em0stacr = stacr;
	eieio();

	/* Wait for data transfer complete bit */
	i = 0;
	while (((stacr = ppc405_enet_regp->em0stacr) & EMAC_STACR_OC) == 0) {
	    	eieio();
		udelay(7);
		if (i == 5) {
			return (-1);
		}
		i++;
	}

	 /* Check for a read error */
	if ((stacr & EMAC_STACR_PHYE) !=0) {
		return (-1);
	}

	*value = (uint16_t)(stacr >> 16);

	return (0);
}

/*
 * static int ppc405_enet_mdio_write()
 *
 * Description:
 *   This routine writes to a specified register on a PHY over the MII
 *   Management Interface.
 *
 * Input(s):
 *   phy   - The address of the PHY to write to. May be 0 through 31.
 *   reg   - The PHY register to write. May be 0 through 31.
 *   value - The value to write to the PHY register.
 *
 * Output(s):
 *   N/A
 *
 * Returns:
 *   0 if OK, otherwise -1 on error.
 *
 */
static int
ppc405_enet_mdio_write(int phy, int reg, uint16_t value)
{
	register int i;
	uint32_t stacr;

	/* Wait for data transfer complete bit */
	i = 0;
	while ((ppc405_enet_regp->em0stacr & EMAC_STACR_OC) == 0) {
		udelay(7);
		if (i == 5) {
			return (-1);
		}
		i++;
	}

	/* Clear the speed bits and make a write request to the PHY */
	stacr = reg & 0x1F;
	stacr = (stacr | EMAC_STACR_WRITE) & ~EMAC_STACR_CLK_100MHZ;
	stacr = stacr | ((phy & 0x1F) << 5);
	stacr = stacr | ((value & 0xFFFF) << 16);

	ppc405_enet_regp->em0stacr = stacr;
	asm("eieio");

	/* Wait for data transfer complete bit */
	i = 0;
	while (((stacr = ppc405_enet_regp->em0stacr) & EMAC_STACR_OC) == 0) {
		eieio();
		udelay(7);
		if (i == 5) {
			return (-1);
		}
		i++;
	}

	/* Check for a write error */
	if ((stacr & EMAC_STACR_PHYE) !=0)
		return (-1);

	return (0);
}

/*
 * static int ppc405_enet_ioctl()
 *
 * Description:
 *   This routine performs the specified I/O control command on the
 *   specified net device.
 *
 * Input(s):
 *   *dev - Pointer to the device structure for this driver.
 *   *rq  - Pointer to data to be written and/or storage for return data.
 *    cmd - I/O control command to perform.
 *
 * Output(s):
 *   *rq  - If OK, pointer to return data for a read command.
 *
 * Returns:
 *   0 if OK, otherwise an error number on error.
 *
 */
static int
ppc405_enet_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
        u16 *data = (u16 *)&rq->ifr_data;

        switch (cmd) {

        case SIOCDEVPRIVATE:
                data[0] = PHY_ADDR;
                /*FALLTHRU*/

        case SIOCDEVPRIVATE + 1:
                if (MDIO_READ(data[0], data[1], &data[3]) < 0)
                        return (-EIO);

                return (0);

        case SIOCDEVPRIVATE + 2:
                if (!capable(CAP_NET_ADMIN))
                        return (-EPERM);

                if (MDIO_WRITE(data[0], data[1], data[2]) < 0)
                        return (-EIO);

                return (0);

        default:
                return (-EOPNOTSUPP);
        }
 }


static int
ppc405_enet_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	unsigned short			ctrl;
	int				frame_length;
	unsigned long			flags;
	struct ppc405_enet_private	*lo_priv;


	lo_priv      = dev->priv;

	save_flags(flags);
	cli();

	if (netif_queue_stopped(dev) ||
			(lo_priv->ep_xmit_active_count == NUM_TX_BUFF)) {

		lo_priv->ep_stats.tx_dropped++;

		restore_flags(flags);

		return -EBUSY;
	}

	frame_length = (int)(skb->tail - skb->data);

	if (++lo_priv->ep_xmit_active_count == NUM_TX_BUFF)
		netif_stop_queue(dev);

	/* Store the skb buffer for later ack by the transmit end of buffer
	 * interrupt.
	 */
	lo_priv->ep_xmit_skb[ppc405_enet_tx_slot] = skb;

	dma_cache_wback_inv((unsigned long)skb->data, frame_length);

	ctrl = EMAC_TX_CTRL_DFLT;
	if ((NUM_TX_BUFF - 1) == ppc405_enet_tx_slot)
		ctrl |= MAL_TX_CTRL_WRAP;

	ppc405_enet_tx[ppc405_enet_tx_slot].data_ptr = (char *)virt_to_phys(skb->data);
	ppc405_enet_tx[ppc405_enet_tx_slot].data_len = (short)frame_length;
	ppc405_enet_tx[ppc405_enet_tx_slot].ctrl     = ctrl;

	/* Send the packet out. */
	ppc405_enet_regp->em0tmr0 = EMAC_TXM0_GNP0;

	if(++ppc405_enet_tx_slot == NUM_TX_BUFF)
		ppc405_enet_tx_slot = 0;

	lo_priv->ep_stats.tx_packets++;
	lo_priv->ep_stats.tx_bytes += frame_length;

	restore_flags(flags);

	return 0;
}


static int
ppc405_enet_close(struct net_device *dev)
{
	int delay;

	/* Free the irq's */
	free_irq(BL_MAC0_WOL,   dev);
	free_irq(BL_MAL0_SERR,  dev);
	free_irq(BL_MAL0_TXEOB, dev);
	free_irq(BL_MAL0_RXEOB, dev);
	free_irq(BL_MAL0_TXDE,  dev);
	free_irq(BL_MAL0_RXDE,  dev);
	free_irq(BL_MAC0_ETH,   dev);

	/* reset the MAL and EMAC */
	mtdcr(DCRN_MALCR, MALCR_MMSR);

        ppc405_enet_regp->em0mr0 = EMAC_M0_SRST;
	eieio();

        for (delay = 0; delay < 1000; delay++);

        ppc405_enet_regp->em0mr0 = ppc405_enet_regp->em0mr0 & ~EMAC_M0_SRST;
	eieio();

	/*
	 * Unmap the non cached memory space.
	 */
#if 1    
	iounmap((void *)ppc405_enet_tx);
	iounmap((void *)ppc405_enet_rx);
#else
	consistent_free((void *)ppc405_enet_tx, PAGE_SIZE);
	consistent_free((void *)ppc405_enet_rx, PAGE_SIZE);
#endif    
	/*
	 * revisit
	 * need to free memory allocated by get_free_page()
	 *   tx_cached_addr
	 *   rx_cached_addr
	 * need to free any skbufs we were transmitting
	 * need to free any skbufs in the receive queue
	 * ...
	 */

	return 0;
}


static void
ppc405_enet_set_multicast_list(struct net_device *dev)
{
	struct dev_mc_list *dmi = dev->mc_list;
	unsigned char *mc_addr;
	unsigned long mc_crc;
	int bit_number;
	int dmi_count;

	/* If promiscuous mode is set then we do not need anything else */
	if (dev->flags & IFF_PROMISC) {	
		ppc405_enet_regp->em0rmr = EMAC_RMR_PME;
		eieio();
		return;
	}

	/* If multicast mode is not set then we are turning it off at this point */
	if (!(dev->flags & IFF_MULTICAST)) {
		ppc405_enet_regp->em0rmr = EMAC_RMR_IAE | EMAC_RMR_BAE;
		eieio();
		return;
	}

	/* Must be setting up to use multicast.  Now check for promiscuous
	 * multicast
	 */
	if (dev->flags & IFF_ALLMULTI) {
		ppc405_enet_regp->em0rmr |= EMAC_RMR_PMME;
		eieio();
		return;
	}

	/* Turn on multicast addressing */
	ppc405_enet_regp->em0rmr |= EMAC_RMR_MAE;

	/* Need to hash on the multicast address. */
	for (dmi_count = 0; dmi_count < dev->mc_count; dmi_count++) {
		mc_addr = dmi->dmi_addr;
		dmi = dmi->next;
		if (*mc_addr & 0x01) {		/* This is a multicast address */
			mc_crc = crc32(mc_addr, 6);
			bit_number = 63 - (int)(mc_crc >> 26);

			switch (bit_number & 0x30) {  /* determine the group register */
			case 0x00:
				ppc405_enet_regp->em0gaht1 =
						ppc405_enet_regp->em0gaht1 | (0x8000 >> bit_number);
				break;
			case 0x10:
				ppc405_enet_regp->em0gaht2 =
						ppc405_enet_regp->em0gaht2 | (0x8000 >> bit_number);
				break;
			case 0x20:
				ppc405_enet_regp->em0gaht3 =
						ppc405_enet_regp->em0gaht3 | (0x8000 >> bit_number);
				break;
			case 0x30:
				ppc405_enet_regp->em0gaht4 =
						ppc405_enet_regp->em0gaht4 | (0x8000 >> bit_number);
				break;
			}
		}
	};

	return;
}


static struct net_device_stats *
ppc405_enet_stats(struct net_device *dev)
{
	struct ppc405_enet_private *lo_priv = dev->priv;

	return &lo_priv->ep_stats;
}


static int
ppc405_enet_probe1(struct net_device *dev)
{
	int	delay, i;
	bd_t	*bd;
	struct ppc405_enet_private *lo_priv;

	/* Reset the MAL */
	mtdcr(DCRN_MALCR, MALCR_MMSR);

	/* Reset the EMAC */
	ppc405_enet_regp->em0mr0 = EMAC_M0_SRST;
	eieio();
	for (delay = 0; delay < 1000; delay++);
	ppc405_enet_regp->em0mr0 = ppc405_enet_regp->em0mr0 & ~EMAC_M0_SRST;
	eieio();
	
	/* read the MAC Address */
	bd = (bd_t *)__res;
	for (i=0; i<6; i++) {
		dev->dev_addr[i] = bd->bi_enetaddr[i];
	}

	dev->base_addr = 0xffe0;  /* indicate no actual physical probing */
	dev->irq = BL_MAC0_WOL;   /* the first ethernet irq - need something here */

	/* initialize the private data pointer */
	lo_priv = (void *)(((long)kmalloc(
		sizeof(struct ppc405_enet_private),
		GFP_KERNEL | GFP_DMA) + 7) & ~7);
	memset(lo_priv, 0, sizeof(struct ppc405_enet_private));
	dev->priv = lo_priv;

	/* Find out the default network settings from the phy */
	lo_priv->ep_speed  = ppc405_phy_speed();
	lo_priv->ep_duplex = ppc405_phy_duplex();

	printk(KERN_NOTICE "%s: PPC405 EMAC %d Mbs %s duplex ",
			dev->name, lo_priv->ep_speed,
			(lo_priv->ep_duplex == HALF)? "Half": "Full");

	/* no KERN_NOTICE, this continues previous printk */
	printk("MAC %02x:%02x:%02x:%02x:%02x:%02x\n",
			dev->dev_addr[0], dev->dev_addr[1], dev->dev_addr[2],
			dev->dev_addr[3], dev->dev_addr[4], dev->dev_addr[5]);

	/* Fill in the driver function table */
	dev->open               = &ppc405_enet_open;
	dev->hard_start_xmit    = &ppc405_enet_start_xmit;
	dev->stop               = &ppc405_enet_close;
	dev->get_stats          = &ppc405_enet_stats;
	dev->set_multicast_list = &ppc405_enet_set_multicast_list;
	dev->do_ioctl = &ppc405_enet_ioctl;
	
	/* Fill in the generic fields of the device structure. */
	ether_setup(dev);
	return 0;
}


static int
ppc405_phy_speed(void)
{
	int speed = _10BASET;
	unsigned short bmcr = 0x0;

	if (MDIO_READ(PHY_ADDR, PHY_ANLPAR, &bmcr)){
		printk(KERN_ERR "phy speed read failed \n");
		ppc405_phy_dump(KERN_ERR);
	}

	if ((bmcr & PHY_ANLPAR_100) != 0) {
		speed = _100BASET;
	}

	return (speed);
}


static int
ppc405_phy_duplex(void)
{
	int duplex = HALF;          /* Assume HALF */
	unsigned short bmcr = 0x0;

	if (MDIO_READ(PHY_ADDR, PHY_ANLPAR, &bmcr)){
		printk(KERN_ERR "phy duplex read failed \n\r");
		ppc405_phy_dump(KERN_ERR);
	}

	if ((bmcr & ( PHY_ANLPAR_10FD | PHY_ANLPAR_TXFD)) != 0) 
		duplex = FULL;
  
	return (duplex);
}


static int
ppc405_phy_dump(char *log_level)
{
	unsigned long i;
	unsigned short data;

	for (i = 0; i < 0x1A; i++) {
                if (MDIO_READ(PHY_ADDR, i, &data)) {
			return -1;
		}

		printk("%sPhy reg 0x%x ==> %4x\n", log_level, (unsigned int)i, data);

		if (i == 0x07) i = 0x0f;
	}
	return 0;
}


static void
ppc405_eth_wakeup(int irq, void *dev_instance, struct pt_regs *regs)
{
	/* On Linux the 405 ethernet will always be active if configured
	 * in.  This interrupt should never occur.
	 */
	printk(KERN_INFO "interrupt ppc405_eth_wakeup\n");
}


static void
ppc405_eth_serr(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *dev;
	struct ppc405_enet_private *lo_priv;
	unsigned long mal_error;
#ifdef CONFIG_PPC405_ENET_ERROR_MSG
	unsigned long plb_error;
	unsigned long plb_addr;
#endif

	dev     = (struct net_device *)dev_instance;
	lo_priv = dev->priv;

	mal_error = mfdcr(DCRN_MALESR);

	if (mal_error & MALESR_EVB) {

		if (mal_error & MALESR_CIDRX) {
			lo_priv->ep_stats.rx_errors++;
		} else {
			lo_priv->ep_stats.tx_errors++;
		}

#ifdef CONFIG_PPC405_ENET_ERROR_MSG
		printk(KERN_ERR "ppc405_eth_serr: %s channel %ld \n",
					(mal_error & 0x40000000)? "Receive" : "Transmit",
		 			(mal_error & 0x3d000000) >> 25);

		printk(KERN_ERR "  -----  latched error  -----\n");
		if (mal_error & MALESR_DE)
			printk(KERN_ERR "  DE: descriptor error\n");
		if (mal_error & MALESR_OEN)
			printk(KERN_ERR "  ONE: OPB non-fullword error\n");
		if (mal_error & MALESR_OTE)
			printk(KERN_ERR "  OTE: OPB timeout error\n");
		if (mal_error & MALESR_OSE)
			printk(KERN_ERR "  OSE: OPB slave error\n");

		if (mal_error & MALESR_PEIN) {
			plb_error = mfdcr(DCRN_PLB0_BESR);
			printk(KERN_ERR "  PEIN: PLB error, PLB0_BESR is 0x%x\n",
							(unsigned int)plb_error);
			plb_addr = mfdcr(DCRN_PLB0_BEAR);
			printk(KERN_ERR "  PEIN: PLB error, PLB0_BEAR is 0x%x\n",
							(unsigned int)plb_addr);
		}
#endif
	} else {
		/*
		 * Don't know whether it is transmit or receive error, blame
		 * both.
		 */
		lo_priv->ep_stats.rx_errors++;
		lo_priv->ep_stats.tx_errors++;
	}


#ifdef CONFIG_PPC405_ENET_ERROR_MSG

	printk(KERN_ERR "  -----  cumulative errors  -----\n");
	if (mal_error & MALESR_DEI)
	    printk(KERN_ERR "  DEI: descriptor error interrupt\n");
	if (mal_error & MALESR_ONEI)
	    printk(KERN_ERR "  OPB non-fullword error interrupt\n");
	if (mal_error & MALESR_OTEI)
	    printk(KERN_ERR "  OTEI: timeout error interrupt\n");
	if (mal_error & MALESR_OSEI)
	    printk(KERN_ERR "  OSEI: slave error interrupt\n");
	if (mal_error & MALESR_PBEI)
	    printk(KERN_ERR "  PBEI: PLB bus error interrupt\n");
#endif /* CONFIG_PPC405_ENET_ERROR_MSG */



	/* Clear the error status register */
	mtdcr(DCRN_MALESR, mal_error);
}


static void
ppc405_eth_txeob(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device		*dev;
	int				first_time;
	struct ppc405_enet_private	*lo_priv;

	dev        = (struct net_device *)dev_instance;
	lo_priv    = dev->priv;
	first_time = 1;

do_it_again:

	while (lo_priv->ep_xmit_active_count &&
		!(ppc405_enet_tx[ppc405_enet_ack_slot].ctrl & MAL_TX_CTRL_READY)) {

		/* Tell the system the transmit completed. */
		dev_kfree_skb_irq(lo_priv->ep_xmit_skb[ppc405_enet_ack_slot]);

		if (ppc405_enet_tx[ppc405_enet_ack_slot].ctrl &
		    (EMAC_TX_ST_EC | EMAC_TX_ST_MC | EMAC_TX_ST_SC)
		   ) {
			lo_priv->ep_stats.collisions++;
		}

		lo_priv->ep_xmit_skb[ppc405_enet_ack_slot] = (struct sk_buff *)NULL;
		if (++ppc405_enet_ack_slot == NUM_TX_BUFF)
			ppc405_enet_ack_slot = 0;

		lo_priv->ep_xmit_active_count--;

		netif_wake_queue(dev);
	}

	/*
	** Don't stay stuck in this handler forever.
	** The first time through:
	**   Acknowledge the interrupt from the MAL.
	**   If another interrupt has come in, go back and process it.
	**   (Otherwise, return; the interrupt has been cleared in the device.)
	** The second time through:
	**   Don't acknowledge the interrupt from the MAL, just return.
	**   If another interrupt has come in, ignore it.
	**   Didn't acknowledge the interrupt.  That means the UIC interrupt
	**     will be reasserted as soon as it is acknowledged and we'll end
	**     up in this handler again soon (possibly with no new data to
	**     process).  But, in the meantime, other interrupt handlers will
	**     have had a shot at the cpu.
	*/
	if (first_time) {

		/* Clear the interrupt bits. */
		mtdcr(DCRN_MALTXEOBISR, mfdcr(DCRN_MALTXEOBISR));

		/* make sure no interrupt gets lost */
		if (lo_priv->ep_xmit_active_count &&
		    !(ppc405_enet_tx[ppc405_enet_ack_slot].ctrl & MAL_TX_CTRL_READY)
		   ) {
			first_time = 0;
			goto do_it_again;
		}
	}

	return;
}


static void
ppc405_eth_rxeob(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device		*dev;
	int				error;
	int				first_time;
	int				frame_length;
	struct ppc405_enet_private	*lo_priv;
	mal_desc_t			*rx_desc;
	struct sk_buff			*skb_rx;
	struct sk_buff			**skb_rx_p;

	dev          = (struct net_device *)dev_instance;
	lo_priv      = dev->priv;
	first_time   = 1;
	frame_length = 0;

do_it_again:

	rx_desc = &ppc405_enet_rx[ppc405_enet_rx_slot];

	while (!(rx_desc->ctrl & MAL_RX_CTRL_EMPTY)) {

		if (rx_desc->ctrl & EMAC_BAD_RX_PACKET) {
		
			lo_priv->ep_stats.rx_errors++;
			lo_priv->ep_stats.rx_dropped++;

			if (rx_desc->ctrl & EMAC_RX_ST_OE)
				lo_priv->ep_stats.rx_fifo_errors++;
			if (rx_desc->ctrl & EMAC_RX_ST_AE)
				lo_priv->ep_stats.rx_frame_errors++;
			if (rx_desc->ctrl & EMAC_RX_ST_BFCS)
				lo_priv->ep_stats.rx_crc_errors++;
			if (rx_desc->ctrl & (EMAC_RX_ST_RP  |
					     EMAC_RX_ST_PTL |
					     EMAC_RX_ST_ORE |
					     EMAC_RX_ST_IRE  ))
				lo_priv->ep_stats.rx_length_errors++;

		} else {
		
			/* Send the skb up the chain. */

			frame_length     = rx_desc->data_len;

			skb_rx_p = &ppc405_skb_rx[ppc405_enet_rx_slot];
			skb_rx   = *skb_rx_p;

			skb_rx->tail    += frame_length;
			skb_rx->len     += frame_length;
			skb_rx->dev      = dev;
			skb_rx->protocol = eth_type_trans(skb_rx, dev);

			error = netif_rx(skb_rx);

			if ((error == NET_RX_DROP) || (error == NET_RX_BAD))
				lo_priv->ep_stats.rx_dropped++;

			*skb_rx_p = dev_alloc_skb(DESC_BUF_SIZE);
			skb_rx    = *skb_rx_p;
			dma_cache_wback_inv((unsigned long)skb_rx->data, DESC_BUF_SIZE);

			rx_desc->data_ptr = (char *)virt_to_phys(skb_rx->data);

			lo_priv->ep_stats.rx_packets++;
			lo_priv->ep_stats.rx_bytes += frame_length;
		}


		rx_desc->ctrl |= MAL_RX_CTRL_EMPTY;

		if(++ppc405_enet_rx_slot >= NUM_RX_BUFF )
			ppc405_enet_rx_slot = 0;

		rx_desc = &ppc405_enet_rx[ppc405_enet_rx_slot];
	}

	/*
	** Don't stay stuck in this handler forever.
	** The first time through:
	**   Acknowledge the interrupt from the MAL.
	**   If another interrupt has come in, go back and process it.
	**   (Otherwise, return; the interrupt has been cleared in the device.)
	** The second time through:
	**   Don't acknowledge the interrupt from the MAL, just return.
	**   If another interrupt has come in, ignore it.
	**   Didn't acknowledge the interrupt.  That means the UIC interrupt
	**     will be reasserted as soon as it is acknowledged and we'll end
	**     up in this handler again soon (possibly with no new data to
	**     process).  But, in the meantime, other interrupt handlers will
	**     have had a shot at the cpu.
	*/
	if (first_time) {

		/* Ack the interrupt bits */
		mtdcr(DCRN_MALRXEOBISR, mfdcr(DCRN_MALRXEOBISR));

		/* make sure no interrupt gets lost */
		if (!(rx_desc->ctrl & MAL_RX_CTRL_EMPTY)) {
			first_time = 0;
			goto do_it_again;
		}
	}

	return;
}


/*
 * This interrupt should be very rare at best.  This occurs when
 * the hardware has a problem with the transmit descriptor.
 */
static void
ppc405_eth_txde(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *dev;
	struct ppc405_enet_private *lo_priv;
	mal_desc_t tmp_desc[NUM_TX_BUFF];
	int k;
	int loop;


	dev     = (struct net_device *)dev_instance;
	lo_priv = dev->priv;

	printk(KERN_ERR "PPC405 Ethernet driver received a transmit descriptor error\n");

	lo_priv->ep_stats.tx_errors++;

#ifdef CONFIG_PPC405_ENET_ERROR_MSG
	ppc405_eth_emac_dump(KERN_DEBUG);
	ppc405_eth_mal_dump(KERN_DEBUG);
#endif

	/* The manual states that when the interface is stoped and restarted
	 * it reset processing to the first descriptor in the table.
	 * Move descriptor entries to the beginning of the table.
	 */

	lo_priv->ep_xmit_active_count = 0;
	k                             = ppc405_enet_ack_slot;
	ppc405_enet_ack_slot          = 0;
	ppc405_enet_tx_slot           = 0;

	for (loop = 0; loop < NUM_TX_BUFF;loop++) {
		if (ppc405_enet_tx[k].ctrl & MAL_TX_CTRL_READY) {
			lo_priv->ep_xmit_active_count++;
			tmp_desc[loop].ctrl     = ppc405_enet_tx[k].ctrl;
			tmp_desc[loop].data_len = ppc405_enet_tx[k].data_len;
			tmp_desc[loop].data_ptr = ppc405_enet_tx[k].data_ptr;
			if (++k == NUM_TX_BUFF)
				k = 0;
		}
	}

	for (loop = 0; loop < NUM_TX_BUFF;loop++) {
		ppc405_enet_tx[loop].ctrl     = tmp_desc[loop].ctrl;
		ppc405_enet_tx[loop].data_len = tmp_desc[loop].data_len;
		ppc405_enet_tx[loop].data_ptr = tmp_desc[loop].data_ptr;
	}
	

	/* Clear the interrupt */
	/* (could report the channel number here) */
	mtdcr(DCRN_MALTXDEIR, mfdcr(DCRN_MALTXDEIR));


	/* Reenable the transmit channel */
	mtdcr(DCRN_MALTXCASR, 0x80000000);

	return;
}


/*
 * This interrupt should be very rare at best.  This occurs when
 * the hardware has a problem with the receive descriptors.  The manual
 * states that it occurs when the hardware cannot find a receive
 * descriptor with the empty bit set.  The recovery mechanism will be to
 * traverse through the descriptors, handle any that are marked to be
 * handled and reinitialize each along the way.  At that point the driver
 * will be restarted.
 */
static void
ppc405_eth_rxde(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *dev;
	struct ppc405_enet_private *lo_priv;
	int loop;
	int end;
	int frame_length;


	dev     = (struct net_device *)dev_instance;
	lo_priv = dev->priv;

	printk(KERN_ERR "PPC405 Ethernet driver received a receive descriptor error\n");

	lo_priv->ep_stats.rx_errors++;

#ifdef CONFIG_PPC405_ENET_ERROR_MSG
	ppc405_eth_emac_dump(KERN_DEBUG);
	ppc405_eth_mal_dump(KERN_DEBUG);
	ppc405_eth_desc_dump(KERN_DEBUG);
#endif

	loop = ppc405_enet_rx_slot;
	end  = ppc405_enet_rx_slot;

	if (end == -1)
			end = NUM_RX_BUFF - 1;
	
	do {
		/*  If this descriptor not marked empty */
		if (!(ppc405_enet_rx[loop].ctrl & MAL_RX_CTRL_EMPTY)) {
			/* Send the skb up the chain. */
			frame_length = ppc405_enet_rx[loop].data_len;
			ppc405_skb_rx[loop]->tail += frame_length;
			ppc405_skb_rx[loop]->len  += frame_length;

			ppc405_skb_rx[loop]->dev = dev;
			ppc405_skb_rx[loop]->protocol =
					eth_type_trans(ppc405_skb_rx[loop], dev);

			netif_rx(ppc405_skb_rx[loop]);

			/* Allocate the next skb */
			ppc405_skb_rx[loop] = dev_alloc_skb(DESC_BUF_SIZE);
			ppc405_enet_rx[loop].data_ptr =
					(char *)virt_to_phys(ppc405_skb_rx[loop]->data);
			dma_cache_wback_inv((unsigned long)
					ppc405_skb_rx[loop]->data, DESC_BUF_SIZE);
		}

		/* Reset the control bits for every descriptor */
		ppc405_enet_rx[loop].ctrl |= MAL_RX_CTRL_EMPTY;

		if (++loop >= NUM_RX_BUFF)
			loop = 0;
	} while (loop != end);

	/* The manual states that when the interface is stoped and restarted
	 * it resets processing to the first descriptor in the table.
	 */
	ppc405_enet_rx_slot = 0;

	/* Reenable the receive channel */
	mtdcr(DCRN_MALRXCASR, 0x80000000);

	/* Clear the interrupt */
	mtdcr(DCRN_MALRXDEIR, mfdcr(DCRN_MALRXDEIR));

}


static void
ppc405_eth_mac(int irq, void *dev_instance, struct pt_regs *regs)
{
	unsigned long tmp_em0isr;
	struct net_device *dev;
	struct ppc405_enet_private *lo_priv;


	/* EMAC interrupt */

	dev     = (struct net_device *)dev_instance;
	lo_priv = dev->priv;

	tmp_em0isr  = ppc405_enet_regp->em0isr;

	if (tmp_em0isr & (EMAC_ISR_TE0 | EMAC_ISR_TE1)) {
		/* This error is a hard transmit error - could retransmit */
		lo_priv->ep_stats.tx_errors++;

		/* Reenable the transmit channel */
		mtdcr(DCRN_MALTXCASR, 0x80000000);

	} else {		
		lo_priv->ep_stats.rx_errors++;
	}


/*	if (tmp_em0isr & EMAC_ISR_OVR ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_PP  ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_BP  ) lo_priv->ep_stats.ZZZ++;		*/
	if (tmp_em0isr & EMAC_ISR_RP  ) lo_priv->ep_stats.rx_length_errors++;
/*	if (tmp_em0isr & EMAC_ISR_SE  ) lo_priv->ep_stats.ZZZ++;		*/
	if (tmp_em0isr & EMAC_ISR_ALE ) lo_priv->ep_stats.rx_frame_errors++;
	if (tmp_em0isr & EMAC_ISR_BFCS) lo_priv->ep_stats.rx_crc_errors++;
	if (tmp_em0isr & EMAC_ISR_PTLE) lo_priv->ep_stats.rx_length_errors++;
	if (tmp_em0isr & EMAC_ISR_ORE ) lo_priv->ep_stats.rx_length_errors++;
/*	if (tmp_em0isr & EMAC_ISR_IRE ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_DBDM) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_DB0 ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_SE0 ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_TE0 ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_DB1 ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_SE1 ) lo_priv->ep_stats.ZZZ++;		*/
	if (tmp_em0isr & EMAC_ISR_TE1 ) lo_priv->ep_stats.tx_aborted_errors++;
/*	if (tmp_em0isr & EMAC_ISR_MOS ) lo_priv->ep_stats.ZZZ++;		*/
/*	if (tmp_em0isr & EMAC_ISR_MOF ) lo_priv->ep_stats.ZZZ++;		*/


#ifdef CONFIG_PPC405_ENET_ERROR_MSG
	printk(KERN_ERR "%s: on-chip ethernet error:\n", dev->name);

	if (tmp_em0isr & EMAC_ISR_OVR)
	    printk(KERN_ERR "  OVR: overrun\n");
	if (tmp_em0isr & EMAC_ISR_PP)
	    printk(KERN_ERR "  PP: control pause packet\n");
	if (tmp_em0isr & EMAC_ISR_BP)
	    printk(KERN_ERR "  BP: packet error\n");
	if (tmp_em0isr & EMAC_ISR_RP)
	    printk(KERN_ERR "  RP: runt packet\n");
	if (tmp_em0isr & EMAC_ISR_SE)
	    printk(KERN_ERR "  SE: short event\n");
	if (tmp_em0isr & EMAC_ISR_ALE)
	    printk(KERN_ERR "  ALE: odd number of nibbles in packet\n");
	if (tmp_em0isr & EMAC_ISR_BFCS)
	    printk(KERN_ERR "  BFCS: bad FCS\n");
	if (tmp_em0isr & EMAC_ISR_PTLE)
	    printk(KERN_ERR "  PTLE: oversized packet\n");
	if (tmp_em0isr & EMAC_ISR_ORE)
	    printk(KERN_ERR "  ORE: packet length field > max allowed LLC\n");
	if (tmp_em0isr & EMAC_ISR_IRE)
	    printk(KERN_ERR "  IRE: In Range error\n");
	if (tmp_em0isr & EMAC_ISR_DBDM)
	    printk(KERN_ERR "  DBDM: xmit error or SQE\n");
	if (tmp_em0isr & EMAC_ISR_DB0)
	    printk(KERN_ERR "  DB0: xmit error or SQE on TX channel 0\n");
	if (tmp_em0isr & EMAC_ISR_SE0)
	    printk(KERN_ERR "  SE0: Signal Quality Error test failure from TX channel 0\n");
	if (tmp_em0isr & EMAC_ISR_TE0)
	    printk(KERN_ERR "  TE0: xmit channel 0 aborted\n");
	if (tmp_em0isr & EMAC_ISR_DB1)
	    printk(KERN_ERR "  DB1: xmit error or SQE on TX channel \n");
	if (tmp_em0isr & EMAC_ISR_SE1)
	    printk(KERN_ERR "  SE1: Signal Quality Error test failure from TX channel 1\n");
	if (tmp_em0isr & EMAC_ISR_TE1)
	    printk(KERN_ERR "  TE1: xmit channel 1 aborted\n");
	if (tmp_em0isr & EMAC_ISR_MOS)
	    printk(KERN_ERR "  MOS\n");
	if (tmp_em0isr & EMAC_ISR_MOF)
	    printk(KERN_ERR "  MOF\n");


	ppc405_eth_emac_dump(KERN_DEBUG);
	ppc405_eth_mal_dump(KERN_DEBUG);
#endif

	ppc405_enet_regp->em0isr = tmp_em0isr;
	eieio();
}


#ifdef CONFIG_PPC405_ENET_ERROR_MSG
static void
ppc405_eth_desc_dump(char *log_level)
{
	int curr_slot;

	printk("%s\ndumping the receive descriptors:  current slot is %d\n",
		log_level, ppc405_enet_rx_slot);
	for (curr_slot = 0; curr_slot < NUM_RX_BUFF; curr_slot++) {
		printk("%sDesc %02d: status 0x%04x, length %3d, addr 0x%x\n", 
			log_level,
			curr_slot,
			ppc405_enet_rx[curr_slot].ctrl,
			ppc405_enet_rx[curr_slot].data_len,
			(unsigned int)ppc405_enet_rx[curr_slot].data_ptr);
	}
}


static void
ppc405_eth_emac_dump(char *log_level)
{
	printk("%sEMAC DEBUG ********** \n", log_level);
	printk("%sEMAC_M0  ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0mr0);
	eieio();
	printk("%sEMAC_M1  ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0mr1);
	eieio();
	printk("%sEMAC_TXM0==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0tmr0);
	eieio();
	printk("%sEMAC_TXM1==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0tmr1);
	eieio();
	printk("%sEMAC_RXM ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0rmr);
	eieio();
	printk("%sEMAC_ISR ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0isr);
	eieio();
	printk("%sEMAC_IER ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0iser);
	eieio();
	printk("%sEMAC_IAH ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0iahr);
	eieio();
	printk("%sEMAC_IAL ==> 0x%x\n", log_level, (unsigned int)ppc405_enet_regp->em0ialr);
	eieio();
	printk("%sEMAC_VLAN_TPID_REG ==> 0x%x\n",
				        log_level, (unsigned int)ppc405_enet_regp->em0vtpid);
	eieio();
	printk("%s\n", log_level);
}


static void
ppc405_eth_mal_dump(char *log_level)
{
	printk("%s MAL DEBUG ********** \n", log_level);
	printk("%s MCR      ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALCR));
	printk("%s ESR      ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALESR));
	printk("%s IER      ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALIER));
	printk("%s DBR      ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALDBR));
	printk("%s TXCASR   ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALTXCASR));
	printk("%s TXCARR   ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALTXCARR));
	printk("%s TXEOBISR ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALTXEOBISR));
	printk("%s TXDEIR   ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALTXDEIR));
	printk("%s RXCASR   ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALRXCASR));
	printk("%s RXCARR   ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALRXCARR));
	printk("%s RXEOBISR ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALRXEOBISR));
	printk("%s RXDEIR   ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALRXDEIR));
	printk("%s TXCTP0R  ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALTXCTP0R));
	printk("%s TXCTP1R  ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALTXCTP1R));
	printk("%s RXCTP0R  ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALRXCTP0R));
	printk("%s RCBS0    ==> 0x%x\n", log_level, (unsigned int)mfdcr(DCRN_MALRCBS0));
	printk("%s\n", log_level);
}
#endif /* CONFIG_PPC405_ENET_ERROR_MSG */


/*
 *  Finds the CRC32 of a set of bytes.
 *    Again, from Peter Cammaert's code.
 *
 *    This code is stolen from another driver.
 */
static unsigned long
crc32(char * s, int length)
{
	/* indices */
	int perByte;
	int perBit;
	int carry;

	/* crc polynomial for Ethernet */
	const unsigned long poly = 0xedb88320;
	/* crc value - preinitialized to all 1's */
	unsigned long crc_value = 0xffffffff;

	for ( perByte = 0; perByte < length; perByte ++ ) {
		unsigned char   c;

		c = *(s++);

		for ( perBit = 0; perBit < 8; perBit++ ) {
			if (crc_value & 0x80000000)
				carry = 0x1 ^ (c & 0x1);
			else
				carry = 0x0 ^ (c & 0x1);

			crc_value <<= 1;

			if (carry)
				crc_value = (crc_value ^ poly) | carry;

			c >>= 1;
		}
	}

	return  crc_value;
}


/*
 * Local variables:
 *  c-indent-level: 8
 *  c-basic-offset: 8
 *  tab-width: 8
 * End:
 */
