/*
COPYRIGHT (C) 2001  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#include <linux/module.h>

#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include <rtai.h>
#include <rtai_shm.h>

#include "kurtnet.h"

static DECLARE_MUTEX_LOCKED(mtx);
static spinlock_t sysrq_lock = SPIN_LOCK_UNLOCKED;

static struct sock_t *runsock, *socks;

int rt_socket(int domain, int type, int protocol)
{
	int i;
	for (i = 0; i < (MAX_SOCKS + MAX_STUBS); i++) {
		if (!socks[i].opnd) {
			socks[i].opnd = 1;
			return i;
		}
	}
	return -1;
}

int rt_close(int sock)
{
	if (sock >= 0 && sock < (MAX_SOCKS + MAX_STUBS)) {
		return socks[sock].opnd = 0;
	}
	return -1;
}

int rt_bind(int sock, struct sockaddr *addr, int addrlen)
{
	memcpy(&socks[sock].bindaddr, addr, addrlen);
	return 0;
}

int rt_socket_callback(int sock, int (*func)(int sock, void *arg), void *arg)
{
	if (func > 0) {
		socks[sock].callback = func;
		socks[sock].arg      = arg;
		return 0;
	}
	return -1;
}

#define MAX_SOCK_SRQ 128
static struct { int srq, in, out, sockindx[MAX_SOCK_SRQ]; } sysrq;

int rt_sendto(int sock, const void *msg, int len, unsigned int sflags, struct sockaddr *to, int tolen)
{
	unsigned long flags;

	memcpy(socks[sock].msg, msg, len);
	memcpy(&socks[sock].addr, to, tolen);
	socks[sock].len = len < MAX_MSG_SIZE ? len : MAX_MSG_SIZE;
	flags = rt_spin_lock_irqsave(&sysrq_lock);
	sysrq.sockindx[sysrq.in] = sock;
        sysrq.in = (sysrq.in + 1) & (MAX_SOCK_SRQ - 1);
	rt_pend_linux_srq(sysrq.srq);
	rt_spin_unlock_irqrestore(flags, &sysrq_lock);
	return 0;
}

int rt_recvfrom(int sock, void *msg, int len, unsigned int flags, struct sockaddr *from, int *fromlen)
{
	memcpy(msg, socks[sock].msg, socks[sock].recv);
	if (from && fromlen) { 
		memcpy(from, &socks[sock].addr, socks[sock].addrlen);
		*fromlen = socks[sock].addrlen;
	}
	return socks[sock].recv;
}

static int urtnet_pid;

long long user_rtnet(unsigned int whatever)
{
	unsigned long flags;
	unsigned int arg[2];
        int i;
        sigset_t signal, blocked;

        copy_from_user(arg, (unsigned int *)whatever, 2*sizeof(unsigned int));
	if (arg[0] == SENDTO) {
		if (!urtnet_pid) {
			urtnet_pid = current->pid;
		}
		if (sysrq.out == sysrq.in) {
			down_interruptible(&mtx);
        		signal = current->pending.signal;
        		blocked = current->blocked;
		        for (i = 0; i < _NSIG_WORDS; i++) {
		                if (signal.sig[i] & ~blocked.sig[i]) {
                			return -ERESTARTSYS;
		                }
			}		
		}
		flags = rt_spin_lock_irqsave(&sysrq_lock);
		if (sysrq.out != sysrq.in) {
			arg[0] = sysrq.sockindx[sysrq.out];
               		sysrq.out = (sysrq.out + 1) & (MAX_SOCK_SRQ - 1);
		} else {
			arg[0] = -1;
		}
		rt_spin_unlock_irqrestore(flags, &sysrq_lock);
		return arg[0];
	} else if (arg[0] == RECVFROM) {
		if (socks[arg[1]].callback) {
			socks[arg[1]].callback(arg[1], socks[arg[1]].arg);
		}
		return 0;
	}
	return 0;
}

void rtai_rtnet(void)
{
	up(&mtx);
	current->need_resched = 1;
}

int init_krtnet(void)
{
	rt_mount_rtai();
        runsock = (struct sock_t *)rtai_kmalloc(0xcacca0, (MAX_SOCKS + MAX_STUBS + 1)*sizeof(struct sock_t));
        socks = runsock + 1;
	if ((sysrq.srq = rt_request_srq(0xcacca1, rtai_rtnet, user_rtnet)) < 0) {
                printk("KRTNET: no sysrq available.\n");
                return sysrq.srq;
	}
	return 0;
}

void cleanup_krtnet(void)
{
	runsock->opnd = 0;
	up(&mtx);
	while (urtnet_pid && find_task_by_pid(urtnet_pid)) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/20);
	}
	rt_free_srq(sysrq.srq);
        rtai_kfree(0xcacca0);
	rt_umount_rtai();
	return;
}
#include <linux/init.h> /* this is required on PowerPC */
module_init(init_krtnet);
module_exit(cleanup_krtnet);
