/*
 * drivers/char/serial_ep7211.c
 *
 * Copyright (C) 2000 Blue Mug, Inc.  All Rights Reserved.
 * Based on serial_7211.c from Ben Williamson's 2.2.1 7211 port,
 * Based on serial_7110.c from the Psion/Geofox port.
 *
 * Devices supported:
 *
 * (204, 4) is UART1
 * (204, 5) is UART2 *untested*
 * (204, 6) is UART1 in IR mode *untested*
 *
 * Questionable things:
 * -- Device name in this file is "ttyFB", which is straight from
 *    serial_21285.c but which seems icky.
 * -- I haven't even looked to see what major 204 is; it's also
 *    pulled straight from serial_21285.c.
 * -- The serial console code has been updated for 2.4, and a few
 *    bugs have been fixed, but in general the driver hasn't been
 *    checked over for 2.4 driver correctness.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/major.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
//#include <linux/malloc.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/serial.h>
#include <linux/console.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include <linux/serial_core.h>
 
#include <asm/hardware/ep7211.h>	/* XXX should be EP7211 */
#include <asm/hardware/clps7111.h>	/* XXX should be EP7211 */

#define NR_PORTS 2
#define LINE_IR_BIT 0x02	/* in the minor number */
#define CONSOLE_PORT 0
#define WBUF_SIZE 1000
#define BAUD_BASE 230400
#define RS7211_BH DIGI_BH	/* not very likely to find any of these in a PDA */
#define WAKEUP_CHARS 256

#define IRQ_NONE	0xffffffff

#if 1
/* We want to stay compatible with the legacy /dev/ttyS0,1,2 devices for now */
/* 
 * For the regular 16x50 serial driver to work with this, we defined
 * SERIAL_DEV_OFFSET to 3 in serial.c.
 */
#define  SERIAL_EP7211_MAJOR	TTY_MAJOR
#define CALLOUT_EP7211_MAJOR	TTYAUX_MAJOR
#define MINOR_START		64
#define TTY_NAME		"ttyS"
#define  CU_NAME		"cua"
#else
/* We've been assigned a range on the "Low-density serial ports" major */
#define  SERIAL_EP7211_MAJOR	204
#define CALLOUT_EP7211_MAJOR	205
#define MINOR_START		16
#define TTY_NAME		"ttyFB"
#define  CU_NAME		"cusa"
#endif

/* Baud rates */
#define BR_115200    1
#define BR_57600     3
#define BR_38400     5
#define BR_19200     11
#define BR_9600      23
#define BR_2400      95
#define BR_1200      191


static struct tty_driver rs7211_driver, callout_driver;
static int rs7211_refcount;
static struct tty_struct *rs7211_table [NR_PORTS];
static struct termios *rs7211_termios [NR_PORTS];
static struct termios *rs7211_termios_locked [NR_PORTS];

struct driver_data {
	/* constant: */
	int line;
	unsigned long		syscon;
	unsigned long 		ubrlcr;
	unsigned long 		sysflg;
	unsigned long 		uartdr;
	unsigned long		uarten;
	unsigned long		urxfe;
	unsigned long		utxff;
	unsigned long		ubusy;
	unsigned long		irq_urxint;
	unsigned long		irq_utxint;
	unsigned long		irq_umsint;

	/* initialised: */
	struct tty_struct *rs7211_tty;
	int use_count, ir_flag;
	struct async_icount icount;

	/* not initialised: */
	char *putp, *getp, x_char;
	char wbuf [WBUF_SIZE];
	wait_queue_head_t	delta_msr_wait;
};

static struct driver_data driver_data[NR_PORTS] = {
	{
		0,
		SYSCON1, UBRLCR1, SYSFLG1, UARTDR1,
		SYSCON_UARTEN, SYSFLG_URXFE, SYSFLG_UTXFF, SYSFLG_UBUSY, 
		IRQ_URXINT1, IRQ_UTXINT1, IRQ_UMSINT, NULL, 0, 0
	},
	{
		1,
		SYSCON2, UBRLCR2, SYSFLG2, UARTDR2,
		SYSCON_UARTEN, SYSFLG_URXFE, SYSFLG_UTXFF, SYSFLG_UBUSY, 
		IRQ_URXINT2, IRQ_UTXINT2, IRQ_NONE, NULL, 0, 0
	}
};

#define SYSCON	(info->syscon)
#define SYSFLG	(info->sysflg)
#define UARTDR	(info->uartdr)
#define UBRLCR	(info->ubrlcr)

static int rs7211_write_room(struct tty_struct *tty)
{
	struct driver_data *info = tty->driver_data;
	return info->putp >= info->getp ?
		(WBUF_SIZE-(long)(info->putp)+(long)(info->getp)):
		((long)(info->getp)-(long)(info->putp)-1);
}

static void rs7211_rx_int(int irq, void *dev_id, struct pt_regs *regs)
{ 
	struct driver_data *info = dev_id;
	if (!info->rs7211_tty) {
		disable_irq(info->irq_urxint);
		return;
	}

	info->icount.rx++;

	/* if (info->rs7211_tty->ldisc.receive_room(info->rs7211_tty))
	 * pm_activity(&rs7211_pm_dev); */
	while( !(clps_readl(SYSFLG) & info->urxfe)) {
		char flag = 0;
		unsigned long ch;

		ch = clps_readl(UARTDR); 

		/*
		printk( "<1>read: %02x, %03d, %c, %c, %c\n", ch&0xff, ch&0xff,
			(ch & OVERR) ? 'o' : '-',
			(ch & PARERR) ? 'p' : '-',
			(ch & FRMERR) ? 'f' : '-' );
		*/

		if ( ch & UARTDR_OVERR )
		   {
		     // clear overrun
		     unsigned long c = clps_readl(UARTDR); 

		     // clear fifo
		     while( !(clps_readl(SYSFLG) & info->urxfe))
			  c = clps_readl(UARTDR);

		     // nur wenn kein frame error
		     if ( (ch & UARTDR_FRMERR) == 0 )
			  tty_insert_flip_char(info->rs7211_tty,0,TTY_OVERRUN);
		   }

		if (ch & UARTDR_PARERR)
			flag=TTY_PARITY;
		else if (ch & UARTDR_FRMERR)
			flag=TTY_FRAME;
		tty_insert_flip_char(info->rs7211_tty,ch & 0xff,flag);
	}
	tty_flip_buffer_push(info->rs7211_tty);
}

static void rs7211_send_xchar(struct tty_struct *tty, char ch)
{
	struct driver_data *info = tty->driver_data;
	info->x_char=ch;
	enable_irq(info->irq_utxint);
}

static void rs7211_throttle(struct tty_struct *tty)
{
	if(I_IXOFF(tty))
		rs7211_send_xchar(tty,STOP_CHAR(tty));
}

static void rs7211_unthrottle(struct tty_struct *tty)
{
	struct driver_data *info = tty->driver_data;
	if(I_IXOFF(tty)) {
		if(info->x_char)
			info->x_char=0;
		else
			rs7211_send_xchar(tty,START_CHAR(tty));
	}
}

static void do_rs7211_bh(void)
{
	int line;
	for (line = 0; line < NR_PORTS; line++) {
		struct driver_data *info = &(driver_data[line]);

		if (!info->use_count)
			continue;
		/* this bit me because of serial console */
		if (!info->rs7211_tty)
			panic("rs7211: BH called on closed tty\n");
		if ((info->rs7211_tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    info->rs7211_tty->ldisc.write_wakeup)
			info->rs7211_tty->ldisc.write_wakeup(info->rs7211_tty);
		wake_up_interruptible(&info->rs7211_tty->write_wait);
	}
}

static void rs7211_tx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct driver_data *info = dev_id;

	info->icount.tx++;

	while(!(clps_readl(SYSFLG) & info->utxff)) {
		if(info->x_char) {
			clps_writel(info->x_char, UARTDR);
			info->x_char=0;
			continue;
		}
		if(info->putp==info->getp) {
			disable_irq(info->irq_utxint);
			break;
		}
		clps_writel(*(info->getp), UARTDR);
		if(++(info->getp)>=info->wbuf+WBUF_SIZE)
			info->getp=info->wbuf;
	}
	if (info->rs7211_tty && rs7211_write_room(info->rs7211_tty) > WBUF_SIZE-WAKEUP_CHARS)
		mark_bh(RS7211_BH);
}

static void rs7211_ms_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct driver_data *info = dev_id;

	if (clps_readl(SYSFLG) & SYSFLG1_DCD)
		info->icount.dcd++;
	if (clps_readl(SYSFLG) & SYSFLG1_DSR)
		info->icount.dsr++;
	if (clps_readl(SYSFLG) & SYSFLG1_CTS) 
		info->icount.cts++;
#ifdef CONFIG_ARCH_EDB7211
	if (clps_readb(PBDR) & 0x4)
		info->icount.rng++;
#endif
	
	wake_up_interruptible(&info->delta_msr_wait);

	clps_writel(0, UMSEOI);
}

static inline int rs7211_xmit(struct tty_struct *tty, int ch)
{
	struct driver_data *info = tty->driver_data;
	if(info->putp+1==info->getp || (info->putp+1==info->wbuf+WBUF_SIZE && info->getp==info->wbuf))
		return 0;
	*(info->putp)=ch;
	if(++(info->putp)>=info->wbuf+WBUF_SIZE)
		info->putp=info->wbuf; 
	enable_irq(info->irq_utxint);
	return 1;
}

static int rs7211_write(struct tty_struct *tty,int from_user,
	const u_char *buf,int count)
{ 
	int i;
	if(from_user) {
		if(verify_area(VERIFY_READ, buf, count))
			return -EINVAL;
	}
	for(i=0;i<count;i++) {
		char ch;
		if(from_user)
			get_user(ch,buf+i);
		else
			ch=buf[i];
		if(!rs7211_xmit(tty, ch))
			break;
	}
	return i;
} 

static void rs7211_put_char(struct tty_struct *tty,u_char ch)
{
	rs7211_xmit(tty, ch);
}

static int rs7211_chars_in_buffer(struct tty_struct *tty)
{
	return WBUF_SIZE-rs7211_write_room(tty);
}

static void rs7211_flush_buffer(struct tty_struct *tty)
{
	struct driver_data *info = tty->driver_data;
	disable_irq(info->irq_utxint);
	info->putp=info->getp=info->wbuf;
	if(info->x_char)
		enable_irq(info->irq_utxint);
	mark_bh(RS7211_BH);
}

static inline void rs7211_set_cflag(struct driver_data *info, int cflag)
{
	int cr;

	switch(cflag & CSIZE) {
	case CS5: cr=UBRLCR_WRDLEN5; break;
	case CS6: cr=UBRLCR_WRDLEN6; break;
	case CS7: cr=UBRLCR_WRDLEN7; break;
	default: cr=UBRLCR_WRDLEN8; break;	// CS8
	}

	if (cflag & CSTOPB) cr|=UBRLCR_XSTOP;
	if (cflag & PARENB) cr|=UBRLCR_PRTEN;
	if (!(cflag & PARODD)) cr|=UBRLCR_EVENPRT;

	switch(cflag & CBAUD) {
	case B1200: cr |= BR_1200; break;
	case B2400: cr |= BR_2400; break;
	case B19200: cr |= BR_19200; break;
	case B38400: cr |= BR_38400; break;
	case B57600: cr |= BR_57600; break;
	case B115200: cr |= BR_115200; break;
	default: cr |= BR_9600;
	}
	clps_writel(cr, info->ubrlcr);
}

static void rs7211_set_termios(struct tty_struct *tty, struct termios *old)
{
	struct driver_data *info = tty->driver_data;
	if(old && tty->termios->c_cflag == old->c_cflag)
		return;
	rs7211_set_cflag(info, tty->termios->c_cflag);
}


static void rs7211_stop(struct tty_struct *tty)
{
	struct driver_data *info = tty->driver_data;
	disable_irq(info->irq_utxint);
}

static void rs7211_start(struct tty_struct *tty)
{
	struct driver_data *info = tty->driver_data;
	enable_irq(info->irq_utxint);
}

static void rs7211_wait_until_sent(struct tty_struct *tty,int timeout)
{
	struct driver_data *info = tty->driver_data;
	int orig_jiffies=jiffies;
	while (clps_readl(SYSFLG) & info->ubusy) {
		current->state=TASK_INTERRUPTIBLE;
		schedule_timeout(1);
		if(signal_pending(current))
			break;
		if(timeout && ((orig_jiffies + timeout)<jiffies))
			break;
	}
	current->state=TASK_RUNNING;
}

static int rs7211_open(struct tty_struct *tty, struct file *filp)
{
	struct driver_data *info;
	int line, ir_flag;

	line = MINOR(tty->device) - tty->driver.minor_start;
	ir_flag = line & LINE_IR_BIT;
	line = line & ~LINE_IR_BIT;
	if ((line < 0) || (line >= NR_PORTS)) return -ENODEV;
	info = &(driver_data[line]);
	if ((info->use_count) && info->ir_flag != ir_flag) return -EBUSY;
	info->ir_flag = ir_flag;
	/* if (!pm_change_wait(&rs7211_pm_dev,1)) return -EIO; */
	/* pm_switch_on_state(&rs7211_pm_dev,1); */
	tty->driver_data = info;
	info->rs7211_tty = tty;
	info->use_count++;
	enable_irq(info->irq_urxint);
	return 0;
}

static void rs7211_close(struct tty_struct *tty, struct file *filp)
{
	struct driver_data *info = tty->driver_data;
	if (tty != info->rs7211_tty || --info->use_count)
		return;
	rs7211_wait_until_sent(tty, 0);
	disable_irq(info->irq_urxint);
	disable_irq(info->irq_utxint);
	info->rs7211_tty = NULL;
}

/* Helper functions for rs7211_ioctl() */
static int get_serial_info(struct driver_data * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
   
	if (!retinfo)
		return -EFAULT;

	memset(&tmp, 0, sizeof(tmp));
	tmp.type = PORT_16550;
	tmp.line = info->line;
	tmp.irq = info->irq_urxint;
	tmp.xmit_fifo_size = (clps_readl(UBRLCR) & UBRLCR_FIFOEN) ? 16 : 1;
	tmp.baud_base = BAUD_BASE;
	tmp.io_type = SERIAL_IO_PORT;
	tmp.custom_divisor = clps_readl(UBRLCR) & UBRLCR_BAUD_MASK;

	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

/* XXX: Fix this to us the proper set stuff... */
static int set_serial_info(struct driver_data * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
	unsigned int		change_irq;
	int 			retval = 0;

	if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
		return -EFAULT;

	change_irq = new_serial.irq != info->irq_urxint;
  
	if (!capable(CAP_SYS_ADMIN)) {
		if (change_irq ||
		    (new_serial.baud_base != BAUD_BASE) ||
		    (new_serial.type != PORT_16550) ||
		    (new_serial.xmit_fifo_size != 
			((clps_readl(UBRLCR) & UBRLCR_FIFOEN)?16:1)))
			return -EPERM;
		goto check_and_exit;
	}

	if (change_irq) {
		return -EINVAL;
	}

	if (new_serial.baud_base != BAUD_BASE) {
		return -EINVAL;
	}

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	clps_writel((clps_readl(UBRLCR) & ~UBRLCR_BAUD_MASK) | 
		new_serial.custom_divisor, UBRLCR);

	if (new_serial.xmit_fifo_size > 1) {
		clps_writel(clps_readl(UBRLCR) | UBRLCR_FIFOEN, UBRLCR);
	} else {
		clps_writel(clps_readl(UBRLCR) & ~UBRLCR_FIFOEN, UBRLCR);

	}

	
check_and_exit:
	return retval;
}

/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space. 
 *
 * XXX: The UARTs on the EP7211 don't let you know when the TX buffer is
 *      empty; they only tell you if it's full. In an attempt to get this
 *      half-working, I'm having this set the TIOCSER_TEMT flag when the
 *      TX buffer is not full. Again, this is not the same as setting it
 *      when the TX buffer is empty.
 */
static int get_lsr_info(struct driver_data * info, unsigned int *value)
{
	unsigned char status;
	unsigned int result;
	unsigned long flags;

	save_flags(flags); cli();
	status = clps_readl(SYSFLG);
	restore_flags(flags);

	/* 
	 * This works because the tx buffer full bit is in the same place
	 * for both UARTs. 
	 */
	result = ((status & SYSFLG_UTXFF) ? 0 : TIOCSER_TEMT);

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}


static int get_modem_info(struct driver_data * info, unsigned int *value)
{
	unsigned char control, status;
	unsigned int result;
	unsigned long flags;

	/* If it's not UART1, return an empty structure. */
	if (info->line != 0) {
		result = 0;
		copy_to_user(value, &result, sizeof(int));
		return 0;
	}
	
	save_flags(flags); cli();
#ifdef CONFIG_ARCH_EDB7211
	control = clps_readb(PBDR);
#endif
	status = clps_readl(SYSFLG1);
	restore_flags(flags);

	result =  ((control & 0x2) ? TIOCM_RTS : 0)
		| ((control & 0x4) ? TIOCM_RNG : 0)
		| ((1) ? TIOCM_DTR : 0) /* This is jumpered on the EDB7211 */
		| ((status  & SYSFLG1_DCD) ? TIOCM_CAR : 0)
		| ((status  & SYSFLG1_DSR) ? TIOCM_DSR : 0)
		| ((status  & SYSFLG1_CTS) ? TIOCM_CTS : 0);

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}

static int set_modem_info(struct driver_data * info, unsigned int cmd,
			  unsigned int *value)
{
	unsigned int arg;
	unsigned long flags;

	/* If it's not UART1, do nothing. */
	if (info->line != 0) {
		return 0;
	}

	if (copy_from_user(&arg, value, sizeof(int)))
		return -EFAULT;

	if (cmd != TIOCMSET) {
		return -EINVAL;
	}

#ifdef CONFIG_ARCH_EDB7211
	save_flags(flags); cli();
	clps_writeb((clps_readb(PBDR) & ~0x6) |
		    ((arg & TIOCM_RTS) ? 0x2 : 0) |
		    ((arg & TIOCM_RNG) ? 0x4 : 0), 
		    PBDR);
	restore_flags(flags);
#endif
	
	return 0;
}

/*
 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
 * - mask passed in arg for lines of interest
 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
 * Caller should use TIOCGICOUNT to see which one it was
 */
static int modem_info_wait(struct driver_data * info, unsigned int arg)
{
	struct async_icount cprev, cnow;	/* kernel counter temps */
	unsigned long flags;

	if (info->irq_umsint == IRQ_NONE) {
		return -EINVAL;
	}

	save_flags(flags); cli();
	/* note the counters on entry */
	cprev = info->icount;
	restore_flags(flags);
	enable_irq(info->irq_umsint);
	while (1) {
		interruptible_sleep_on(&info->delta_msr_wait);
		/* see if a signal did it */
		if (signal_pending(current))
			return -ERESTARTSYS;

		save_flags(flags); cli();
		cnow = info->icount; /* atomic copy */
		restore_flags(flags);
		if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr && 
				cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
			return -EIO; /* no change => error */
		if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
			return 0;
		}
		cprev = cnow;
	}

	return 0;
}

static int rs7211_ioctl(struct tty_struct *tty, struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct driver_data * info = (struct driver_data *)tty->driver_data;
	struct async_icount cnow;	/* kernel counter temps */
	struct serial_icounter_struct icount;
	unsigned long flags;

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGSTRUCT) &&
	    (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT) &&
	    (cmd != TIOCMGET) && (cmd != TIOCMSET)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
		    return -EIO;
	}
	
	switch (cmd) {
		case TIOCMGET:
			return get_modem_info(info, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			return set_modem_info(info, cmd, (unsigned int *) arg);
		case TIOCGSERIAL:
			return get_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSSERIAL:
			return set_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSERCONFIG:
			return -ENOIOCTLCMD;

		case TIOCSERGETLSR: /* Get line status register */
			return get_lsr_info(info, (unsigned int *) arg);

		case TIOCSERGSTRUCT:
			return -ENOIOCTLCMD;

		case TIOCMIWAIT:
			return modem_info_wait(info, arg);

		/* 
		 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
		 * Return: write counters to the user passed counter struct
		 * NB: both 1->0 and 0->1 transitions are counted except for
		 *     RI where only 0->1 is counted.
		 */
		case TIOCGICOUNT:
			save_flags(flags); cli();
			cnow = info->icount;
			restore_flags(flags);
			icount.cts = cnow.cts;
			icount.dsr = cnow.dsr;
			icount.rng = cnow.rng;
			icount.dcd = cnow.dcd;
			icount.rx = cnow.rx;
			icount.tx = cnow.tx;
			icount.frame = cnow.frame;
			icount.overrun = cnow.overrun;
			icount.parity = cnow.parity;
			icount.brk = cnow.brk;
			icount.buf_overrun = cnow.buf_overrun;
			
			if (copy_to_user((void *)arg, &icount, sizeof(icount)))
				return -EFAULT;
			return 0;

		case TIOCSERGWILD:
		case TIOCSERSWILD:
			/* "setserial -W" is called in Debian boot */
			printk ("TIOCSER?WILD ioctl obsolete, ignored.\n");
			return 0;

		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

static int __init rs7211_init(void)
{
	int line, ret;

	init_bh(RS7211_BH, do_rs7211_bh);

	memset(&rs7211_driver, 0, sizeof(rs7211_driver));
	rs7211_driver.magic = TTY_DRIVER_MAGIC;
	rs7211_driver.driver_name = "serial_ep7211";
#if (defined(CONFIG_DEVFS_FS))
	rs7211_driver.name = "tts/%d";
#else
	rs7211_driver.name = TTY_NAME;
#endif 
	rs7211_driver.major = SERIAL_EP7211_MAJOR;
	rs7211_driver.minor_start = MINOR_START;
	rs7211_driver.num = 3;
	rs7211_driver.type = TTY_DRIVER_TYPE_SERIAL;
	rs7211_driver.subtype = SERIAL_TYPE_NORMAL;
	rs7211_driver.init_termios = tty_std_termios;
	rs7211_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	rs7211_driver.flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
	rs7211_driver.refcount = &rs7211_refcount; 
	rs7211_driver.table = rs7211_table;
	rs7211_driver.termios = rs7211_termios;
	rs7211_driver.termios_locked = rs7211_termios_locked;

	rs7211_driver.open = rs7211_open;
	rs7211_driver.close = rs7211_close;
	rs7211_driver.write = rs7211_write;
	rs7211_driver.ioctl = rs7211_ioctl;
	rs7211_driver.put_char = rs7211_put_char;
	rs7211_driver.write_room = rs7211_write_room;
	rs7211_driver.chars_in_buffer = rs7211_chars_in_buffer;
	rs7211_driver.flush_buffer = rs7211_flush_buffer;
	rs7211_driver.throttle = rs7211_throttle;
	rs7211_driver.unthrottle = rs7211_unthrottle;
	rs7211_driver.send_xchar = rs7211_send_xchar;
	rs7211_driver.set_termios = rs7211_set_termios;
	rs7211_driver.stop = rs7211_stop;
	rs7211_driver.start = rs7211_start;
	rs7211_driver.wait_until_sent = rs7211_wait_until_sent; 

	for (line = 0; line < NR_PORTS; line++) {
		struct driver_data *info = &(driver_data[line]);

#ifdef CONFIG_SERIAL_EP7211_CONSOLE
		/* 
		 * don't setup serial console again, it might run on a 
		 * different baudrate 
		 */
		if (line != 0)
#endif
		rs7211_set_cflag(info, rs7211_driver.init_termios.c_cflag);
		info->putp = info->getp = info->wbuf;
		info->x_char = 0;
		info->line = line;

		init_waitqueue_head(&info->delta_msr_wait);
		
		if (request_irq(info->irq_urxint, rs7211_rx_int, 0, "rs7211rx", info))
			panic("Couldn't get rx irq for rs7211\n"); 
		if (request_irq(info->irq_utxint, rs7211_tx_int, 0, "rs7211tx", info))
			panic("Couldn't get tx irq for rs7211\n"); 

		if ((info->irq_umsint != IRQ_NONE) &&
		    (request_irq(info->irq_umsint, rs7211_ms_int, 0, "rs7211ms", info)))
			panic("Couldn't get ms irq for rs7211\n");

	}

#ifdef CONFIG_ARCH_EDB7211
	/* Enable output on PB2, the GPIO pin for RTS on UART1 */
	clps_writeb(clps_readb(PBDDR) | 0x2, PBDDR);
#endif

	ret = tty_register_driver(&rs7211_driver);
	if (ret) {
		printk(KERN_ERR "Couldn't register EP7211 serial driver\n");
		return ret;
	}

	/* devfs (kanne) */
	for( line = 0; line < NR_PORTS; line++ )
	     tty_register_devfs( &rs7211_driver, 0, MINOR_START + line );

	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	callout_driver = rs7211_driver;
#ifdef CONFIG_DEVFS_FS
	callout_driver.name = "cua/%d";
#else
	callout_driver.name = CU_NAME;
#endif
	callout_driver.major = CALLOUT_EP7211_MAJOR;
	callout_driver.subtype = SERIAL_TYPE_CALLOUT;

	if (tty_register_driver(&callout_driver))
		panic("Couldn't register callout driver\n");
	
	return 0;
}

static void __exit rs7211_fini(void)
{
	unsigned long flags;
	int line, ret;

	save_flags(flags);
	cli();
	for (line = 0; line < NR_PORTS; line++) {
		struct driver_data *info = &(driver_data[line]);
		free_irq(info->irq_urxint, NULL);
		free_irq(info->irq_utxint, NULL);

		/* devfs (kanne) */
		tty_unregister_devfs( &rs7211_driver, MINOR_START + line );
	}

	ret = tty_unregister_driver(&rs7211_driver);
	if (ret)
		printk(KERN_ERR "Couldn't unregister EP7211 serial driver (%d)\n", ret);
	restore_flags(flags);
}

module_init(rs7211_init);
module_exit(rs7211_fini);

#ifdef CONFIG_SERIAL_EP7211_CONSOLE
/************** console driver *****************/

static void rs7211_console_write(struct console *co, const char *s, u_int count)
{
  struct driver_data *info = driver_data; // + co->index;
	unsigned int status, syscon;
	int i;

	/* ensure port is enabled */
	syscon = clps_readl(SYSCON);
	clps_writel(syscon | SYSCON_UARTEN, SYSCON);

	/* transmit data */
	for (i = 0; i < count; i++) {
		do {
			status = clps_readl(SYSFLG);
		} while (status & SYSFLG_UTXFF);
		clps_writel(s[i], UARTDR);
		if (s[i] == '\n') {
			do {
				status = clps_readl(SYSFLG);
			} while (status & SYSFLG_UTXFF);
			clps_writel('\r', UARTDR);
		}
	}

	/* wait for transmitter to empty; restore state */
	do {
		status = clps_readl(SYSFLG);
	} while (status & SYSFLG1_UBUSY);

	clps_writel(syscon, SYSCON);
}

static int rs7211_console_wait_key(struct console *co)
{ 
	int c;

	disable_irq(IRQ_URXINT1);
	while (clps_readl(SYSFLG1) & SYSFLG_URXFE);
	c = clps_readl(UARTDR1);
	if (driver_data[0].rs7211_tty)
		enable_irq(IRQ_URXINT1);
	return c;
}

static kdev_t rs7211_console_device(struct console *co)
{
	return MKDEV(SERIAL_EP7211_MAJOR, MINOR_START);
}

static int __init rs7211_console_setup(struct console *co, char *options)
{
	int baud = 9600;
	int bits = 8;
	int parity = 'n';
	int cflag = CREAD | HUPCL | CLOCAL;

	if (options) {
		char *s = options;
		baud = simple_strtoul(options, NULL, 10);
		while (*s >= '0' && *s <= '9')
			s++;
		if (*s)
			parity = *s++;
		if (*s)
			bits = *s - '0';
	}

	/* Construct a cflag setting. */
	switch (baud) {
	case 1200:	cflag |= B1200; break;
	case 2400:	cflag |= B2400; break;
	case 4800:	cflag |= B4800; break;
	case 19200:	cflag |= B19200; break;
	case 38400:	cflag |= B38400; break;
	case 57600:	cflag |= B57600; break;
	case 115200:	cflag |= B115200; break;
	default:
		cflag |= B9600; break;
	}

	switch (bits) {
	case 7: cflag |= CS7; break;
	default:
		cflag |= CS8; break;
	}

	switch (parity) {
	case 'o': 
	case 'O':
		cflag |= PARODD; break;
	case 'e': 
	case 'E':
		cflag |= PARENB; break;
	}

	co->cflag = cflag;
	rs7211_set_cflag(&(driver_data[0]), cflag);
	rs7211_console_write( co, "boot ", 5 );
	if (!options)
		options = "no options";
	rs7211_console_write(NULL, options, strlen(options));
	rs7211_console_write(NULL, "\n", 1);
	return 0;
}

static struct console rs7211_cons =
{
	TTY_NAME,
	rs7211_console_write,
	NULL,
	rs7211_console_device,
	rs7211_console_wait_key,
	NULL,
	rs7211_console_setup,
	CON_PRINTBUFFER,
	-1,
	0,
	NULL
};

void __init rs7211_console_init(void)
{
	register_console(&rs7211_cons);
}

#endif /* CONFIG_SERIAL_EP7211_CONSOLE */
