/* drivers/video/stpcfb.c
 *
 * Frame buffer device driver for the STPC video card.
 * Supported models: Atlas, Consumer-II, in theory also the other STPC models.
 * Derived from vga256fb.c
 *
 * Copyright (C) 2001 Salvatore Sanfilippo. All rights reserved.
 * Copyright (C) 2001 Marco Pantaleoni. All rights reserved.
 * Copyright (C) 2001 SYSGO RTS GmbH. All rights reserved.
 *
 * AUTHORS:
 *   Salvatore Sanfilippo   <antirez@invece.org>
 *   Marco Pantaleoni       <panta@elasticworld.org>
 *   Armin Schindler        <acs@sysgo.de>
 *
 * TODO:
 * DONE          Add the right dependences to cfb16 and cfb24 in Config.in
 * DONE          Check the Makefile
 *               Support for 8bpp, 24bpp modes, 32bpp
 * DONE          Video blanking code, if needed
 * DONE          Arbitrary video modes support (TODO: use only one `par' structure as a template)
 * DONE          Detect the framebuffer size
 * DONE          Fix hardware detection (used in dot clock setting)
 * DONE          Remove references to Atlas in the code.
 *               Hardware panning
 *               Console hardware acceleration
 *               Also search for XXX and FIXME inside the code
 *
 * BUGS:
 *   Only 16bpp is guaranteed to work. Maybe 24bpp could work for some resolutions.
 *   8bpp has to be implemented.
 *
 * REFERENCES:
 *    STPC Atlas Programming Manual - Issue 0.1, 19 April 2001 - ST Microelectronics
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive
 * for more details.
 */

/* $Id: stpcfb.c,v 1.9 2001/08/03 14:29:09 acs Exp $ */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/selection.h>
#include <linux/init.h>
#include <linux/wrapper.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/pci.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/pgalloc.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>

#include "vga.h"			    /* FIXME: add the relevant stuff to stpcfb.h */
#include "stpcfb.h"

#define MNAME "stpcfb"
/*
 *  Debug macros
 */
#undef DEBUG
/* #define DEBUG */
#ifdef DEBUG
#define KDBG_WAIT(jdelay) \
do { \
        unsigned long j = jiffies + jdelay; \
        while (jiffies < j) \
                schedule(); \
} while (0)

#define KDBG_WAIT_SEC(sec_delay) \
do { \
        unsigned long j = jiffies + sec_delay * HZ; \
        while (jiffies < j) \
                schedule(); \
} while (0)
#  define DPRINTK(fmt, args...) do { printk(MNAME " [%s]: " fmt , __FUNCTION__ , ## args); } while (0)
#else
#  define DPRINTK(fmt, args...)
#endif

/* Hardware specific data for video modes.
 *
 * xres: X resolution
 * yres: Y resolution
 * bpp:  bits per pixel
 * mode: pixel RGBA bits organization
 *
 * Just for clearness:
 *
 * 69 crt		registers  index      : 0x3[BD]4  read/write: 0x3[BD]5
 * 21 attrib		registers  index/write: 0x3C0	  read/write: 0x3C1
 * 9  graphics     	registers  index      : 0x3CE	  read/write: 0x3CF
 * 5  sequencer		registers  index      : 0x3C4	  read/write: 0x3C5
 * 1  misc		register   read       : 0x3CC	  write     : 0x3C2
 * 2  dclk		registers  (host register pair, see the code and the programming manual)
 *
 * It is not useful to include the DAC registers and other vga registers
 */
struct stpcfb_par {
	u32             xres, yres, bpp;
	u32             refresh;
	u32             pixclock, left, right, upper, lower;
	u32             hslen, vslen;
	u32             hsync, vsync, csync, gsync, extsync;
	u32             vm_laced, vm_double;

	u32             mode;		    /* color mode: 565 (16 bpp), 888 (24 bpp), 8888 (32 bpp) */

	unsigned char   crt[0x45];	    /* 69 crt sub-registers (standard VGA: 24) */
	unsigned char   attrib[0x15];	    /* 21 attribute sub-registers */
	unsigned char   graphic[0x09];	    /* 9 graphic sub-registers */
	unsigned char   sequencer[0x05];    /* 5 sequencer sub-registers               */
	unsigned char   host[0x08];	    /* DCLK control registers */
	unsigned char   misc;		    /* misc register */

	struct fb_var_screeninfo var;
};

struct stpcfb_info {
	struct fb_info           info;
	struct fb_fix_screeninfo fix;
	struct fb_var_screeninfo var;
	struct display           disp;
	struct stpcfb_par        par;	    /* current parameters */
	unsigned char           *vbase;	    /* on-screen video memory mmap */
	unsigned char           *regs;	    /* memory mapped registers */
	unsigned char           *offscreen; /* off-screen video memory mmap */

	int                      blanked;   /* true if blanked */

	struct {
#ifdef FBCON_HAS_CFB16
		u16 cfb16[16];
#endif
#ifdef FBCON_HAS_CFB24
		u32 cfb24[16];
#endif
#ifdef FBCON_HAS_CFB32
		u32 cfb32[16];
#endif
	} fbcon_cmap;
	struct {
		__u8 red, green, blue;
	} palette[256];
	unsigned long linelength[MAX_NR_CONSOLES];
};

static unsigned long stpc_fb_addr;
static unsigned long stpc_offscreen_fb_addr;
static unsigned long stpc_fb_size;
static unsigned long stpc_mmap_regs;
static unsigned long stpc_gbase;

static struct stpcfb_info fb_info;	/* frame buffer info */
static int currcon = 0;

#define DEFAULT_OPT_XRES  640	
#define DEFAULT_OPT_YRES	480
#define DEFAULT_OPT_BPP		24

static const char *mode_option __initdata = NULL;
static char fontname[FONTNAME_SZ] __initdata = { '\0' };

static u32 opt_xres      = DEFAULT_OPT_XRES;
static u32 opt_yres      = DEFAULT_OPT_YRES;
static u32 opt_bpp       = DEFAULT_OPT_BPP;
static u32 opt_refresh   = 0;
static u32 opt_pixclock  = 0;
static u32 opt_left      = 0;
static u32 opt_right     = 0;
static u32 opt_upper     = 0;
static u32 opt_lower     = 0;
static u32 opt_hslen     = 0;
static u32 opt_vslen     = 0;
static int opt_hsync     = 0;
static int opt_vsync     = 0;
static int opt_csync     = 0;
static int opt_gsync     = 0;
static int opt_extsync   = 0;
static int opt_laced     = 0;
static int opt_double    = 0;

static int opt_preserve  = 0;		    /* if TRUE, screen will not be cleared on initialization,
					       and console messages will not be displayed */

#define CLOCKNUM	0x02		    /* use second pair of synthesizers */
#define RR_LOW		50L		    /* minimum refresh rate: 50 Hz     */
#define RR_HIGH		140L		    /* minimum refresh rate: 140 Hz    */

/* Supported platforms */
typedef enum
{
	stpc_platform_invalid = 0,
	stpc_platform_unknown,		    /* Unknown platform */
	stpc_platform_gloria,		    /* STPC Client      */
	stpc_platform_eval200pc,	    /* STPC Industrial  */
	stpc_platform_eval110,		    /* STPC Consumer    */
	stpc_platform_eval210,		    /* STPC ATLAS       */
	stpc_platform_eval120,		    /* STPC Consumer S  */
	stpc_platform_eval130		    /* STPC Consumer II */
} stpc_platform;


/* --------------------------- chained modes -------------------------------- */

/* Atlas 640x480x8bpp */
/* linelength: 640 */
struct stpcfb_par atlas640x480x8bpp = {
	xres:	640,
	yres:	480,
	bpp:	8,
	mode:	0,

	/* TODO FIXME: missing clocking parameters */

	crt: {

		0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0x0B, 0x3E, 
		0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0xEA, 0x0C, 0xDF, 0x50, 0x00, 0xE7, 0x04, 0xE3, 
		0xFF, 0x00, 0x3E, 0x3B, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x11, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x00, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {

		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {

		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {

		0x01, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0xEB,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x74, 0xC5, 0x00, 0x00 }
};

/* Atlas 640x480x16bpp */
/* linelength: 1280 */
struct stpcfb_par atlas640x480x16bpp = {
	xres:	640,
	yres:	480,
	bpp:	16,
	mode:	565,

	/* TODO FIXME: missing clocking parameters */

	crt: {

		0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0x0B, 0x3E, 
		0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0xEA, 0x0C, 0xDF, 0xA0, 0x00, 0xE7, 0x04, 0xE3, 
		0xFF, 0x00, 0x3E, 0x6B, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x1B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x00, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {

		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {

		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {

		0x01, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0xEB,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x74, 0xC5, 0x00, 0x00 }
};

/* Atlas 640x480x24bpp */
/* linelength: 2048 */
struct stpcfb_par atlas640x480x24bpp = {
	xres:	640,
	yres:	480,
	bpp:	24,
	mode:	888,

	pixclock:  34920,
	left:      64,
	right:     96,
	upper:     30,
	lower:     9,
	hslen:     112,
	vslen:     2,
	hsync:     0,
	vsync:     0,
	csync:     0,
	gsync:     0,
	extsync:   0,
	vm_laced:  0,
	vm_double: 0,


	crt: {

		0x6d, 0x4F, 0x4f, 0x90, 0x5c, 0x8a, 0x07, 0x3E, 
		0x00, 0x40, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0xE8, 0x0a, 0xDF, 0xf0, 0x1f, 0xdf, 0x07, 0xE3, 
		0xFF, 0x00, 0x3E, 0x9a, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x00, 0x00, 0x00, 0x10, 0x14, 0xB8, 
		0x1C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x00, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {

		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {

		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {

		0x03, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0xCB,
	host:   { 0x74, 0xc5, 0x5c, 0xae, 0x4b, 0x48, 0x6e, 0x69 }
};

/* Atlas 800x600x8bpp */
/* linelength: 800 */
struct stpcfb_par atlas800x600x8bpp = {
	xres:	800,
	yres:	600,
	bpp:	8,
	mode:	0,

	/* TODO FIXME: missing clocking parameters */

	crt: {

		0x7F, 0x63, 0x64, 0x82, 0x6B, 0x1B, 0x72, 0xF0, 
		0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0x59, 0x0D, 0x57, 0x64, 0x00, 0x57, 0x73, 0xE3, 
		0xFF, 0x00, 0x3E, 0x5B, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x11, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x00, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {

		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x01, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {

		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {

		0x01, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0x2B,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x5B, 0x7B, 0x00, 0x00 }
};

/* Atlas 800x600x16bpp */
/* linelength: 1600 */
struct stpcfb_par atlas800x600x16bpp = {
	xres:	800,
	yres:	600,
	bpp:	16,
	mode:	565,

#if ORIGINAL
	/* 800x600 @ 50 Hz */
	pixclock:  32199,
	left:      48,
	right:     16,
	upper:     39,
	lower:     8,
	hslen:     96,
	vslen:     2,
	hsync:     0,
	vsync:     0,
	csync:     0,
	gsync:     0,
	extsync:   0,
	vm_laced:  0,
	vm_double: 0,
#endif
	/* 800x600 @ 76 Hz */
	pixclock:  20943,
	left:      64,
	right:     8,
	upper:     39,
	lower:     8,
	hslen:     96,
	vslen:     2,
	hsync:     0,
	vsync:     0,
	csync:     0,
	gsync:     0,
	extsync:   0,
	vm_laced:  0,
	vm_double: 0,

	crt: {
		0x7F, 0x63, 0x64, 0x82, 0x6B, 0x1B, 0x72, 0xF0, 
		0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0x59, 0x0D, 0x57, 0xC8, 0x00, 0x57, 0x73, 0xE3, 
		0xFF, 0x00, 0x3E, 0xAB, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x1B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x60, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00
	},
	attrib: {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00
	},
	graphic: {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF
	},
	sequencer: {
		0x01, 0x01, 0x0F, 0x00, 0x0E
	},
	misc:	0x2B,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x5B, 0x7B, 0x00, 0x00 }
};

/* Atlas 800x600x24bpp */
/* linelength: 3272 */
struct stpcfb_par atlas800x600x24bpp = {
	xres:	800,
	yres:	600,
	bpp:	24,
	mode:	888,

	pixclock:  32199,
	left:      48,
	right:     16,
	upper:     39,
	lower:     8,
	hslen:     96,
	vslen:     2,
	hsync:     0,
	vsync:     0,
	csync:     0,
	gsync:     0,
	extsync:   0,
	vm_laced:  0,
	vm_double: 0,

	crt: {
		0x7F, 0x63, 0x64, 0x82, 0x6B, 0x1B, 0x72, 0xF0, 
		0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0x59, 0x0D, 0x57, 0x99, 0x00, 0x57, 0x73, 0xE3, 
		0xFF, 0x10, 0x3E, 0xBB, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x1C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x60, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {
		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {
		0x01, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0x2B,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x5B, 0x7B, 0x00, 0x00 }
};

/* Atlas 1024x768x8bpp */
/* linelength: 1024 */
struct stpcfb_par atlas1024x768x8bpp = {
	xres:	1024,
	yres:	768,
	bpp:	8,
	mode:	0,

	/* TODO FIXME: missing clocking parameters */

	crt: {

		0x99, 0x7F, 0x80, 0x9C, 0x83, 0x19, 0x2F, 0xF5, 
		0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0x01, 0x09, 0xFF, 0x80, 0x00, 0xFF, 0x30, 0xE3, 
		0xFF, 0x00, 0x3E, 0x93, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x11, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x00, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {

		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {

		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {

		0x01, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0x2B,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x6B, 0xEC, 0x00, 0x00 }
};

/* Atlas 1024x768x16bpp */
/* linelength: 2048 */
struct stpcfb_par atlas1024x768x16bpp = {
	xres:	1024,
	yres:	768,
	bpp:	16,
	mode:	565,

	/* TODO FIXME: missing clocking parameters */

	crt: {

		0x99, 0x7F, 0x80, 0x9C, 0x83, 0x19, 0x2F, 0xF5, 
		0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x6D, 0x20, 
		0x01, 0x09, 0xFF, 0x00, 0x00, 0xFF, 0x30, 0xE3, 
		0xFF, 0x10, 0x3E, 0xBB, 0x04, 0x00, 0x00, 0x80, 
		0x00, 0xFF, 0x45, 0x00, 0x00, 0x00, 0x14, 0xB8, 
		0x1B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0xFF, 0x01, 0x01, 0x00, 0x00, 0x00, 0xFF, 
		0x00, 0x00, 0x00, 0x00, 0x00, 
	},
	attrib: {

		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 
		0x41, 0x00, 0x0F, 0x00, 0x00, 
	},
	graphic: {

		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 
		0xFF, 
	},
	sequencer: {

		0x01, 0x01, 0x0F, 0x00, 0x0E, 
	},
	misc:		0x2B,
	host:   { 0x00, 0x00, 0x00, 0x00, 0x6B, 0xEC, 0x00, 0x00 }
};

struct stpcfb_par atlas400x234x24bpp = {
	xres:	400,
	yres:	234,
	bpp:	24,
	mode:	888,

	pixclock:  125000,
	left:      50,
	right:     10,
	upper:     17,
	lower:     10,
	hslen:     37,
	vslen:     3,
	hsync:     0,
	vsync:     1,
	csync:     0,
	gsync:     0,
	extsync:   0,
	vm_laced:  0,
	vm_double: 0,

	crt: {

    0x3b, 0x31, 0x31, 0x9e, 0x34, 0x99, 0x06, 0x11, 
    0x00, 0x40, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
    0xf3, 0x06, 0xe9, 0x96, 0x1f, 0xe9, 0x06, 0xe3,
    0xff, 0x00, 0x3e, 0x9a, 0x04, 0x00, 0x00, 0x80,
    0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x14, 0xb8,
    0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
    0x00, 0xff, 0x01, 0x00, 0x00, 0x00, 0xff, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00,
	},
	attrib: {

    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x41, 0x00, 0x0f, 0x00, 0x00,
	},
	graphic: {

    0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0f,
    0xff,
	},
	sequencer: {

    0x03, 0x01, 0x0f, 0x00, 0x0e,
	},
	misc:		0x8b,
	host:   { 0x74, 0xc5, 0x5c, 0xae, 0x75, 0x7d, 0x6e, 0x69 }
};

/* Available preset modes */
struct stpcfb_par *stpcfb_modes[] = {
	&atlas640x480x8bpp,
	&atlas640x480x16bpp,
	&atlas640x480x24bpp,
	&atlas800x600x8bpp,
	&atlas800x600x16bpp,
	&atlas800x600x24bpp,
	&atlas1024x768x8bpp,
	&atlas1024x768x16bpp,
	&atlas400x234x24bpp,
	NULL
};

/* The default mode */
struct stpcfb_par *default_mode = &atlas640x480x24bpp;


/* ========================================================================
 *   P R O T O T Y P E S
 * ------------------------------------------------------------------------ */

/* Framebuffer API: exported functions */
static int  stpc_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info);
static int  stpc_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info);
static int  stpc_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *fb);
static int  stpc_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info);
static int  stpc_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info);
static void do_install_cmap(int con, struct fb_info *info);
static int  stpc_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info);
static int  stpc_updatevar(int con, struct fb_info *info);
static void stpc_blank(int blank, struct fb_info *info);
static int  stpc_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
			     u_int *transp, struct fb_info *info);
static int  stpc_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
			     u_int transp, struct fb_info *info);

/* utility functions used by the above API ones */
static void stpcfb_encode_var(struct fb_var_screeninfo *var,
			      const struct stpcfb_par *par,
			      const struct stpcfb_info *info);
static int  stpcfb_decode_var(struct fb_var_screeninfo *var,
			      struct stpcfb_par *par,
			      const struct stpcfb_info *info);

/* VGA hardware access */
static void          stpc_disable_video(void);
static void          stpc_enable_video(void);
static void          stpc_unlock_crt_registers(void);
static int           stpc_get_current_dot_clock(struct stpcfb_par *par);
static void          stpc_select_current_dot_clock(struct stpcfb_par *par, int clocknum);
static void          stpc_set_dot_clock(struct stpcfb_par *par, int clocknum, long frequency);
#if __UNUSED__
static void          stpc_set_refresh_rate(struct stpcfb_par *par, unsigned long refresh_rate);
static u32           stpc_get_video_hardware_id(void);
#endif
#ifdef DEBUG
static u32           stpc_dclk_to_refresh(struct stpcfb_par *par, unsigned long dclk);
#endif
static u32           stpc_refresh_to_dclk(struct stpcfb_par *par, unsigned long refresh_rate);
static void          stpc_load_regs(struct stpcfb_par *regs);
static void          stpc_load_par(struct stpcfb_par *par);
static void __init   stpc_hardware_init(void);
static unsigned long stpc_get_framebuffer_size(void);

/* misc functions */
static stpc_platform stpc_get_hardware_id(void);
static int                stpc_get_modes_count(void);
static void               store_rgb_param(struct fb_var_screeninfo *var,
					  struct stpcfb_par *p);
static unsigned long stpc_get_offset(struct stpcfb_par *par);

static u16 stpc_read_host_2x8(u8 index);

/* framebuffer init */
struct stpcfb_par *stpcfb_modedb_like(const char *mode);
int        __init    stpcfb_setup(char *options);
static int __init    init_stpc(struct stpcfb_info *vi);
static int __init    stpc_of_init(void);
int        __init    stpcfb_init(void);

/* ------------------------- Access to VGA registers ------------------------ */

/* Extended registers */
#define VGA_CRTC_REPAINT_CTL0	0x19	    /* CRTC Offset bits 9:8 - CRTC Start Address bits 19:16 */
#define VGA_CRTC_REPAINT_CTL4	0x25	    /* various bits  */
#define VGA_CRTC_PALETTE_CTL	0x28	    /* Pixel Format and other amenities                     */

/* sequence registers */
#define V_SEQ_INDEX	0x3C4	/* Sequencer address (index) register */
#define V_SEQ_RW	0x3C5	/* Sequencer data register */

#define put_sequence(i,b) \
	do { \
		outb(i, V_SEQ_INDEX); \
		outb(b, V_SEQ_RW); \
	} while(0)

#define get_sequence(i,v) \
	do { \
		outb(i, V_SEQ_INDEX); \
		v = inb(V_SEQ_RW); \
	} while(0)

/* graphics registers */
#define V_GR_INDEX	0x3CE	/* VGA address (index) register */
#define V_GR_RW		0x3CF	/* VGA data register */

#define put_graph(i,b) \
	do { \
		outb(i, V_GR_INDEX); \
		outb(b, V_GR_RW); \
	} while(0)

#define get_graph(i,v) \
	do { \
		outb(i, V_GR_INDEX); \
		v = inb(V_GR_RW); \
	} while(0)

/* misc registers */
#define V_MISC_R	0x3CC	/* VGA misc register (read) */
#define V_MISC_W	0x3C2	/* VGA misc register (write) */

#define put_misc(b) do { outb(b, V_MISC_W); } while(0)
#define get_misc(v) do { v = inb(V_MISC_R); } while(0)

/* attribute registers */
#define V_ATTR_IW	0x3C0	/* Attribute index and data register (write) */
#define V_ATTR_R	0x3C1	/* Attribute data register (read) */

/* warning! the attribute index/write_data register may be in desync
 * since there is a flip/flop inside the VGA.
 *
 * To reset the attribute register flip/flop we must read the Input Status #1
 * register, this can be done using this macro.
 */
#define reset_attr() do { (void) inb(V_ISTAT1_R); } while(0)

/* safe attribute registers read and write,
   warning: as you can see the 0x20 bit will be set to zero
   so the video output will be disabled, if you want read or write
   (since some VGA cards allows you to write without setting te PAS
   bit) without put the video off OR the index with 0x20 */
#define put_attr(i, b) \
	do { \
		volatile unsigned char t; \
		t = inb(V_ISTAT1_R);	/* reset the flip/flop */ \
		udelay(1); \
		outb(i, V_ATTR_IW);     /* set the index */ \
		udelay(1);		/* delay */ \
		outb(b, V_ATTR_IW);	/* write data */ \
		udelay(1); \
		(void) inb(V_ISTAT1_R);	/* reset the flip/flop */ \
	} while(0)

#define get_attr(i, v) \
	do { \
		volatile unsigned char t; \
		t = inb(V_ISTAT1_R);	/* reset the flip/flop */ \
		udelay(1); \
		outb(i, V_ATTR_IW);     /* set the index */ \
		udelay(1);		/* delay */ \
		v = inb(V_ATTR_R);	/* read data */ \
		udelay(1); \
		(void) inb(V_ISTAT1_R);	/* reset the flip/flop */ \
	} while(0)

/* color registers */
#define V_DAC_WRMODE_RW	0x3C8	/* DAC address write mode register read/write */
#define V_DAC_RDMODE_W	0x3C7	/* DAC address read mode register, write */
#define V_DAC_RW	0x3C9	/* DAC data register, read/write */
#define V_DAC_STATE_R	0x3C7	/* DAC state register, read */

#define put_dac_wrmode(b)	do { outb(b, V_DAC_WRMODE_RW); } while(0)
#define get_dac_wrmode(v)	do { v = inb(V_DAC_WRMODE_RW); } while(0)
#define put_dac_rdmode(b)	do { outb(b, V_DAC_RDMODE_W);  } while(0)
#define put_dac(b)		do { outb(b, V_DAC_RW);        } while(0)
#define get_dac(v)		do { v = inb(V_DAC_RW);        } while(0)
#define get_dac_state(v)	do { v = inb(V_DAC_STATE_R);   } while(0)

/* V_DAC_STATE_R and V_DAC_RDMODE_W are the same register, but the first
 * in write and the second in read */

/* How to manage the palette using the color registers:
 *
 * To write a palette entry (from the 2 between 256)
 *   - write the index to the DAC address write mode register (V_DAC_WRMODE_RW)
 *   - then performs 3 writes (RGB) to the DAC data register (V_DAC_RW)
 * The index will be incremented after this operation, so you will be able
 * to load all the palette just writing all values to the DAC data register.
 *
 * To read a palette entry
 *   - write the index to the DAC address read mode register (V_DAC_RDMODE_R)
 *   - then performs 3 reads (RGB) from the DAC data register (V_DAC_RW)
 * The index will be incremented as explained.
 */

#define V_ISTAT0_R	0x3C2	/* Input status register #0 */
#define V_FEATURE_R	0x3CA	/* feature control register (read) */

#define get_feature(v) do { v = inb(V_FEATURE_R); } while(0)
#define get_istat0(v)  do { v = inb(V_ISTAT0_R);  } while(0)

/* this registers are obtained at runtime */
static int V_CRT_INDEX;		/* CRT address (index) register */
static int V_CRT_RW;		/* CRT data register */
static int V_ISTAT1_R;		/* Input status register #1 */
static int V_FEATURE_W;		/* feature control register, (write) */

#define put_crt(i,b) \
	do { \
		outb(i, V_CRT_INDEX); \
		outb(b, V_CRT_RW); \
	} while(0)

#define get_crt(i,v) \
	do { \
		outb(i, V_CRT_INDEX); \
		v = inb(V_CRT_RW); \
	} while(0)

#define get_istat1(v)  do { v = inb(V_ISTAT1_R);  } while(0)
#define put_feature(b) do { outb(b, V_FEATURE_W); } while(0)

#define stpc_select_write_plane(plane)  do { put_sequence(0x02, 1 << plane); } while (0)
#define stpc_select_read_plane(plane)   do { put_graph(0x04, plane);         } while (0)
#define stpc_sequencer_sync_reset_on()  do { put_sequence(0x00, 0x01);       } while (0)
#define stpc_sequencer_reset_off()      do { put_sequence(0x00, 0x03);       } while (0)

/* ------------------------ Framebuffer API: exported function  ----------------------- */
/* stpc_get_fix -- get the fix structure for the given console */
static int stpc_get_fix(struct fb_fix_screeninfo *fix, int con,
			struct fb_info *info)
{
	struct stpcfb_info *vi = (struct stpcfb_info*) info;

	DPRINTK("ENTERED con=%d\n", con);

	/* Copy the fix from the fb_info to get the fields
	 * that are really fixed */
	*fix = vi->fix;

	/* handle the fields relative to var */
	if (con >= 0)			    /* TODO FIXME: con < 0 ????? */
		fix->line_length = vi->linelength[con];
	else
		fix->line_length = vi->linelength[currcon];

	return 0;
}

/* stpc_get_var -- get the var structure for the given console */
static int stpc_get_var(struct fb_var_screeninfo *var, int con,
			struct fb_info *info)
{
	struct stpcfb_info *vi = (struct stpcfb_info*) info;

	DPRINTK("ENTERED con=%d\n", con);

	if (con == -1)
		*var = vi->var;
	else
		*var = fb_display[con].var;
	return 0;
}

/* stpc_set_var -- set new var screen info for a console.
 * Called from others to change the video mode
 * Called from stpc_switch to setup the new console */
static int stpc_set_var(struct fb_var_screeninfo *var, int con,
			struct fb_info *fb)
{
	struct stpcfb_info *vi = (struct stpcfb_info*) fb;
	struct display     *d  = (con < 0) ? &vi->disp : &fb_display[con];
	struct stpcfb_par   par;
	int    err;

	DPRINTK("ENTERED con=%d\n", con);

	/* this driver doesn't support virtual resolutions */
	if ((var->xres != var->xres_virtual) ||
	    (var->yres != var->yres_virtual))
	{
		DPRINTK("ERROR effective resolution != virtual\n");
		return -EINVAL;
	}

	par = atlas640x480x24bpp;	    /* use this as a template */

	if ((err = stpcfb_decode_var(var, &par, vi)) != 0)
	{
		DPRINTK("ERROR in decode_var\n");
		return err;
	}
	stpcfb_encode_var(var, &par, vi);

	/* Test? The mode is valid, return 0 */
	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_TEST)
		return 0;

	/* Activate now! */
	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		int oldxres, oldyres, oldbpp;

		/* Make a backup of the old values before to
		 * overwrite the var structure in the display structure */
		oldxres     = d->var.xres;
		oldyres     = d->var.yres;
		oldbpp      = d->var.bits_per_pixel;

		/* Copy the user supplied var structure */
		d->var = *var;
		store_rgb_param(&d->var, &par);

		/* Set the new display parameters only if something
		 * interesting changed */
		if (oldxres != var->xres           ||
		    oldyres != var->yres           ||
		    oldbpp  != var->bits_per_pixel)
		{
			struct fb_fix_screeninfo f;

			DPRINTK("*** RELEVANT PARAMETERS CHANGED ***\n");

			/* Set the display structure */
			stpc_get_fix(&f, con, fb);
			d->screen_base    = vi->vbase;
			d->visual         = f.visual;
			d->type           = f.type;
			d->type_aux       = f.type_aux;
			d->ypanstep       = f.ypanstep;
			d->ywrapstep      = f.ywrapstep;
			d->line_length    = stpc_get_offset(&par);
			d->can_soft_blank = 1;
			d->inverse        = 0;
			switch (par.bpp) {
#ifdef FBCON_HAS_CFB16
			case 16:
				d->dispsw      = &fbcon_cfb16;
				d->dispsw_data = vi->fbcon_cmap.cfb16; /* console palette */
				break;
#endif
#ifdef FBCON_HAS_CFB24
			case 24:
				d->dispsw      = &fbcon_cfb24;
				d->dispsw_data = vi->fbcon_cmap.cfb24; /* console palette */
				break;
#endif
#ifdef FBCON_HAS_CFB32
			case 32:
				d->dispsw      = &fbcon_cfb32;
				d->dispsw_data = vi->fbcon_cmap.cfb32; /* console palette */
				break;
#endif
			default:
				d->dispsw      = &fbcon_dummy;
			}
			if (opt_preserve) {
				d->dispsw = &fbcon_dummy;
      }
			d->scrollmode = SCROLL_YREDRAW;

			/* Update the console */
			if (vi->info.changevar)
				(*vi->info.changevar)(con);
		}

		/* Store the line length for this console */
		if (con >= 0)
			vi->linelength[con]     = stpc_get_offset(&par);
		else
			vi->linelength[currcon] = stpc_get_offset(&par);

		/* Current console? Write the new hardware settings */
		if (con == currcon) {
			DPRINTK("con == currcon: writing new hardware settings\n");
			stpc_load_par(&par);
			vi->par = par;

			/* XXX: hack.
			 * We need to copy the display structure in the
			 * global one: the /dev/fb* char device uses
			 * the global display structure to get the
			 * screen_base. Probably all the other framebuffers
			 * don't change the framebuffer address for
			 * different modes, but we need to do it */
			if (d != &vi->disp)
				vi->disp = *d;
		}
	}
	return 0;
}

static int stpc_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			 struct fb_info *info)
{
	if (con == currcon)		/* current console? */
		return fb_get_cmap(cmap, kspc, stpc_getcolreg, info);
	if (fb_display[con].cmap.len)	/* non default cmap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else
		fb_copy_cmap(fb_default_cmap(32), cmap, kspc ? 0 : 2);
	return 0;
}

static int stpc_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			   struct fb_info *info)
{
	int err;

	if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		err = fb_alloc_cmap(&fb_display[con].cmap, 32, 0);
		if (err) return err;
	}
	if (con == currcon)
		return fb_set_cmap(cmap, kspc, stpc_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	return 0;
}

static void do_install_cmap(int con, struct fb_info *info)
{
	if (con != currcon)
		return;
	if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, 1, stpc_setcolreg, info);
	else
		fb_set_cmap(fb_default_cmap(32), 1, stpc_setcolreg, info);
}

/* stpc_switch -- switch the console
 * All the real work is done inside stpc_set_var() */
static int stpc_switch(int con, struct fb_info *info)
{
	static int init = 1;

	DPRINTK("ENTERED con=%d%s\n", con, init ? "  INITIALIZATION" : "");

	/* Set the hardware parameters the first time */
	if (init) {
		if (! opt_preserve) {
  		stpc_load_par(default_mode);
	  	memset(fb_info.vbase, 0, stpc_fb_size); /* Clear it */
    }
	}

	/* Do we have to save the colormap? */
	if (fb_display[currcon].cmap.len)
		fb_get_cmap(&fb_display[currcon].cmap, 1, stpc_getcolreg,
			    info);
	currcon = con;
	if (! opt_preserve) {
  	stpc_set_var(&fb_display[con].var, con, info);
  }
	do_install_cmap(con, info);

  init = 0;
	return 0;
}

static int stpc_updatevar(int con, struct fb_info *info)
{
	/* Nothing to update */
	return 0;
}

/* stpc_blank -- blank the screen */
static void stpc_blank(int blank, struct fb_info *info)
{
	struct stpcfb_info *vi = (struct stpcfb_info*) info;
#if OLD
	int i;

	/* FIXME: this works only for 8bpp modes */
	if (blank && !vi->blanked) {
		/* save the current palette to restore it at unblanking */
		put_dac_rdmode(0);
		udelay(1);
		for (i = 0; i < 256; i++) {
			get_dac(vi->palette[i].red);
			get_dac(vi->palette[i].green);
			get_dac(vi->palette[i].blue);
		}
		/* blank it */
		put_dac_wrmode(0);
		udelay(1);
		for (i = 0; i < 256; i++) {
			put_dac(0);
			put_dac(0);
			put_dac(0);
		}
		vi->blanked++;
	} else {
		/* restore the palette */
		put_dac_wrmode(0);
		udelay(1);
		for (i = 0; i < 256; i++) {
			put_dac(vi->palette[i].red);
			put_dac(vi->palette[i].green);
			put_dac(vi->palette[i].blue);
		}
		vi->blanked = 0;
	}
#endif
	u8 sr1;

	/* Blank/unblank by turning the screen off/on */

	get_sequence(0x01, sr1);
	if (blank & !vi->blanked)
	{
		sr1 |= 0x20;
	} else
	{
		sr1 &= (u8) ~((u8) 0x20);
	}
	put_sequence(0x01, sr1);
}

/* Get the color registers */
static int stpc_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
			    u_int *transp, struct fb_info *info)
{
	struct stpcfb_info *vi = (struct stpcfb_info*) info;

	if (regno > 255)
		return 1;
	*red    = vi->palette[regno].red;
	*green  = vi->palette[regno].green;
	*blue   = vi->palette[regno].blue;
	*transp = 0; /* unsupported */
	return 0;
}

/* Set the color registers */
static int stpc_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
			  u_int transp, struct fb_info *info)
{
	struct stpcfb_info *vi = (struct stpcfb_info*) info;

	if (regno > 255)
		return 1;

	vi->palette[regno].red   = red;
	vi->palette[regno].green = green;
	vi->palette[regno].blue  = blue;

	/* to support 8bpp we need to set the palette */
	if (vi->par.bpp == 8) {
		put_dac_wrmode(regno);
		udelay(1);
		put_dac((red >> 10) & 0x3F);
		put_dac((green >> 10) & 0x3F);
		put_dac((blue >> 10) & 0x3F);
	}

	if (regno < 16) {
		switch(vi->par.bpp) {
#ifdef FBCON_HAS_CFB16
		case 16:
			vi->fbcon_cmap.cfb16[regno] = (((red & 0xf800))        |
						       ((green & 0xfc00) >> 5) |
						       ((blue & 0xf800) >> 11 ));
			break;
#endif
#ifdef FBCON_HAS_CFB24
		case 24:
			vi->fbcon_cmap.cfb24[regno] = (((blue & 0xff00) << 8) |
						       (green & 0xff00)       |
						       ((red & 0xff00) >> 8));
			break;
#endif
#ifdef FBCON_HAS_CFB32
		case 32:
			vi->fbcon_cmap.cfb32[regno] = (((blue & 0xff00) << 8) |
						       (green & 0xff00) |
						       ((red & 0xff00) >> 8));
			break;
#endif
		}
	}
	return 0;
}

/* utility functions used by the above API ones */

static void stpcfb_encode_var(struct fb_var_screeninfo *var,
			      const struct stpcfb_par *par,
			      const struct stpcfb_info *info)
{
	*var = par->var;
}

#define FAIL(X) \
do                                          \
{                                           \
	printk(MNAME ": ERROR (" X ")\n");  \
	return -EINVAL;                     \
} while (0)

static int stpcfb_decode_var(struct fb_var_screeninfo *var,
			     struct stpcfb_par              *par,
			     const struct stpcfb_info       *info)
{
	u32 xres, right, hslen, left, xtotal;
	u32 yres, lower, vslen, upper, ytotal;
	u32 vxres, xoffset, vyres, yoffset;
	u32 pos;
	u8  r7, r25, rMode;
	u32 linelength;

	u32 dclk_freq, dclk_lo, dclk_hi;
	int i;

	DPRINTK("ENTERED\n");

	if ((var->bits_per_pixel != 16) &&
	    (var->bits_per_pixel != 24))    /* only 16 and 24 bpp are supported right now */
	{
		DPRINTK("invalid depth requested (%d)\n", var->bits_per_pixel);
		return -EINVAL;
	}

	/* round resolution values */
	xres    = (var->xres + 7) & ~7;
/*
	vxres   = (var->xres_virtual + 0xF) & ~0xF;
*/
	vxres   = xres;
	xoffset = (var->xoffset + 7) & ~7;
	left    = (var->left_margin + 7) & ~7;
	right   = (var->right_margin + 7) & ~7;
	hslen   = (var->hsync_len + 7) & ~7;

	/* clip values */
	if (vxres < xres)
		vxres = xres;
	if (xres + xoffset > vxres)
		xoffset = vxres - xres;

	par->var.xres         = xres;
	par->var.right_margin = right;
	par->var.hsync_len    = hslen;
	par->var.left_margin  = left;
	par->var.xres_virtual = vxres;
	par->var.xoffset      = xoffset;

	linelength = xres * var->bits_per_pixel / 8; /* TODO FIXME: BOGUS */

	r25 = 0x00;

	/* scale down horizontal values by 8 (the width of a character) */
	xres  >>= 3;
	right >>= 3;
	hslen >>= 3;
	left  >>= 3;
	vxres >>= 3;
	xtotal = xres + right + hslen + left;
	if (xtotal >= 256)
		FAIL("xtotal too big");
	if (hslen > 32)
		FAIL("hslen too big");
	if (right + hslen + left > 64)
		FAIL("hblank too big");
	par->crt[VGA_CRTC_H_TOTAL]       = xtotal - 5;		/* Horizontal Total          (CR0) */
	par->crt[VGA_CRTC_H_BLANK_START] = xres - 1;		/* Horizontal Display End    (CR1) */
	par->crt[VGA_CRTC_H_DISP]        = xres - 1; 		/* Horizontal Blanking Start (CR2) */ /* TODO FIXME: or xres ??? */
	pos = xres + right;
	par->crt[VGA_CRTC_H_SYNC_START]  = pos;			/* Horizontal Retrace Start  (CR4) */
	pos += hslen;
	par->crt[VGA_CRTC_H_SYNC_END]    = pos & 0x1F;		/* Horizontal Retrace End    (CR5) */
	pos += left - 2;		    /* blank_end + 2 <= total + 5 */
	par->crt[VGA_CRTC_H_BLANK_END]   = (pos & 0x1F) | 0x80;	/* Horizontal Blanking End   (CR3) */
	if (pos & 0x20)
		par->crt[VGA_CRTC_H_SYNC_END] |= 0x80;		/* bit 6 of horizontal blanking end */
	if (pos & 0x40)
		r25 |= 0x10;					/* bit 7 of horizontal blanking end */

	yres    = var->yres;
	lower   = var->lower_margin;
	vslen   = var->vsync_len;
	upper   = var->upper_margin;
	vyres   = var->yres_virtual;
	yoffset = var->yoffset;

	if (yres > vyres)
		vyres = yres;
	if (yoffset + yres > vyres)
		yoffset = vyres - yres;
	par->var.yres         = yres;
	par->var.lower_margin = lower;
	par->var.vsync_len    = vslen;
	par->var.upper_margin = upper;
	par->var.yres_virtual = vyres;
	par->var.yoffset      = yoffset;

	/* Fill the video mode stuff of the par structure */
	par->xres = xres;
	par->yres = yres;
	par->bpp  = var->bits_per_pixel;
	switch (par->bpp)
	{
	case 16:
		par->mode = 565;
		break;
	case 24:
		par->mode = 888;
		break;
	case 32:
		par->mode = 8888;
		break;
	}

	if (var->vmode & FB_VMODE_DOUBLE) {
		yres  <<= 1;
		lower <<= 1;
		vslen <<= 1;
		upper <<= 1;
	}
	ytotal = yres + lower + vslen + upper;
	if (ytotal > 1024) {
		ytotal >>= 1;
		yres   >>= 1;
		lower  >>= 1;
		vslen  >>= 1;
		upper  >>= 1;
		rMode = 0x04;
	} else
		rMode = 0x00;
	if (ytotal > 1024)
		FAIL("ytotal too big");
	if (vslen > 16)
		FAIL("vslen too big");

	par->crt[VGA_CRTC_V_TOTAL] = (ytotal - 2) & 0xFF;	/* Vertical Total            (CR6) */ /* 11th bit in CR25 */
	r25 |= ((ytotal - 2) >> 10) & 0x01; /* 11th bit of Vertical Total */
	r7 = 0x10;			    /* disable linecompare */
	if (ytotal & 0x100) r7 |= 0x01;
	if (ytotal & 0x200) r7 |= 0x20;
	par->crt[VGA_CRTC_PRESET_ROW] = 0;			/* Screen A Preset Row Scan  (CR8) */
	par->crt[VGA_CRTC_MAX_SCAN]   = 0x40;			/* Character Cell Height     (CR9) */ /* 1 scanline, no linecmp */
	par->var.vmode = var->vmode;
	if (var->vmode & FB_VMODE_DOUBLE)
		par->crt[VGA_CRTC_MAX_SCAN] |= 0x80;		/* Scan Double */
	par->crt[VGA_CRTC_CURSOR_START] = 0x20;			/* Cursor Start              (CRA) */
	par->crt[VGA_CRTC_CURSOR_END]   = 0x00;			/* Cursor End                (CRB) */
	pos = yoffset * vxres + (xoffset >> 3);
	par->crt[VGA_CRTC_START_HI]     = pos >> 8;		/* Start Address High        (CRC) */
	par->crt[VGA_CRTC_START_LO]     = pos & 0xFF;		/* Start Address Low         (CRD) */
	par->crt[VGA_CRTC_REPAINT_CTL0] = (pos >> 16) & 0x0F;	/* Repaint Control 0         (CR19) */
                                                                /* Upper 4 bits of the display buffer address */
	par->crt[VGA_CRTC_CURSOR_HI]    = 0x00;			/* Text Cursor Offset High   (CRE)  */
	par->crt[VGA_CRTC_CURSOR_LO]    = 0x00;			/* Text Cursor Offset Low    (CRF)  */
	pos = yres - 1;
	par->crt[VGA_CRTC_V_DISP_END]    = (yres - 1) & 0xFF;	/* Vertical Display End      (CR12) */
	par->crt[VGA_CRTC_V_BLANK_START] = (yres - 1) & 0xFF;	/* Vertical Blanking Start   (CR15) */
	r25 |= ((yres - 1) >> 9) & 0x02; /* 11th bit of Vertical Display End    */
	r25 |= ((yres - 1) >> 7) & 0x08; /* 11th bit of Vertical Blanking Start */
	if (pos & 0x100)
		r7 |= 0x0A;	/* 0x02 -> DISP_END, 0x08 -> BLANK_START */
	if (pos & 0x200) {
		r7 |= 0x40;	/* 0x40 -> DISP_END */
		par->crt[VGA_CRTC_MAX_SCAN] |= 0x20; /* BLANK_START */
	}
	pos += lower;
	par->crt[VGA_CRTC_V_SYNC_START] = pos & 0xFF;		/* Vertical Retrace Start    (CR10) */
	r25 |= (pos >> 8) & 0x04; /* 11th bit of Vertical Retrace Start */
	if (pos & 0x100)
		r7 |= 0x04;
	if (pos & 0x200)
		r7 |= 0x80;
	pos += vslen;
	par->crt[VGA_CRTC_V_SYNC_END] = (pos & 0x0F) & ~0x10;	/* Vertical Retrace End      (CR11) */
                                                                /* disabled IRQ */
	pos += upper - 1; /* blank_end + 1 <= ytotal + 2 */
	par->crt[VGA_CRTC_V_BLANK_END] = pos & 0xFF;		/* Vertical Blanking End     (CR16) */
                                                                /* 0x7F for original VGA,
							           but some SVGA chips requires all 8 bits to set */

	par->crt[VGA_CRTC_REPAINT_CTL4] = r25;			/* Repaint Control Register 4 (CR25) */

	if (vxres >= 512)
		FAIL("vxres too long");
	par->crt[VGA_CRTC_OFFSET]        = (u8)((linelength >> 3) & 0xFF); /* Offset         (CR13) */
	par->crt[VGA_CRTC_REPAINT_CTL0] |= (u8)(((linelength >> 3) >> 4) & 0xF0); /* in CR19 the value is left-shifted by 4 bits */
	par->crt[VGA_CRTC_UNDERLINE]    = 0x1F;
	par->crt[VGA_CRTC_MODE]         = rMode | 0xE3;
	par->crt[VGA_CRTC_LINE_COMPARE] = 0xFF;
	par->crt[VGA_CRTC_OVERFLOW]     = r7;			/* Overflow                  (CR7) */

	/* pixel format (CR28) */
	par->crt[VGA_CRTC_PALETTE_CTL] = 0x10;			/* Palette Control Register  (CR28) */
                                                                /* bit 4 is reserved and must be 1 */
	switch (var->bits_per_pixel)
	{
	case 8:
		par->crt[VGA_CRTC_PALETTE_CTL] |= 0x01;	/* LUT bypass inactive - 8 bit color (non-VGA) */
		break;
	case 15:
		par->crt[VGA_CRTC_PALETTE_CTL] |= 0x0A;	/* LUT bypass - 15 bit color                   */
		break;
	case 16:
		par->crt[VGA_CRTC_PALETTE_CTL] |= 0x0B;	/* LUT bypass - 16 bit color                   */
		break;
	case 24:
		par->crt[VGA_CRTC_PALETTE_CTL] |= 0x0C;	/* LUT bypass - 24 bit color                   */
		break;
	case 32:
		par->crt[VGA_CRTC_PALETTE_CTL] |= 0x0D;	/* LUT bypass - 32 bit color                   */
		break;
	}

	for (i = 0x00; i < 0x10; i++)
		par->attrib[i] = i;
	par->attrib[VGA_ATC_MODE]         = 0x41; /* vga16: 0x81 */
	par->attrib[VGA_ATC_OVERSCAN]     = 0x00; /* 0 for EGA, 0xFF for VGA */
	par->attrib[VGA_ATC_PLANE_ENABLE] = 0x0F;
	par->attrib[VGA_ATC_PEL]          = xoffset & 7;
	par->attrib[VGA_ATC_COLOR_PAGE]   = 0x00;

	par->misc = 0xC3;	/* enable CPU, ports 0x3Dx, positive sync */
	par->var.sync = var->sync;
	if (var->sync & FB_SYNC_HOR_HIGH_ACT)
		par->misc &= ~0x40;
	if (var->sync & FB_SYNC_VERT_HIGH_ACT)
		par->misc &= ~0x80;

	par->sequencer[VGA_SEQ_CLOCK_MODE]    = 0x01;
	par->sequencer[VGA_SEQ_PLANE_WRITE]   = 0x0F;
	par->sequencer[VGA_SEQ_CHARACTER_MAP] = 0x00;
	par->sequencer[VGA_SEQ_MEMORY_MODE]   = 0x0E; /* vga16: 0x06; */

	par->graphic[VGA_GFX_SR_VALUE]      = 0x00;
	par->graphic[VGA_GFX_SR_ENABLE]     = 0x00; /* vga16: 0x0F; */
	par->graphic[VGA_GFX_COMPARE_VALUE] = 0x00;
	par->graphic[VGA_GFX_DATA_ROTATE]   = 0x00; /* vga16: 0x20; */
	par->graphic[VGA_GFX_PLANE_READ]    = 0x00;
	par->graphic[VGA_GFX_MODE]          = 0x40; /* vga16: 0x00; */
	par->graphic[VGA_GFX_MISC]          = 0x05;
	par->graphic[VGA_GFX_COMPARE_MASK]  = 0x0F;
	par->graphic[VGA_GFX_BIT_MASK]      = 0xFF;

	/* DCLK */

	if (var->pixclock <= 0)
	{
		u8   cr6, cr7, cr25;
		u8   sr1, cr0;
		u32  vertical_total, horizontal_total;

		/* Get current vertical total (number of lines including blank) */
		cr6  = par->crt[0x06];
		cr7  = par->crt[0x07];
		cr25 = par->crt[0x25];
		vertical_total = (u32)cr6 + (((u32)cr7 & 0x1) << 8) + (((u32)cr7 & 0x20) << 4) + (((u32)cr25 & 0x1) << 10);

		/* Get current horizontal total (number of pixels per lines including blank) */
		sr1 = par->sequencer[0x01];
		cr0 = par->crt[0x00];
		if (sr1 & 0x01)
			horizontal_total = ((u32)cr0 + 5) * 8;
		else
			horizontal_total = ((u32)cr0 + 5) * 9;

		dclk_freq = (horizontal_total * vertical_total * RR_LOW);
		dclk_freq /= 1000L;
		var->pixclock  = 1000000000L / dclk_freq; 
	}
	dclk_freq  = 1000000000L / var->pixclock;
	dclk_freq *= 1000L;

	/* clip dot clock */
	dclk_lo = stpc_refresh_to_dclk(par, RR_LOW);
	dclk_hi = stpc_refresh_to_dclk(par, RR_HIGH);
	if (dclk_freq < dclk_lo)
		dclk_freq = dclk_lo;
	if (dclk_freq > dclk_hi)
		dclk_freq = dclk_hi;

	DPRINTK("effective dot clock: %lu (refresh rate: %d Hz)\n",
		(unsigned long)dclk_freq, stpc_dclk_to_refresh(par, dclk_freq));

	stpc_select_current_dot_clock(par, CLOCKNUM);/* use second pair of synthesizers */
	stpc_set_dot_clock(par, CLOCKNUM, dclk_freq);

	i = stpc_get_current_dot_clock(par);
	
	par->var.bits_per_pixel = var->bits_per_pixel;
	par->var.grayscale      = var->grayscale;

	par->var.nonstd = 0;
	par->var.activate = FB_ACTIVATE_NOW;
	par->var.height = -1;
	par->var.width  = -1;
	par->var.accel_flags = 0;
	par->var.pixclock = var->pixclock;

	DPRINTK("EXITED\n");
	return 0;
}

/* ------------------------- VGA low level access ------------------------ */

/* disable/enable video output using the PAS bit */
static void stpc_disable_video(void)
{
        volatile unsigned char t;

        t = inb(V_ATTR_IW);     /* get the current value */
        udelay(1);
        (void) inb(V_ISTAT1_R); /* reset the flip/flop */
        udelay(1);
        t &= 0xDF;              /* set the PAS bit */
        outb(t, V_ATTR_IW);     /* set the port */
        udelay(1);
        (void) inb(V_ISTAT1_R); /* reset the flip/flop */
}

static void stpc_enable_video(void)
{
        volatile unsigned char t;

        t = inb(V_ATTR_IW);     /* get the current value */
	udelay(1);
        (void) inb(V_ISTAT1_R); /* reset the flip/flop */
	udelay(1);
        t |= 0x20;              /* set the PAS bit */
        outb(t, V_ATTR_IW);     /* set the port */
	udelay(1);
        (void) inb(V_ISTAT1_R); /* reset the flip/flop */
}

static void stpc_unlock_crt_registers(void)
{
        volatile unsigned char aux;
        get_crt(0x11, aux);
        aux &= 0x7f;
        put_crt(0x11, aux);
}

#define STPC_HOST_INDEX		0x22
#define STPC_HOST_DATA		0x23
#define STPC_SDRAM_GRAPH_MEM	0x36

#define stpc_put_host(i, v) \
do {                              \
	outb(i, STPC_HOST_INDEX); \
	outb(v, STPC_HOST_DATA);  \
} while(0)

#define stpc_get_host(i, var) \
do {                                \
	outb((i), STPC_HOST_INDEX); \
	var = inb(STPC_HOST_DATA);  \
} while(0)

static u16 stpc_read_host_2x8(u8 index)
{
        u16 val;

	outb(index, STPC_HOST_INDEX);
        val  = inb(STPC_HOST_DATA);
        val += inb(STPC_HOST_DATA) << 8;

        return val;
}

static int stpc_get_current_dot_clock(struct stpcfb_par *par)
{
	int DCLK;

	/* get_misc(DCLK); */
	DCLK = par->misc;
	DCLK >>= 2;			    /* Clock Select are bits [3:2] in
					       Miscellaneous Output Register */
	DCLK  &= 0x3;

	return DCLK;
}

static void stpc_select_current_dot_clock(struct stpcfb_par *par, int clocknum)
{
	if ((clocknum < 0) || (clocknum > 3))
	{
		DPRINTK("ERROR wrong clock synthesizer pair number\n");
		return;
	}

	/* get_misc(misc); */
	/* put_misc(misc & 0xF3) | ((u8)clocknum << 2)); */
	par->misc = (par->misc & 0xF3) | ((u8)clocknum << 2);
}

static void stpc_set_dot_clock(struct stpcfb_par *par, int clocknum, long frequency)	/* frequency in Hertz */
{
	stpc_platform hardware_id;
	long M, N, P, minM = 0xFF, minN = 0xFF, minP = 0xFF;
	long VCOOut, PhaseRef;
	long error, error_min;
	long maxP;
	int  has_hcmos6pll    = 0;
	int  has_hcmos7newpll = 0;

	/* The way to program the PLL differs from chip to chip */

	hardware_id = stpc_get_hardware_id();
	if ((hardware_id == stpc_platform_gloria)    ||
	    (hardware_id == stpc_platform_eval110)   ||
	    (hardware_id == stpc_platform_eval200pc) ||
	    (hardware_id == stpc_platform_eval120))
		has_hcmos6pll = 1;	    /* HCMOS 6 PLL     */
	else
		has_hcmos6pll = 0;

	if ((hardware_id == stpc_platform_eval130)    ||
	    (hardware_id == stpc_platform_eval210))
		has_hcmos7newpll = 1;	    /* HCMOS 7 new  PLL */
	else
		has_hcmos7newpll = 0;

	if (has_hcmos6pll)
	{
		maxP = 4;
	}
	else
	{
		maxP = 5;
	}

	/*
	 * Find M, N and P best parameters to program the PLL
	 * by scanning the whole range of possibilities and finding
	 * the nearest generated value
	 * M = input divisor, N = feedback divisor, P = post divisor
	 * (see the PLL programmation description)
	 */

	/* Loop to cover the whole range of M values */
	error_min = 100000000L;
	for (M = 1; M <= 255; M++)
	{
		PhaseRef = 14318180L / M;   /* 14.31818MHz = PLL input frequency */

		/* Phase ref need to reach this range to be a valid M */
		/* (see the PLL programming description) */
		if ((PhaseRef < 1000000L) || (PhaseRef > 2000000L))
			continue;

		/* Loop to cover the whole range of N values */
		for (N = 1; N <= 255; N++)
		{
			/* VCO output need to reach this range to be a valid N */
			/* (see the PLL programmation description) */
			if (has_hcmos6pll)
			{
				VCOOut = PhaseRef * N;
				if ((VCOOut < 128000000L) || (VCOOut > 256000000L))
					continue;

			}
			else
			{
				VCOOut = PhaseRef * N * 2;
				if ((VCOOut < 200000000L) || (VCOOut > 622000000L))
					continue;

			}

			/* Loop to cover the whole range of P value */
			for (P = 0; P <= maxP; P++)
			{
				/* Compute the difference between frequency */
				/* generated by this set of (M, N, P) and   */
				/* the required value                       */
				error = (VCOOut / (1 << P)) - frequency;
				if (error < 0)
					error = -error;

				/* If nearest than previous value, select it */
				if (error < error_min)
				{
					error_min = error;
					minM = M;
					minN = N;
					minP = P;
				}
			}
		}
	}

	/* Set new M, N and P PLL parameters */
	if (has_hcmos7newpll)
	{
		par->host[clocknum * 2]     = (u8)(((minM & 0x0f) << 3) | minP);
		par->host[clocknum * 2 + 1] = (u8)minN;
	}
	else
	{
		par->host[clocknum * 2]     = (u8)((((minM & 0x0f) << 3) |
						    ((minN & 0xe0) >> 5)));
		par->host[clocknum * 2 + 1] = (u8)((((minP & 0x01) << 7) |
						    ((minN & 0x1f) << 2) |
						    ((minP & 0x06) >> 1)));
	}
}

#if __UNUSED__
static void stpc_set_refresh_rate(struct stpcfb_par *par, unsigned long refresh_rate)
{
	int  DCLK;
	long frequency;

	DCLK = stpc_get_current_dot_clock(par);

	frequency = stpc_refresh_to_dclk(par, refresh_rate);

	/* Set DCLK */
	stpc_set_dot_clock(par, DCLK, frequency);
}
#endif

#ifdef DEBUG
static u32 stpc_dclk_to_refresh(struct stpcfb_par *par, unsigned long dclk)
{
	u8   cr6, cr7, cr25;
	u8   sr1, cr0;
	u32  vertical_total, horizontal_total;
	u32  refresh_rate;

	/* Get current vertical total (number of lines including blank) */
	cr6  = par->crt[0x06];
	cr7  = par->crt[0x07];
	cr25 = par->crt[0x25];
	vertical_total = (u32)cr6 + (((u32)cr7 & 0x1) << 8) + (((u32)cr7 & 0x20) << 4) + (((u32)cr25 & 0x1) << 10);

	/* Get current horizontal total (number of pixels per lines including blank) */
	sr1 = par->sequencer[0x01];
	cr0 = par->crt[0x00];
	if (sr1 & 0x01)
		horizontal_total = ((u32)cr0 + 5) * 8;
	else
		horizontal_total = ((u32)cr0 + 5) * 9;

	/* clock frequency = horizontal * vertical * refresh */
	refresh_rate = dclk / (horizontal_total * vertical_total);

	return refresh_rate;
}
#endif

static u32 stpc_refresh_to_dclk(struct stpcfb_par *par, unsigned long refresh_rate)
{
	u8   cr6, cr7, cr25;
	u8   sr1, cr0;
	u32  vertical_total, horizontal_total;
	long frequency;

	/* Get current vertical total (number of lines including blank) */
	cr6  = par->crt[0x06];
	cr7  = par->crt[0x07];
	cr25 = par->crt[0x25];
	vertical_total = (u32)cr6 + (((u32)cr7 & 0x1) << 8) + (((u32)cr7 & 0x20) << 4) + (((u32)cr25 & 0x1) << 10);

	/* Get current horizontal total (number of pixels per lines including blank) */
	sr1 = par->sequencer[0x01];
	cr0 = par->crt[0x00];
	if (sr1 & 0x01)
		horizontal_total = ((u32)cr0 + 5) * 8;
	else
		horizontal_total = ((u32)cr0 + 5) * 9;

	/* Required clock frequency = horizontal * vertical * refresh */
	frequency = refresh_rate * horizontal_total * vertical_total;

	return frequency;
}

static void stpc_load_regs(struct stpcfb_par *regs)
{
	int j;

	put_misc(regs->misc);			/* the  misc register */
	stpc_sequencer_sync_reset_on();
	for (j = 0; j <= 0x04; j++)		/* sequencer registers */
		put_sequence(j, regs->sequencer[j]);
	stpc_sequencer_reset_off();
	stpc_unlock_crt_registers();
	for (j = 0; j <= 0x44; j++)		/* crt registers */
		put_crt(j, regs->crt[j]);
	for (j = 0; j <= 0x08; j++)		/* graphic registers */
		put_graph(j, regs->graphic[j]);
	for (j = 0; j <= 0x14; j++) 		/* attrib registers */
		put_attr(j, regs->attrib[j]);

	/* set the DCLK registers, select the right
	 * pair using the 3-2 bits of the MISC register */
	j = (regs->misc >> 2) & 0x03;
	stpc_put_host(0x42 + j * 2,
		       regs->host[j * 2]);
	stpc_put_host(0x43 + j * 2,
		       regs->host[j * 2 + 1]);
}

static void stpc_load_par(struct stpcfb_par *par)
{
	DPRINTK("ENTERED\n");

	stpc_disable_video();
	stpc_load_regs(par);
	stpc_enable_video();
}

static void __init stpc_hardware_init(void)
{
	volatile unsigned char aux;

	DPRINTK("ENTERED\n");

	aux = (volatile unsigned char) inb(V_MISC_R);
	/* Check the I/O AS bit in order to set the variable registers */
	if (aux & 1) {  /* color compatibility */
		V_CRT_INDEX = 0x3D4;
		V_CRT_RW    = 0x3D5;
		V_FEATURE_W = V_ISTAT1_R  = 0x3DA;
	} else {        /* mono compatibility */
		V_CRT_INDEX = 0x3B4;
		V_CRT_RW    = 0x3B5;
		V_FEATURE_W = V_ISTAT1_R  = 0x3BA;
	}

	/* unlock the CRT extended registers */
	put_sequence(0x06, 0x57);

	/* enable mapping of registers in high memory address */
	get_crt(0x1f, aux);
	put_crt(0x1f, aux | 0x80);

	/* enable access to the whole frame buffer */
	get_crt(0x1a, aux);
	put_crt(0x1a, aux & 0xfe);

	/* obtain the (physical) frame buffer address */
	get_crt(0x20, stpc_gbase);
	stpc_gbase &= 0x07;		    /* get only the right bits */
	stpc_fb_addr           = (stpc_gbase | 0x08) << 24;
	stpc_mmap_regs         = stpc_fb_addr + 0x400000;
	stpc_offscreen_fb_addr = stpc_fb_addr + 0x800000;

	/* obtain the video size */
	stpc_fb_size = stpc_get_framebuffer_size();
}

static unsigned long stpc_get_framebuffer_size(void)
{
	unsigned long size;

	stpc_get_host(STPC_SDRAM_GRAPH_MEM, size);
	size = (size & 0x3F) << 17;
	return size;
}

#if __UNUSED__
static u32 stpc_get_video_hardware_id(void)
{
	unsigned char implement;	/* graphics implementation number */
	unsigned char version;		/* graphics version number */
	u32 hwid;

	get_crt(0x3A, implement);
	get_crt(0x3B, version);
	hwid = (implement << 8) | version;

	return hwid;
}
#endif

/* ---------------------------- Misc functions ------------------------------ */

static stpc_platform stpc_get_hardware_id(void)
{
	static stpc_platform hwid = stpc_platform_invalid;
	struct pci_dev *northbridge_dev, *southbridge_dev, *pcmcia_dev;
	u32             northbridge_id,   southbridge_id,   pcmcia_id;
	u32             id;

	if (hwid != stpc_platform_invalid)
		return hwid;

	northbridge_dev = pci_find_slot(/* bus */ 0, /* dev + fn */ PCI_DEVFN(0x0b, 0));
	southbridge_dev = pci_find_slot(/* bus */ 0, /* dev + fn */ PCI_DEVFN(0x0c, 0));
	pcmcia_dev      = pci_find_slot(/* bus */ 0, /* dev + fn */ PCI_DEVFN(0x0d, 0));

	if (! (northbridge_dev && southbridge_dev && pcmcia_dev))
		return stpc_platform_unknown;

	if (pci_read_config_dword(northbridge_dev, 0, &northbridge_id) ||
	    pci_read_config_dword(southbridge_dev, 0, &southbridge_id) ||
	    pci_read_config_dword(pcmcia_dev,      0, &pcmcia_id))
		return stpc_platform_unknown;

	id = (northbridge_id & (u32)0xffff0000) + (southbridge_id >> 16);
	DPRINTK("northbridge_id: 0x%08lX  southbridge_id: 0x%08lX  pcmcia_id: 0x%08lX  id: 0x%08lX\n",
		(unsigned long)northbridge_id, (unsigned long)southbridge_id, (unsigned long)pcmcia_id, (unsigned long)id);
	switch (id)
	{
	case 0x056455cc:		    /* STPC Client */
		printk(KERN_INFO MNAME ": detected STPC Client (Gloria)\n");
		hwid = stpc_platform_gloria;
		break;

	case 0x02090209:		    /* STPC Industrial || STPC Consumer */
		if ((pcmcia_id == 0x220104a) || (stpc_read_host_2x8(0x1d)))
		{
			printk(KERN_INFO MNAME ": detected STPC Industrial (EVAL200PC)\n");
			hwid = stpc_platform_eval200pc;
	}
		else
		{
			printk(KERN_INFO MNAME ": detected STPC Consumer (EVAL110)\n");
			hwid = stpc_platform_eval110;
		}
		break;

	case 0x020a0210:		    /* STPC ATLAS */
		printk(KERN_INFO MNAME ": detected STPC ATLAS (EVAL210)\n");
		hwid = stpc_platform_eval210;
		break;

	case 0x020a021a:		    /* STPC Consumer S */
		printk(KERN_INFO MNAME ": detected STPC Consumer S (EVAL120)\n");
		hwid = stpc_platform_eval120;
		break;

	case 0x020a021b:		    /* STPC Consumer II */
		printk(KERN_INFO MNAME ": detected STPC Consumer II (EVAL130)\n");
		hwid = stpc_platform_eval130;
		break;

	default:			    /* Unknown */
		printk(KERN_INFO MNAME ": unknown chip (id: 0x%08lX)\n", (unsigned long)id);
		hwid = stpc_platform_unknown;
		break;
	}

	return hwid;
}

/* stpc_get_modes_count -- return the number of video modes available */
static int stpc_get_modes_count(void)
{
	int i;

	for (i = 0; stpcfb_modes[i]; i++)
		/* empty */;

	return i;
}

/* Fill the RGB parameters */
static void store_rgb_param(struct fb_var_screeninfo *var,
			    struct stpcfb_par *p)
{
	switch (p->bpp) {
	case 16:
		if (p->mode == 565) {
			var->red.offset   = 11;
			var->red.length   = 5;
			var->green.offset = 5;
			var->green.length = 6;
			var->blue.offset  = 0;
			var->blue.length  = 5;
		}
		break;
	case 24:
		var->blue.offset   = 0;
		var->blue.length   = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->red.offset  = 16;
		var->red.length  = 8;
		break;
	case 32:
		if (p->mode == 8888) {
			var->blue.offset = 0;
			var->blue.length = 8;
			var->green.offset = 8;
			var->green.length = 8;
			var->red.offset = 16;
			var->red.length = 8;
			var->transp.offset = 24;
			var->transp.length = 8;
		}
		break;
	}
}

static unsigned long stpc_get_offset(struct stpcfb_par *par)
{
	int CR13 = par->crt[0x13];	    /* 5-4 are the 9-8 offset bits */
	int CR19 = par->crt[0x19];	    /* 7-0 are the 7-0 offset bits */
	return (CR13 | ((CR19 & 0x30) << 4)) << 3;
}

/* --------------------------- Framebuffer init -------------------------- */

struct fb_ops stpcfb_ops = {
	owner:		THIS_MODULE,
	fb_get_fix:	stpc_get_fix,
	fb_get_var:	stpc_get_var,
	fb_set_var:	stpc_set_var,
	fb_get_cmap:	stpc_get_cmap,
	fb_set_cmap:	stpc_set_cmap,
};

/* my_atoi comes from modedb.c */
static int __init my_atoi(const char *name)
{
	int val = 0;

	for (;; name++) {
		switch (*name) {
		case '0'...'9':
			val = 10*val+(*name-'0');
			break;
		default:
			return val;
		}
	}
}

struct stpcfb_par *stpcfb_modedb_like(const char *mode)
{
	int xres = 0, yres = 0, bpp = 0, refresh = 0;
	int i;

	/* Parse xres */
	while (*mode && (*mode < '0' || *mode > '9'))
		mode++;
	xres = my_atoi(mode);

	/* Parse yres */
	while (*mode) {
		switch (*mode) {
		case 'x':
		case 'X':
			mode++;
			yres = my_atoi(mode);
			break;
		case '0' ... '9':
			break;
		}
		if (yres)
			break;
		mode++;
	}

	/* Parse bpp */
	if (*mode == '-')
	{
		while (*mode) {
			switch (*mode) {
			case '-':
				mode++;
				bpp = my_atoi(mode);
				break;
			case '0' ... '9':
				break;
			}
			if (bpp)
				break;
			mode++;
		}
	}

	/* Parse refresh */
	if (*mode == '@')
	{
		while (*mode) {
			switch (*mode) {
			case '@':
				mode++;
				refresh = my_atoi(mode);
				break;
			case '0' ... '9':
				break;
			}
			if (refresh)
				break;
			mode++;
		}
	}

	if (xres)
		opt_xres    = xres;
	if (yres)
		opt_yres    = yres;
	if (bpp)
		opt_bpp     = bpp;
	if (refresh)
		opt_refresh = refresh;

	for (i = 0; stpcfb_modes[i]; i++) {
		if (stpcfb_modes[i]->xres == xres &&
		    stpcfb_modes[i]->yres == yres &&
		    stpcfb_modes[i]->bpp  == bpp)
			break;
	}
	return stpcfb_modes[i]; /* This is NULL if the mode wasn't found */
}

int __init stpcfb_setup(char *options)
{
	char *this_opt;

	DPRINTK("ENTERED\n");

	if (!options || !*options)
		goto out;

	for (this_opt=strtok(options,","); this_opt; this_opt=strtok(NULL,",")) {
		if (!*this_opt) continue;
		
		if (! strncmp(this_opt, "font:", 5)) {
			strncpy(fontname, this_opt+5, FONTNAME_SZ);
			fontname[FONTNAME_SZ-1] = '\0';
		}
		else if (! strncmp(this_opt, "xres:", 5)) {
			opt_xres = my_atoi(this_opt+5);
		}
		else if (! strncmp(this_opt, "yres:", 5)) {
			opt_yres = my_atoi(this_opt+5);
		}
		else if (! strncmp(this_opt, "depth:", 6)) {
			opt_bpp = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "bpp:", 4)) {
			opt_bpp = my_atoi(this_opt+4);
		}
		else if (! strncmp(this_opt, "refresh:", 8)) {
			opt_refresh = my_atoi(this_opt+8);
		}
		else if (! strncmp(this_opt, "pixclock:", 9)) {
			opt_pixclock = my_atoi(this_opt+9);
		}
		else if (! strncmp(this_opt, "left:", 5)) {
			opt_left = my_atoi(this_opt+5);
		}
		else if (! strncmp(this_opt, "right:", 6)) {
			opt_right = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "upper:", 6)) {
			opt_upper = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "lower:", 6)) {
			opt_lower = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "hslen:", 6)) {
			opt_hslen = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "vslen:", 6)) {
			opt_vslen = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "hsync:", 6)) {
			opt_hsync = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "vsync:", 6)) {
			opt_vsync = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "csync:", 6)) {
			opt_csync = my_atoi(this_opt+6);
		}
		else if (! strncmp(this_opt, "gsync", 5)) {
			opt_gsync = 1;
		}
		else if (! strncmp(this_opt, "extsync", 7)) {
			opt_extsync = 1;
		}
		else if (! strncmp(this_opt, "laced", 5)) {
			opt_laced = 1;
		}
		else if (! strncmp(this_opt, "double", 6)) {
			opt_double = 1;
		}
		else if (! strncmp(this_opt, "preserve", 8)) {
			opt_preserve = 1;
		}
		else
		{
			mode_option = this_opt;
			/* TODO FIXME: parse mode_option */
		}
	}
out:
	return 0;
}

static int __init init_stpc(struct stpcfb_info *vi)
{
	int i;
	int found_mode;
	u32 sync;
	u32 vmode;
	int sync_user;
	int vmode_user;

	DPRINTK("ENTERED\n");

	/* modedb like initial video mode selection */
	if (mode_option) {
		struct stpcfb_par *mo_par;

		if ((mo_par = stpcfb_modedb_like(mode_option)) != NULL)
		{
			default_mode = mo_par;
			found_mode = 1;
		} else
			found_mode = 0;
	} else
		found_mode = 1;

	printk("stpcfb: default mode is %dx%dx%d\n",
	       opt_xres, opt_yres, opt_bpp);

	strcpy(vi->fix.id, "STPC VGA");

	/* Fill fixed part ("fix" structure) in stpcfb_info */

	vi->fix.smem_start = stpc_fb_addr;
	vi->fix.smem_len   = stpc_fb_size;

	vi->fix.type        = FB_TYPE_PACKED_PIXELS;
	vi->fix.type_aux    = 0;
	vi->fix.visual      = FB_VISUAL_TRUECOLOR;
	vi->fix.xpanstep    = vi->fix.ypanstep = vi->fix.ywrapstep = 0;
	if (found_mode)
		vi->fix.line_length = stpc_get_offset(default_mode);
	else
		vi->fix.line_length = opt_xres * opt_bpp / 8; /* TODO FIXME: BOGUS */
	vi->fix.accel       = 0;

	/* Fill variable part ("var" structure) in stpcfb_info */

	sync_user = 0;
	sync = 0;
	if (opt_hsync)
	{
		sync |= FB_SYNC_HOR_HIGH_ACT;
		sync_user = 1;
	}
	if (opt_vsync)
	{
		sync |= FB_SYNC_VERT_HIGH_ACT;
		sync_user = 1;
	}
	if (opt_csync)
	{
		sync |= FB_SYNC_COMP_HIGH_ACT;
		sync_user = 1;
	}
	if (opt_gsync)
	{
		sync |= FB_SYNC_ON_GREEN;
		sync_user = 1;
	}
	if (opt_extsync)
	{
		sync |= FB_SYNC_EXT;
		sync_user = 1;
	}

	if (! sync_user)
	{
		sync = 0;
		if (default_mode->hsync)
			sync |= FB_SYNC_HOR_HIGH_ACT;
		if (default_mode->vsync)
			sync |= FB_SYNC_VERT_HIGH_ACT;
		if (default_mode->csync)
			sync |= FB_SYNC_COMP_HIGH_ACT;
		if (default_mode->gsync)
			sync |= FB_SYNC_ON_GREEN;
		if (default_mode->extsync)
			sync |= FB_SYNC_EXT;
	}

	vmode_user = 0;
	vmode = FB_VMODE_NONINTERLACED;
	if (opt_laced)
	{
		vmode = FB_VMODE_INTERLACED;
		vmode_user = 1;
	}
	if (opt_double)
	{
		vmode = FB_VMODE_DOUBLE;
		vmode_user = 1;
	}

	if (! vmode_user)
	{
		if (default_mode->vm_laced)
			vmode = FB_VMODE_INTERLACED;
		if (default_mode->vm_double)
			vmode = FB_VMODE_DOUBLE;
	}

	vi->var.xres           = opt_xres;
	vi->var.yres           = opt_yres;
	vi->var.xres_virtual   = vi->var.xres;
	vi->var.yres_virtual   = vi->var.yres;
	vi->var.nonstd         = 0;
	vi->var.bits_per_pixel = opt_bpp;
	store_rgb_param(&vi->var, default_mode);
	vi->var.height         = vi->var.width = -1;
	vi->var.activate       = FB_ACTIVATE_NOW;
	vi->var.vmode          = vmode;
	if (opt_pixclock)
		vi->var.pixclock       = opt_pixclock;
	else
		vi->var.pixclock       = default_mode->pixclock;
	if (opt_left)
		vi->var.left_margin    = opt_left;
	else
		vi->var.left_margin    = default_mode->left;
	if (opt_right)
		vi->var.right_margin   = opt_right;
	else
		vi->var.right_margin   = default_mode->right;
	if (opt_upper)
		vi->var.upper_margin   = opt_upper;
	else
		vi->var.upper_margin   = default_mode->upper;
	if (opt_lower)
		vi->var.lower_margin   = opt_lower;
	else
		vi->var.lower_margin   = default_mode->lower;
	if (opt_hslen)
		vi->var.hsync_len      = opt_hslen;
	else
		vi->var.hsync_len      = default_mode->hslen;
	if (opt_vslen)
		vi->var.vsync_len      = opt_vslen;
	else
		vi->var.vsync_len      = default_mode->vslen;
	vi->var.sync                   = sync;

	if ((! opt_pixclock) && opt_refresh)
	{
		u32 horizontal_total, vertical_total;
		u32 dclk_freq;

		/* The user specified the refresh rate but not the dot clock */
		horizontal_total = vi->var.xres + vi->var.right_margin + vi->var.hsync_len + vi->var.left_margin;
		vertical_total   = vi->var.yres + vi->var.lower_margin + vi->var.vsync_len + vi->var.upper_margin;

		dclk_freq = horizontal_total * vertical_total * opt_refresh;
		dclk_freq /= 1000L;
		vi->var.pixclock = 1000000000L / dclk_freq;

		DPRINTK("User specified refresh (%d Hz) -> dot clock: %lu, pixclock (%lu)\n",
			(int)opt_refresh,
			(unsigned long)dclk_freq,
			(unsigned long)vi->var.pixclock);
	}


	/* Fill display structure in stpcfb_info */

	vi->disp.var           = vi->var;   /* propagate `var' part to display field */
	vi->disp.cmap.red      = NULL;
	vi->disp.cmap.green    = NULL;
	vi->disp.cmap.blue     = NULL;
	vi->disp.cmap.transp   = NULL;

	vi->disp.screen_base   = vi->vbase;

	vi->disp.visual         = vi->fix.visual;
	vi->disp.type           = vi->fix.type;
	vi->disp.ypanstep       = vi->fix.ypanstep;
	vi->disp.ywrapstep      = vi->fix.ywrapstep;
	vi->disp.type_aux       = vi->fix.type_aux;
	vi->disp.line_length    = vi->fix.line_length;
	vi->disp.next_line      = vi->fix.line_length;
	vi->disp.can_soft_blank = 1;
	vi->disp.inverse        = 0;
	switch (opt_bpp) {
#ifdef FBCON_HAS_CFB16
	case 16:
		vi->disp.dispsw      = &fbcon_cfb16;
		vi->disp.dispsw_data = vi->fbcon_cmap.cfb16; /* console palette */
		break;
#endif
#ifdef FBCON_HAS_CFB24
	case 24:
		vi->disp.dispsw      = &fbcon_cfb24;
		vi->disp.dispsw_data = vi->fbcon_cmap.cfb24; /* console palette */
		break;
#endif
#ifdef FBCON_HAS_CFB32
	case 32:
		vi->disp.dispsw      = &fbcon_cfb32;
		vi->disp.dispsw_data = vi->fbcon_cmap.cfb32; /* console palette */
		break;
#endif
	default:
		vi->disp.dispsw      = &fbcon_dummy;
	}
	if (opt_preserve) {
		vi->disp.dispsw = &fbcon_dummy;
  }
	vi->disp.scrollmode = SCROLL_YREDRAW;

	/* Fill parameter structure in stpcfb_info */

	vi->par = *default_mode;

	/* Fill fb_info structure in stpcfb_info */

	if (fontname[0] != '\0')
		strcpy(vi->info.fontname, fontname);
	else
		strcpy(vi->info.fontname, "PEARL8x8");
#if 0
		strcpy(vi->info.fontname, "VGA4x8");	/* 80x25 in 320x200 */
#endif

	strcpy(vi->info.modename, "STPC VGA");
	vi->info.node  = -1;
	vi->info.fbops = &stpcfb_ops;

	vi->info.disp       = &vi->disp;    /* Propagate display structure in fb_info part */
	vi->info.changevar  = NULL;
	vi->info.switch_con = &stpc_switch;
	vi->info.updatevar  = &stpc_updatevar;
	vi->info.blank      = &stpc_blank;
	vi->info.flags      = FBINFO_FLAG_DEFAULT;

	vi->blanked = 0;

	/* Set defaults for the first 16 colors */
	for (i = 0; i < 16; i++) {
		int j = color_table[i];
		vi->palette[i].red   = default_red[j];
		vi->palette[i].green = default_grn[j];
		vi->palette[i].blue  = default_blu[j];
	}

	if (register_framebuffer(&vi->info) < 0) {
		printk("stpcfb: can't register the framebuffer\n");
		return -EINVAL;
	}

	printk("fb%d: STPC frame buffer\n", GET_FB_IDX(vi->info.node));
	return 0;
}

static int __init stpc_of_init(void)
{
	memset(&fb_info, 0, sizeof(struct stpcfb_info));
	fb_info.vbase     = ioremap(stpc_fb_addr,           stpc_fb_size);
	fb_info.regs      = ioremap(stpc_mmap_regs,         0x400000);
	fb_info.offscreen = ioremap(stpc_offscreen_fb_addr, stpc_fb_size);
	printk(KERN_INFO "stpcfb: %luK SDRAM mapped to 0x%p\n",
	       stpc_fb_size/1024, fb_info.vbase);
	printk(KERN_INFO "stpcfb: %d available modes\n",
	       stpc_get_modes_count());
	return init_stpc(&fb_info);
}

int __init stpcfb_init(void)
{
	/* Performs the VGA hardware initialization, but does *NOT*
	 * set the registers for the new video mode.
	 * The outgoing vga console needs to read from the text video memory.
	 * 
	 * register_framebuffer() will call the stpcfb_switch() function
	 * that will set the new video mode */
	DPRINTK("ENTERED\n");

	/* probe for STPC boards */
	if (stpc_get_hardware_id() == stpc_platform_unknown)
	{
		printk("stpcfb: unknown platform, bailing out.\n");
		return -EINVAL;
	}

	stpc_hardware_init();

	return stpc_of_init();
}

/* ========================================================================
 *   M O D U L A R I Z A T I O N
 * ------------------------------------------------------------------------ */

MODULE_AUTHOR("Copyright (C) 2001 Marco Pantaleoni <panta@elasticworld.org>, Salvatore Sanfilippo <antirez@invece.org>.");
MODULE_DESCRIPTION("FBDev driver for STPC integrated video units");

#ifdef MODULE
/* bogus, you can't remove the framebuffer AFAIK */
static void __exit stpcfb_exit(void)
{
	/* FIXME XXX: Restore the original VGA settings */
	if (!fb_info.par.chained)
		stpc_le_disable_timer();

	if (fb_info.virtualfb)
		rvfree(fb_info.virtualfb, fb_info.virtualfb_size);

	unregister_framebuffer(&fb_info.info);
	iounmap(fb_info.offscreen);
	iounmap(fb_info.regs);
	iounmap(fb_info.vbase);
}
#endif

/*
 * I don't think we can use the "module_init" stuff here because
 * the fbcon stuff may not be initialized yet.  Hence the #ifdef
 * around module_init.
 */
#ifdef MODULE
module_init(stpcfb_init);
#endif
#ifdef MODULE
module_exit(stpcfb_exit);
#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * tab-width: 8
 * c-indent-level: 4
 * c-argdecl-indent: 0
 * c-label-offset: -4
 * c-continued-statement-indent: 0
 * c-brace-offset: -2
 * c-brace-imaginary-offset: 0
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * comment-column: 44
 * End:
 */
