/*
 *  gunzip definitions
 */
#define OF(args)  args

#ifndef memzero
#define memzero(s, n)     memset ((s), 0, (n))
#endif

typedef unsigned char  uch;
typedef unsigned short ush;
typedef unsigned long  ulg;

#define INBUFSIZ 4096
#define WSIZE 0x8000    /* window size--must be a power of two, and */
            /*  at least 32K for zip's deflate method */

static uch *inbuf;
static uch *window;

static unsigned insize;  /* valid bytes in inbuf */
static unsigned inptr;   /* index of next byte to be processed in inbuf */
static unsigned outcnt;  /* bytes in output buffer */
static int exit_code;
static long bytes_out;
static struct file *crd_infp, *crd_outfp;

#define get_byte()  (inptr < insize ? inbuf[inptr++] : fill_inbuf())

/* Diagnostic functions (stubbed out) */
#define Assert(cond,msg)
#define Trace(x)
#define Tracev(x)
#define Tracevv(x)
#define Tracec(c,x)
#define Tracecv(c,x)

#define STATIC static

static int  fill_inbuf(void);
static void flush_window(void);
static void *malloc(int size);
static void free(void *where);
static void error(char *m);
static void gzip_mark(void **);
static void gzip_release(void **);

#include "../lib/inflate.c"

static void __init *malloc(int size)
{
    return kmalloc(size, GFP_KERNEL);
}

static void __init free(void *where)
{
    kfree(where);
}

static void __init gzip_mark(void **ptr)
{
}

static void __init gzip_release(void **ptr)
{
}

/* ===========================================================================
 * Fill the input buffer. This is called only when the buffer is empty
 * and at least one byte is really needed.
 */
static int __init fill_inbuf(void)
{
    if (exit_code) return -1;

    insize = crd_infp->f_op->read(crd_infp, inbuf, INBUFSIZ,
                      &crd_infp->f_pos);
    if (insize == 0) return -1;

    inptr = 1;
    return inbuf[0];
}

/* ===========================================================================
 * Write the output window window[0..outcnt-1] and update crc and bytes_out.
 * (Used for the decompressed data only.)
 */
static void __init flush_window(void)
{
    ulg c = crc;         /* temporary variable */
    unsigned n;
    uch *in, ch;

    crd_outfp->f_op->write(crd_outfp, window, outcnt, &crd_outfp->f_pos);
    in = window;
    for (n = 0; n < outcnt; n++) {
        ch = *in++;
        c = crc_32_tab[((int)c ^ ch) & 0xff] ^ (c >> 8);
    }
    crc = c;
    bytes_out += (ulg)outcnt;
    outcnt = 0;
}


static void __init error(char *x)
{
    printk(KERN_ERR "%s", x);
    exit_code = 1;
}


static int __init
crd_load(struct file * fp, struct file *outfp)
{
    int result;

    insize = 0;     /* valid bytes in inbuf */
    inptr = 0;      /* index of next byte to be processed in inbuf */
    outcnt = 0;     /* bytes in output buffer */
    exit_code = 0;
    bytes_out = 0;
    crc = (ulg)0xffffffffL; /* shift register contents */

    crd_infp = fp;
    crd_outfp = outfp;
    inbuf = kmalloc(INBUFSIZ, GFP_KERNEL);
    if (inbuf == 0) {
        printk(KERN_ERR "UNTAR: Couldn't allocate gzip buffer\n");
        return -1;
    }
    window = kmalloc(WSIZE, GFP_KERNEL);
    if (window == 0) {
        printk(KERN_ERR "UNTAR: Couldn't allocate gzip window\n");
        kfree(inbuf);
        return -1;
    }
    makecrc();
    result = gunzip();
    kfree(inbuf);
    kfree(window);
    return result;
}


/*
 *  untar definitions
 */


/* untar debugging: uncomment both lines */
//#define DEBUG_UNTAR 
//#include <linux/delay.h>


#include <linux/utime.h>

static unsigned long simple_strtoul_wrapper(const char *cp,char **endp,unsigned int base);

#define E_NAMETOOLONG   1
#define E_CANNOTSTAT    2
#define E_READLINK  3
#define E_SHORTWRITE    4
#define E_FOPEN     5
#define E_MALLOC    6
#define E_OPENDIR   7
#define E_WRONG_FTYPE   8

#define TARBLOCKSIZE    512

#define NAME_FIELD_SIZE   100
#define PREFIX_FIELD_SIZE 155
#define UNAME_FIELD_SIZE   32
#define GNAME_FIELD_SIZE   32

#define TMAGIC   "ustar "   /* ustar and a null */
#define TMAGLEN  6
#define TVERSION " \0"      /* 00 and no null */
#define TVERSLEN 2

#define REGTYPE  '0'        /* regular file */
#define AREGTYPE '\0'       /* regular file */
#define LNKTYPE  '1'        /* link */
#define SYMTYPE  '2'        /* reserved */
#define CHRTYPE  '3'        /* character special */
#define BLKTYPE  '4'        /* block special */
#define DIRTYPE  '5'        /* directory */
#define FIFOTYPE '6'        /* FIFO special */
#define CONTTYPE '7'        /* reserved */

#define BUF_SIZE (1024*64)

//#define strtoul simple_strtoul
#define strtoul simple_strtoul_wrapper

/* POSIX header  */
struct fileHeader
{                   /* byte offset */
  char name[NAME_FIELD_SIZE];       /*   0 */
  char mode[8];             /* 100 */
  char uid[8];              /* 108 */
  char gid[8];              /* 116 */
  char size[12];            /* 124 */
  char mtime[12];           /* 136 */
  char chksum[8];           /* 148 */
  char typeflag;            /* 156 */
  char linkname[NAME_FIELD_SIZE];   /* 157 */
  char magic[TMAGLEN];          /* 257 */
  char version[TVERSLEN];       /* 263 */
  char uname[UNAME_FIELD_SIZE];     /* 265 */
  char gname[GNAME_FIELD_SIZE];     /* 297 */
  char devmajor[8];         /* 329 */
  char devminor[8];         /* 337 */
  char prefix[PREFIX_FIELD_SIZE];   /* 345 */
                    /* 500 */
};


typedef union TarInfo
{
  char buf[TARBLOCKSIZE];
  struct fileHeader header;
} TarInfo;


enum untar_status
{
    READING_HEADER,
    READING_DATA,
    SKIPPING_REST
};


struct untar_context
{
    TarInfo           tarInfo;
    enum untar_status state;
    mode_t            fmode;
    unsigned long     fsize;
    int               out;
    int               rotate;
	unsigned long     initrd_pos;
} *Untar;

extern int sys_write(unsigned int fd, const char * buf,unsigned int count);
extern int sys_chmod(const char * filename, mode_t mode);
extern int sys_chown(const char * filename, uid_t user, gid_t group);
extern int sys_mknod(const char * filename, int mode, dev_t dev);
extern int sys_mkdir(const char * pathname, int mode);
extern int sys_chdir(const char * filename);
extern int sys_link(const char * oldname, const char * newname);
extern int sys_symlink(const char * oldname, const char * newname);
extern int sys_utime(char * filename, struct utimbuf * times);
extern int sys_time(int * tloc);
extern int sys_fchown(unsigned int fd, uid_t user, gid_t group);



unsigned long initrd_start, initrd_end;
int initrd_below_start_ok;


static ssize_t __init initrd_read(
    struct file *fp, /* not used */
    char   *buf,
    size_t count,
    loff_t *fpos)    /* not used */
{
	unsigned long			p;
    unsigned long			remain;
    /* pointer to our context structure is passed thru ioctl fop */
	struct untar_context	*Untar = (struct untar_context*)fp->f_op->ioctl;

    remain=initrd_end-Untar->initrd_pos;
    /* nothing more to read */
    if (remain == 0)
    {
        return(0);
    }
    /* less than count bytes available */
    else if (remain < count)
    {
		/* provide data */
        memcpy((void *)buf, (void *)(Untar->initrd_pos), remain);
		p=Untar->initrd_pos;
        Untar->initrd_pos=initrd_end;
		/* free the remaining pages */
		while (p < initrd_end)
        {
            ClearPageReserved(virt_to_page(p));
            set_page_count(virt_to_page(p), 1);
            free_page(p);
            p+=PAGE_SIZE;
        }
        return(remain);
    }
    /* full count bytes available */
    else
    {
		/* provide data */
        memcpy((void *)buf, (void *)(Untar->initrd_pos), count);
		p=Untar->initrd_pos;
        Untar->initrd_pos+=count;
		/* free all pages, we won't access again */
		while (virt_to_page(p) < virt_to_page(Untar->initrd_pos))
		{
			ClearPageReserved(virt_to_page(p));
			set_page_count(virt_to_page(p), 1);
			free_page(p);
			p+=PAGE_SIZE;
		}
        return(count);
    }
}

/* given a path name, assert that the path to that file really exists.
   We assume that the archive to be unpacked was created by the 
   ELinOS (www.elinos.com) mkefs tool and thus all path names are clean.
   By clean we mean: no ../../ or /./ garbage.
   PATH is the file that is about to be created.
   Returns 0 on succes and non-zero on error.
*/
static int assert_parent_directory (char* path)
{
  char* start = path;                  /* first character in file path */
  char* end = path + strlen(path) - 1; /* last character in path (not 
					  the terminating null) */
  char* slash;                         /* pointer to next slash in path */
  int err = 0;                         /* error */
  char saved;                          /* save the original last string 
					  character here */

  /* paranoia check */
  if (start >= end) return -EINVAL; /* error, null file name */

  /* strip leading dots and slashes (due to our assumption above, leading
     slashes can only result from ../ or ./ constructs. */
  while (*start == '.') start++;
  while (*start == '/') start++;

  /* remove any trailing slashes. mkefs puts them after directory names */
  while (*end == '/') end--;

  /* ok, temporarily cut the string after the last non-slash character */
  saved = *(end + 1);
  *(end + 1) = 0;

  /* now iterate over the path components and attempt to create every
     subdirectory we encounter. At this point our path looks like
     PATH/TO/FILE/file.
     This implies, the current wd must be / for the following to work.
     Since we have stripped all trailing slashes we do not create a
     directory /.../file.
  */
  for (slash = strchr(start, '/'); slash && (! err);
       slash = strchr(slash + 1, '/')) 
  {
    *slash = 0; /* temporarily cut string at slash */
    err = sys_mkdir(start, 0755); /* create directory with default 
				     permissions */
    if (err == -EEXIST)  /* ok, directory did already exist */
      err = 0;
    *slash = '/'; /* re-replace */
  }

  *(end + 1) = saved; /* uncut the string (see above) */

  return err;
}


/*
** This is called by flush_window to flush the gzip output
** buffer (containing part of a tar image).
*/
static ssize_t __init untar_write(
    struct file *fp,        /* not used */
    const char  *buf,
    size_t      count,
    loff_t      *fpos)     /* not used */
{
    int i, scoop, err, file_written = 0;
    char *to;
    kdev_t fdev;
    struct utimbuf ut;
    /* pointer to our context structure is passed thru ioctl fop */
    struct untar_context *Untar = (struct untar_context*)fp->f_op->ioctl;

#ifndef DEBUG_UNTAR
#ifndef CONFIG_ARCH_S390
    char rotator[4] = { '|' , '/' , '-' , '\\' };
    printk("%c\b", rotator[Untar->rotate & 0x3]);
    Untar->rotate++;
#endif
#endif

    /*
    ** Note: the gunzip rotuines call this only when
    **       a) the gunzip buffer is full -> count = WSIZE = 64 * TARBLOCKSIZE
    ** or    b) to flush the last bytes out
    */
    while(count)
    {   /*
        ** What is our current state ?
        */
        err = 0;
        switch(Untar->state)
        {
            case READING_HEADER:
            {   /* get the tar header */
                if(count < TARBLOCKSIZE)
                {
                    count = 0;
                    break;
                }
                to = (char*)&Untar->tarInfo;
                for(i = 0; i < TARBLOCKSIZE; i++)
                {
                    *to++ = *buf++;
                    --count;
                }
                /*
                ** Tar usually pads the output byte to a
                ** multiple of it's block size usually
                ** 10 KB, appending zeroes if necessary.
                ** here we skip those zeroes:
                */
                if(Untar->tarInfo.header.name[0] == '\0'  &&
                   Untar->tarInfo.header.mode[0] == '\0' &&
                   Untar->tarInfo.header.mode[0] == '\0')
                {
                    if(count < TARBLOCKSIZE)
                        count = 0;
                    break;
                }
                /* determine mode and size */
                Untar->fmode = strtoul(Untar->tarInfo.header.mode,NULL,8);
                Untar->fsize = strtoul(Untar->tarInfo.header.size,NULL,8);
#ifdef DEBUG_UNTAR
                printk("\ntype: %c ",Untar->tarInfo.header.typeflag);
                printk("name: %s ",Untar->tarInfo.header.name);
                printk("Untar->fsize: %ld ",Untar->fsize);
                udelay(100000);
#endif

/* shorthand macro definition */
#define assert_parent() do {                                         \
  int assert_err = 0;                                                \
  assert_err = assert_parent_directory(Untar->tarInfo.header.name);  \
  if (assert_err) return assert_err;                                 \
} while (0)
/* end of macro */

                /* check the type */
                switch(Untar->tarInfo.header.typeflag)
                {
                    case AREGTYPE :
                    case REGTYPE  :
		        assert_parent();
                        Untar->out = sys_open(Untar->tarInfo.header.name,
                                 O_CREAT|O_WRONLY|O_TRUNC, Untar->fmode);
                        if(Untar->out < 0)
                        {
                            err = Untar->out;
                            break;
                        }
						err = sys_fchown(Untar->out,
							(uid_t)strtoul(Untar->tarInfo.header.uid, NULL, 8),
							(uid_t)strtoul(Untar->tarInfo.header.gid, NULL, 8));
                        Untar->state = READING_DATA;
                        break;
                    case LNKTYPE  :
		        assert_parent();
                        err = sys_link(Untar->tarInfo.header.linkname,
                                          Untar->tarInfo.header.name);
                        break;

                    case SYMTYPE  :
		        assert_parent();
                        err = sys_symlink(Untar->tarInfo.header.linkname,
                                         Untar->tarInfo.header.name);
                        break;

                    case CHRTYPE  :
                    case BLKTYPE  :
                    case FIFOTYPE :
		        assert_parent();
                        fdev = MKDEV(strtoul(Untar->tarInfo.header.devmajor,NULL,8),
                           strtoul(Untar->tarInfo.header.devminor,NULL,8));
                        err = sys_mknod(Untar->tarInfo.header.name,Untar->fmode,fdev);
			if (err == 0)
			{
				err = sys_chown(Untar->tarInfo.header.name, 
					(uid_t)strtoul(Untar->tarInfo.header.uid, NULL, 8),
					(uid_t)strtoul(Untar->tarInfo.header.gid, NULL, 8));
                		if (err == 0)
				{
					err = sys_chmod(Untar->tarInfo.header.name, Untar->fmode);
				}
			}
                        break;

                    case DIRTYPE  :
                        /* Skip if name is "" "./" "." ".." */
                        if((Untar->tarInfo.header.name[0] != '.' &&
                            Untar->tarInfo.header.name[0] != '\0') &&
                           (Untar->tarInfo.header.name[1] != '/' &&
                            Untar->tarInfo.header.name[1] != '\0') &&
                           (Untar->tarInfo.header.name[1] != '\0')) {
			    assert_parent();
                            err = sys_mkdir(Untar->tarInfo.header.name,Untar->fmode);
                            /* don't treat existing directories as errors */
                            if (err == -EEXIST)
                              err = 0;
                        }
                        break;
                    default:
		        panic("UNTAR: corrupt tar archive (typeflag=%x)\n", 
			      Untar->tarInfo.header.typeflag);
/*                         printk(KERN_ERR "UNTAR: corrupt tar archive (typeflag=%x)\n",  */
/* 							   Untar->tarInfo.header.typeflag); */
                        Untar->state = SKIPPING_REST;
                        return(-1);
                }
                break;
            }
            case READING_DATA:
            {
                scoop = 0;
                while(count > 0 && Untar->fsize > 0)
                {
                    int wrote;
                    scoop = Untar->fsize > TARBLOCKSIZE ? TARBLOCKSIZE : Untar->fsize;
                    wrote = sys_write(Untar->out, buf, scoop);
                    if (wrote < 0)
                    {
                        /* fatal error, skip this file */
                        Untar->fsize = 0;
                        err = wrote;

                        break;
                    }
                    count -= wrote;
                    buf   += wrote;
                    Untar->fsize -= wrote;
                }
                if(Untar->fsize == 0)
                {   /* skip unto the next tar block */
                    sys_close(Untar->out);
                    scoop = count % TARBLOCKSIZE;
                    buf   += scoop;
                    count -= scoop;
                    Untar->state = READING_HEADER;
                    file_written = 1;
                }
                break;
            }
            case SKIPPING_REST:
                count = 0;
                break;
        }
        if(err)
        {
            /*
             * Tell the user what went wrong. If the disk was full, we
             * skip all remaining data. The switch() should be maintained
             * to contain the most common types of errors in order to help
             * the user to track down configuration problems.
             */
            printk("\nError %d making %s", -err, Untar->tarInfo.header.name);
            switch(err)
            {
		case -EEXIST: printk(": File already exists\n"); break;
	        case -EINVAL: printk(": Invalid file name\n"); break;
	        case -ENOENT: 
	        case -ENOTDIR: 
		    printk(": Path component is not a directory\n"); 
		break;
                case -ENOSPC:
                    printk(": No space left on device\n");
                    Untar->state = SKIPPING_REST;
                break;
                case -EPERM:
                    printk(": Operation not permitted\n");
                break;
                case -EACCES:
                    printk(": Permission denied\n");
                break;
                default:
		  printk("\n");
                break;
            }
	    panic("Failed to unpack tar archive\n");
	    return(-1); /* Quit on error */
        }
        if(file_written)
        {
            if(Untar->tarInfo.header.typeflag != LNKTYPE &&
               Untar->tarInfo.header.typeflag != SYMTYPE)
            {
                err = sys_chmod(Untar->tarInfo.header.name, Untar->fmode);
            }
            if(err)
                printk("\nError chown and/or chmod %s", Untar->tarInfo.header.name);
            ut.actime=sys_time(NULL);
            ut.modtime=strtoul(Untar->tarInfo.header.mtime,NULL,8);
            sys_utime(Untar->tarInfo.header.name,&ut);
            file_written = 0;
        }
    }
    return(0);
}


typedef int (*ioctl_fop_t)(struct inode *, struct file *, unsigned int, unsigned  long);


static int __init untar(void)
{
    struct file				infile, outfile;
	struct inode 			in_inode, out_inode;
    struct dentry			in_dentry, out_dentry;
    struct file_operations	in_fops, out_fops;
    struct untar_context	*Untar;
	int						err;


	/* get free buffer for tar extractor */
	if (!(Untar = kmalloc(sizeof(*Untar), GFP_KERNEL)))
	{
		printk(KERN_ERR "UNTAR: Couldn't allocate untar buffer\n");
		return(-1);
	}
	Untar->state  = READING_HEADER; /* begin reading header */
	Untar->rotate = 0;				/* first rotator status */
	Untar->initrd_pos=initrd_start; /* initrd_pos is start of initrd */

    /*
	** build a fake infile descriptor for use with gunzip()
    ** all gunzip will ever do with it is to call the
    ** read function. We misuse the fops.iocl entry
    ** here to pass a pointer to Untar down to initrd_read()
	*/
	memset(&infile, 0, sizeof(infile));
	memset(&in_inode, 0, sizeof(in_inode));
	memset(&in_dentry, 0, sizeof(in_dentry));
	memset(&in_fops, 0, sizeof(in_fops));
	in_inode.i_rdev   = 0;     /* don't care */
    infile.f_mode     = 3;     /* don't care */
	in_fops.read      = initrd_read;
	in_fops.ioctl     = (ioctl_fop_t)Untar;
	infile.f_dentry   = &in_dentry;
    infile.f_op       = &in_fops;
    in_dentry.d_inode = &in_inode;

	/*
    ** build a fake outfile descriptor for use with gunzip()
    ** all gunzip will ever do with it is to call the
    ** write function. We misuse the fops.iocl entry
    ** here to pass a pointer to Untar down to untar_write()
    */
    memset(&outfile, 0, sizeof(outfile));
    memset(&out_inode, 0, sizeof(out_inode));
    memset(&out_dentry, 0, sizeof(out_dentry));
    memset(&out_fops, 0, sizeof(out_fops));
    out_inode.i_rdev   = 0;     /* don't care */
    outfile.f_mode     = 3;     /* don't care */
    out_fops.write     = untar_write;
    out_fops.ioctl     = (ioctl_fop_t)Untar;
    outfile.f_dentry   = &out_dentry;
    outfile.f_op       = &out_fops;
    out_dentry.d_inode = &out_inode;

	printk(KERN_NOTICE "UNTAR: Extracting root archive: ");

	/* now uncompress & untar the initrd image */
	err = crd_load(&infile, &outfile);
	
	/* free untar buffer */
	kfree(Untar);

	if (!err)
        printk("done.\n");

	return(err);
}

/*
 * simple_strtoul() does not strip leading spaces (0x20) from the input
 * string. The 'real' strtoul does. This wrapper works around this.
 * NOTE: This probably should be fixed in: lib/vsprintf.c
 *
 */

static unsigned long __init simple_strtoul_wrapper(
    const char   *cp,
    char         **endp,
    unsigned int base)
{
    while (*cp == ' ') {
        cp++;
    }

    return ( simple_strtoul(cp,endp,base) );
}

